package Client;

use strict;
use warnings;
use JSON;
use IO::Socket qw(SHUT_WR);
use Xormon;
use Data::Dumper;

# Client class to handle client requests and responses
# Attributes:
#   client            - The socket client object.
#   client_address    - The IP address of the connected client.
#   client_port       - The port of the connected client.
#   recv_buffer_size  - The buffer size for receiving data (default: 2048).
#   tasks             - A hash reference mapping task types to Task class names.

sub new {
    my ($class, %args) = @_;
    my $self = {
        client => $args{client},
        client_address => $args{client}->peerhost(),
        client_port => $args{client}->peerport(),
        recv_buffer_size => $args{recv_buffer_size} || 2048,
        tasks => $args{tasks} || {},
    };
    bless $self, $class;
    return $self;
}

# Method to receive data from the client
# Returns a decoded JSON object containing the task data
sub recv_data {
    my ($self) = @_;
    my $data = "";
    $self->{client}->recv($data, $self->{recv_buffer_size});
    return from_json($data);
}

# Method to process the client's request
# It identifies the task type and executes the corresponding Task object
sub process_request {
    my ($self) = @_;
    my $decoded = $self->recv_data();
    my $task_type = $decoded->{task};
    
    Xormon::log("Received task: $decoded->{task} for Device: $decoded->{data}->{hostalias}");

    if (exists $self->{tasks}{$task_type}) {
        my $task = $self->{tasks}{$task_type}->new(data => $decoded, client => $self->{client});
        $task->execute();        
    } else {
        $self->send_response("undefined task");
    }
}

# Method to send a response to the client and close the connection
# Parameters:
#   response - The response message to be sent.
sub send_response {
    my ($self, $response_data) = @_;
    my $response = {
        status => 200,
        headers => {
            'Content-Type' => 'application/json',
            'Host' => $self->{client_address},
        },
        content => to_json($response_data),
    };
    $self->{client}->send(to_json($response));
    $self->{client}->close();
}

# Method to shutdown the client connection
sub shutdown {
    my ($self) = @_;
    eval {
        $self->{client}->shutdown(SHUT_WR);
    };
    if ($@) {
        Xormon::log("Shutdown error: $@", "error");
        die "Shutdown error: $@";
    }
    Xormon::log("Client connection shutdown.");
}

1;
