"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CustomGroupWrapper = void 0;
const pageResponse_constants_1 = require("../../common/constants/pageResponse.constants");
const subsystemMetricsResponse_constants_1 = require("../../common/constants/subsystemMetricsResponse.constants");
const dataSource_1 = require("../../config/dataSource");
const customGroup_entity_1 = require("../../entity/customGroup.entity");
const itemGroup_entity_1 = require("../../entity/itemGroup.entity");
const itemGroupItem_entity_1 = require("../../entity/itemGroupItem.entity");
const itemGroupParent_entity_1 = require("../../entity/itemGroupParent.entity");
const itemGroupRegexParent_entity_1 = require("../../entity/itemGroupRegexParent.entity");
const objectItem_entity_1 = require("../../entity/objectItem.entity");
const typeorm_1 = require("typeorm");
const uuid_1 = require("uuid");
const ACLCheck_1 = require("../acl/ACLCheck");
const Definitions_1 = require("../metrics/Definitions");
const MetricsWrapper_1 = require("../metrics/MetricsWrapper");
const User_1 = require("../users/User");
const class_constants_1 = require("../../common/constants/class.constants");
const subsystem_entity_1 = require("../../entity/subsystem.entity");
const Heatmap_1 = require("../menu/menu/Heatmap");
const ItemGroupWrapper_1 = require("./ItemGroupWrapper");
const dashboard_constants_1 = require("../../common/constants/dashboard.constants");
class CustomGroupWrapper {
    static async getCustomGroup(custom_group_id) {
        return await (0, dataSource_1.getDataSource)()
            .getRepository(customGroup_entity_1.CustomGroup)
            .findOne({
            where: {
                custom_group_id: custom_group_id,
            },
            relations: {
                groups: {
                    items: {
                        _item: true,
                    },
                    parents: {
                        _item: true,
                    },
                    regex_parents: true,
                    tags: true,
                    parent_tags: true,
                },
            },
        });
    }
    static async saveInternalCustomGroup(itemStructure) {
        const check = this.checkItemStructure(itemStructure);
        if (check.valid === false) {
            throw new Error('[CheckInternalCustomGroupError]: ' + check.valid);
        }
        let savedCustomGroup;
        await (0, dataSource_1.getDataSource)().transaction(async (transactionalEntityManager) => {
            const cgId = (0, uuid_1.v4)();
            savedCustomGroup = await transactionalEntityManager.getRepository(customGroup_entity_1.CustomGroup).save({
                custom_group_id: cgId,
                class: itemStructure.class,
                hw_type: itemStructure.hw_type,
                subsystem: itemStructure.subsystem,
                label: 'unlabeled',
                internal: true,
                user_id: null,
            });
            for (const group of itemStructure.groups) {
                const igId = (0, uuid_1.v4)();
                const igItems = [];
                const parentItems = [];
                const parentRegexItems = [];
                if (group.item_ids) {
                    group.item_ids.forEach((itemId) => {
                        igItems.push({
                            item_id: itemId,
                            item_group_id: igId,
                        });
                    });
                }
                if (group.parents && group.parents.item_ids) {
                    group.parents.item_ids.forEach((itemId) => {
                        parentItems.push({
                            item_id: itemId,
                            item_group_id: igId,
                        });
                    });
                }
                if (group.parents && group.parents.regex) {
                    group.parents.regex.forEach((regex) => {
                        parentRegexItems.push({
                            regex: regex.regex,
                            subsystem: regex.subsystem,
                            item_group_id: igId,
                        });
                    });
                }
                await transactionalEntityManager.getRepository(itemGroup_entity_1.ItemGroup).save({
                    item_group_id: igId,
                    custom_group_id: cgId,
                    regex: group.regex,
                });
                if (igItems.length >= 1) {
                    transactionalEntityManager.getRepository(itemGroupItem_entity_1.ItemGroupItem).save(igItems);
                }
                if (parentItems.length >= 1) {
                    transactionalEntityManager.getRepository(itemGroupParent_entity_1.ItemGroupParent).save(parentItems);
                }
                if (parentRegexItems.length >= 1) {
                    transactionalEntityManager.getRepository(itemGroupRegexParent_entity_1.ItemGroupRegexParent).save(parentRegexItems);
                }
            }
        });
        return savedCustomGroup;
    }
    static checkItemStructure(itemStructure) {
        if (itemStructure.class === undefined) {
            return {
                valid: false,
                message: 'class is required',
            };
        }
        for (const group of itemStructure.groups) {
            if (!group.item_ids || !group.regex) {
                return {
                    valid: false,
                    message: 'item_ids or regex is required',
                };
            }
        }
        if (itemStructure.class == class_constants_1.ClassKey.STORAGE || itemStructure.class == class_constants_1.ClassKey.SAN) {
            if (itemStructure.subsystem === undefined) {
                return {
                    valid: false,
                    message: 'subsystem is required',
                };
            }
        }
        else {
            if (itemStructure.hw_type === undefined) {
                return {
                    valid: false,
                    message: 'hw_type is required',
                };
            }
            else if (itemStructure.subsystem === undefined) {
                return {
                    valid: false,
                    message: 'subsystem is required',
                };
            }
        }
        return {
            valid: true,
        };
    }
    static async getItems(custom_group_id, withArch = false) {
        const customGroup = await this.getCustomGroup(custom_group_id);
        if (!customGroup) {
            return [];
        }
        let cgItems = [];
        for (const group of customGroup.groups) {
            const items = await ItemGroupWrapper_1.ItemGroupWrapper.getItemsByData(ItemGroupWrapper_1.ItemGroupWrapper.transformItemGroupEntity(group), customGroup.subsystem, customGroup.hw_type, customGroup.class, withArch
                ? {
                    hw_type: true,
                    subsystem: true
                }
                : undefined);
            cgItems = cgItems.concat(items.map(item => {
                return {
                    item_id: item.item_id,
                    label: item.label,
                    subsystem: withArch ? item.subsystem : undefined,
                    hw_type: withArch ? item.hw_type : undefined
                };
            }));
        }
        const ids = [];
        cgItems = cgItems.filter((cgItem) => {
            const isDuplicate = ids.includes(cgItem.item_id);
            if (!isDuplicate) {
                ids.push(cgItem.item_id);
                return true;
            }
            return false;
        });
        if (customGroup.user_id) {
            const user = await User_1.User.getUserById(customGroup.user_id, true);
            const itemIds = await ACLCheck_1.ACLCheck.filterItemsByRulesTopDown(user.groups.map((group) => group.group_id), cgItems.map((item) => item.item_id));
            cgItems = cgItems.filter((cgItem) => itemIds.includes(cgItem.item_id));
        }
        return cgItems;
    }
    static async getMetrics(custom_group_id, withRawMetric = false) {
        const customGroup = await this.getCustomGroup(custom_group_id);
        if (!customGroup) {
            return [];
        }
        let metrics = [];
        if (!customGroup.hw_type && customGroup.class) {
            if (customGroup.class === 'virtualization') {
                return [
                    {
                        ...Definitions_1.metricTypes.cpu_core,
                        metric: 'cpu_utilization_cores',
                        label: 'CPU',
                    },
                    {
                        ...Definitions_1.metricTypes.cpu_percent,
                        metric: 'cpu_utilization_percent',
                        label: 'CPU %',
                    },
                ];
            }
            const items = await this.getItems(custom_group_id);
            const hwTypes = (await (0, dataSource_1.getDataSource)()
                .getRepository(objectItem_entity_1.ObjectItem)
                .createQueryBuilder('object_items')
                .where({
                item_id: (0, typeorm_1.In)(items.map((item) => item.item_id)),
            })
                .distinctOn(['object_items.hw_type'])
                .getMany()).map((item) => item.hw_type);
            for (const hwType of hwTypes) {
                const hMetrics = await MetricsWrapper_1.MetricsWrapper.getMetrics(hwType, customGroup.subsystem, 'customGroup');
                if (['storage', 'san', 'lan'].includes(customGroup.class)) {
                    if (metrics.length === 0) {
                        metrics = hMetrics;
                        continue;
                    }
                    else {
                        for (const hMetric of hMetrics) {
                            if (metrics.find((metric) => metric.metric === hMetric.metric) === undefined) {
                                metrics.push(hMetric);
                            }
                        }
                    }
                }
                else {
                    if (metrics.length === 0) {
                        metrics = hMetrics;
                        continue;
                    }
                    else {
                        metrics = hMetrics.filter((hMetric) => metrics.find((metric) => metric.metric === hMetric.metric) !== undefined);
                    }
                }
            }
        }
        else if (customGroup.hw_type) {
            metrics = await MetricsWrapper_1.MetricsWrapper.getMetrics(customGroup.hw_type, customGroup.subsystem, 'customGroup');
        }
        return withRawMetric ? metrics : MetricsWrapper_1.MetricsWrapper.handleLongLabel(metrics);
    }
    static async reduceItems(itemIds, metric, max, lastValues) {
        if (max === undefined) {
            max = CustomGroupWrapper.DEFAULT_FREE_MAX_ITEMS;
        }
        const metricsWrapper = new MetricsWrapper_1.MetricsWrapper([metric], itemIds);
        lastValues = lastValues ? lastValues : await metricsWrapper.getLastDailyValuesFromTable();
        let mappedItemData = [];
        if (lastValues === undefined || Object.keys(lastValues).length <= max) {
            return itemIds.slice(0, max);
        }
        for (const itemId in lastValues) {
            if (!itemIds.includes(itemId)) {
                continue;
            }
            if (lastValues[itemId][metric] !== undefined) {
                mappedItemData.push({
                    itemId: itemId,
                    value: lastValues[itemId][metric],
                });
            }
        }
        mappedItemData.sort((a, b) => b.value - a.value);
        return mappedItemData.map((iData) => iData.itemId).slice(0, max);
    }
    static async getInfoAndParent(customGroup) {
        let isClassDefined = false;
        let subsystemLabel = customGroup.subsystem;
        let parentType = undefined;
        if (customGroup.class == class_constants_1.ClassKey.STORAGE || customGroup.class == class_constants_1.ClassKey.SAN || customGroup.class == class_constants_1.ClassKey.LAN) {
            isClassDefined = true;
            if (customGroup.subsystem !== 'device') {
                const subsystemDefinition = await (0, dataSource_1.getDataSource)()
                    .getRepository(subsystem_entity_1.Subsystem)
                    .findOne({
                    where: {
                        hw_type_entity: {
                            class: customGroup.class,
                        },
                        subsystem: customGroup.subsystem,
                    },
                    relations: {
                        hw_type_entity: true,
                    },
                });
                if (subsystemDefinition) {
                    subsystemLabel = subsystemDefinition.label;
                }
                parentType = {
                    label: 'Device',
                    subsystem: 'device',
                };
            }
        }
        else if (!customGroup.hw_type && customGroup.class == class_constants_1.ClassKey.VIRTUALIZATION && customGroup.subsystem === 'vm') {
            isClassDefined = true;
            subsystemLabel = 'VM';
            parentType = {
                label: 'Server',
                subsystem: 'server',
            };
        }
        else {
            const subsystem = await (0, dataSource_1.getDataSource)()
                .getRepository(subsystem_entity_1.Subsystem)
                .findOne({
                where: {
                    hw_type: customGroup.hw_type,
                    subsystem: customGroup.subsystem,
                },
                relations: {
                    hw_type_entity: true,
                },
            });
            if (subsystem) {
                parentType = MetricsWrapper_1.MetricsWrapper.getParent(subsystem.hw_type_entity.class, customGroup.hw_type, customGroup.subsystem);
                subsystemLabel = subsystem.label;
            }
        }
        return {
            isClassDefined: isClassDefined,
            subsystemLabel: subsystemLabel,
            parentType: parentType,
        };
    }
    static async hasItemStructureAnyItems(itemStructure) {
        const hwTypeOrClass = itemStructure.hw_type ? itemStructure.hw_type : itemStructure.class;
        for (const group of itemStructure.groups) {
            if (group.item_ids && group.item_ids.length >= 1) {
                return true;
            }
            const items = await ItemGroupWrapper_1.ItemGroupWrapper.getItemsByData(group, itemStructure.subsystem, itemStructure.hw_type, itemStructure.class);
            if (items.length >= 1) {
                return true;
            }
        }
        return false;
    }
    static async genTab(items, metric, parentType, subsystemLabel, isFree, tabLabel, lastValues, licensedTechnologies) {
        let stacked = true;
        if (metric.unit === subsystemMetricsResponse_constants_1.UNIT.percent ||
            metric.unit === subsystemMetricsResponse_constants_1.UNIT.millisecond ||
            metric.metricType === Definitions_1.MetricTypeKey.BLOCK_SIZE ||
            metric.unit === subsystemMetricsResponse_constants_1.UNIT.sessions ||
            metric.metric.includes('sfp_power')) {
            stacked = false;
        }
        const customMetric = {
            type: stacked ? pageResponse_constants_1.CHART_TYPE.area : pageResponse_constants_1.CHART_TYPE.line,
            stacked: stacked,
            metric: metric.metric,
            order: 0,
            label: metric.label,
            unit: metric.unit,
            shortcut: subsystemMetricsResponse_constants_1.SHORTCUT[metric.unit],
            defaultPrecision: metric.defaultPrecision,
            defaultPrefix: metric.defaultPrefix,
            unitLabel: metric.unitLabel,
        };
        let itemIds = items.map((item) => item.item_id);
        let max;
        let reduced = false;
        if (isFree) {
            max = ['port', 'sanport', 'san port', 'volume'].includes(subsystemLabel.toLowerCase())
                ? 10
                : CustomGroupWrapper.DEFAULT_FREE_MAX_ITEMS;
            if (licensedTechnologies && licensedTechnologies.length >= 1) {
                let finalItemIds = [];
                let mappedItems = {};
                for (const item of items) {
                    if (!mappedItems[item.hw_type]) {
                        mappedItems[item.hw_type] = [];
                    }
                    mappedItems[item.hw_type].push(item.item_id);
                }
                for (const technology of licensedTechnologies) {
                    if (mappedItems[technology]) {
                        finalItemIds.push(...mappedItems[technology]);
                    }
                }
                if (finalItemIds.length >= max) {
                    itemIds = finalItemIds;
                }
                else {
                    itemIds = itemIds.filter((i) => !finalItemIds.includes(i));
                    itemIds = await CustomGroupWrapper.reduceItems(itemIds, metric.metric, max - finalItemIds.length, lastValues);
                    itemIds = itemIds.concat(finalItemIds);
                }
            }
            else {
                reduced = isFree && items.length >= max;
                itemIds = await CustomGroupWrapper.reduceItems(itemIds, metric.metric, max, lastValues);
            }
        }
        const box = {
            type: pageResponse_constants_1.BOX_TYPE.FourGraphBox,
            metrics: [customMetric],
            uuids: itemIds,
            aggregated: true,
            parent: parentType
                ? {
                    label: parentType.label,
                    subsystem: parentType.subsystem,
                }
                : undefined,
            label: subsystemLabel,
        };
        return {
            title: tabLabel ? tabLabel : metric.longLabel ? metric.longLabel : metric.label,
            content: reduced
                ? [
                    box,
                    {
                        type: pageResponse_constants_1.BOX_TYPE.HtmlTextBox,
                        html: `You have reached max number of items in one custom group. Only <b>first ${max ? max : CustomGroupWrapper.DEFAULT_FREE_MAX_ITEMS} items are graphed</b>.<br>It is a limitation of a <b>free Xormon NG</b> edition.<br>Unlimited number of items in a custom group is one of <a href="https://xormon.com/support-ng.php">benefits</a> of the Enterprise edition which is distributed to customers under support.`,
                    },
                ]
                : [box],
        };
    }
    static async genHeatmapTab(customGroup, items, request) {
        if (CustomGroupWrapper.checkCustomGroupHeatmap(customGroup.class, customGroup.subsystem)) {
            if (customGroup.class == class_constants_1.ClassKey.VIRTUALIZATION || customGroup.class == class_constants_1.ClassKey.CLOUD) {
                const heatmap = new Heatmap_1.Heatmap(customGroup.hw_type, customGroup.subsystem, true, items.map((i) => i.item_id));
                return {
                    title: dashboard_constants_1.HEATMAP_TAB_TITLE,
                    content: [],
                    subTabs: [
                        {
                            title: 'Heatmap',
                            content: Object.values(await heatmap.computedHeatmapBox(customGroup.class, request)),
                        },
                        {
                            title: 'Table',
                            content: [heatmap.genTable()],
                        },
                    ],
                };
            }
            else if (customGroup.class == class_constants_1.ClassKey.STORAGE) {
                const heatmap = new Heatmap_1.HeatmapStorage(request, items.map((i) => i.item_id));
                return {
                    title: dashboard_constants_1.HEATMAP_TAB_TITLE,
                    content: [],
                    subTabs: [
                        {
                            title: 'Heatmap',
                            content: Object.values(await heatmap.heatmapBox(customGroup.subsystem)),
                        },
                        {
                            title: 'Table',
                            content: [await heatmap.tableBox(customGroup.subsystem)],
                        },
                    ],
                };
            }
            else if (customGroup.class == class_constants_1.ClassKey.SAN) {
                const heatmap = new Heatmap_1.HeatmapSan(request, items.map((i) => i.item_id));
                return {
                    title: dashboard_constants_1.HEATMAP_TAB_TITLE,
                    content: [],
                    subTabs: [
                        {
                            title: 'Heatmap',
                            content: Object.values(await heatmap.gen()),
                        },
                        {
                            title: 'Table',
                            content: [await heatmap.table()],
                        },
                    ],
                };
            }
            else if (customGroup.class == class_constants_1.ClassKey.LAN) {
                const heatmap = new Heatmap_1.HeatmapLan(request, items.map((i) => i.item_id));
                return {
                    title: dashboard_constants_1.HEATMAP_TAB_TITLE,
                    content: [],
                    subTabs: [
                        {
                            title: 'Heatmap',
                            content: Object.values(await heatmap.gen()),
                        },
                        {
                            title: 'Table',
                            content: [await heatmap.table()],
                        },
                    ],
                };
            }
        }
        else {
            return null;
        }
    }
    static checkCustomGroupHeatmap(_class, subsystem) {
        return ((_class == class_constants_1.ClassKey.VIRTUALIZATION && ['vm', 'lpar', 'host', 'node', 'esxi'].includes(subsystem)) ||
            (_class == class_constants_1.ClassKey.STORAGE && ['volume', 'pool'].includes(subsystem)) ||
            (_class == class_constants_1.ClassKey.SAN && ['sanport'].includes(subsystem)) ||
            (_class == class_constants_1.ClassKey.LAN && ['physical'].includes(subsystem)) ||
            (_class == class_constants_1.ClassKey.CLOUD && ['ec2', 'compute', 'vm'].includes(subsystem)));
    }
}
exports.CustomGroupWrapper = CustomGroupWrapper;
CustomGroupWrapper.DEFAULT_FREE_MAX_ITEMS = 4;
//# sourceMappingURL=CustomGroupWrapper.js.map