"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CustomTop = void 0;
const pageResponse_constants_1 = require("../../../common/constants/pageResponse.constants");
const dataSource_1 = require("../../../config/dataSource");
const objectItem_entity_1 = require("../../../entity/objectItem.entity");
const subsystem_entity_1 = require("../../../entity/subsystem.entity");
const class_constants_1 = require("../../../common/constants/class.constants");
;
const MetricsWrapper_1 = require("../../metrics/MetricsWrapper");
const typeorm_1 = require("typeorm");
const Table_1 = require("./Table");
const URL_1 = require("./URL");
const User_1 = require("../../users/User");
const ACLCheck_1 = require("../../acl/ACLCheck");
class CustomTop {
    constructor(hwType, subsystem, metrics, start, end, type, parentsSubsystem, request) {
        this.hwType = hwType;
        this.subsystem = subsystem;
        this.metrics = metrics;
        this.start = start;
        this.end = end;
        this.type = type;
        this.parentsSubsystem = parentsSubsystem;
        this.request = request;
        this.items = [];
        this.data = {};
        this.metricsMetadata = [];
    }
    async loadItems() {
        this.items = await (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem).find({
            where: {
                hw_type: [class_constants_1.ClassKey.STORAGE, class_constants_1.ClassKey.SAN].includes(this.hwType) ? undefined : this.hwType,
                subsystem: this.subsystem,
                hw_type_entity: [class_constants_1.ClassKey.STORAGE, class_constants_1.ClassKey.SAN].includes(this.hwType)
                    ? {
                        class: this.hwType
                    } : undefined
            }, relations: {
                hw_type_entity: true,
                parents: this.parentsSubsystem ? {
                    _parent: true
                } : undefined,
            }
        });
        if (this.request) {
            const user = await User_1.User.getUserByRequest(this.request, true);
            const ids = await ACLCheck_1.ACLCheck.filterItems(user.groups.map(group => group.group_id), this.items.map(item => item.item_id));
            this.items = this.items.filter(item => ids.includes(item.item_id));
        }
    }
    async loadMetricsData() {
        if (this.items.length === 0) {
            await this.loadItems();
        }
        const itemIds = this.items.map(item => item.item_id);
        if (this.hwType === class_constants_1.ClassKey.STORAGE || this.hwType === class_constants_1.ClassKey.SAN) {
            this._class = this.hwType;
            const hwTypes = [...new Set(this.items.map(item => item.hw_type))];
            this.metricsMetadata = await MetricsWrapper_1.MetricsWrapper.getMetrics(hwTypes, this.subsystem);
            for (const metric of this.metrics) {
                const mData = await new MetricsWrapper_1.MetricsWrapper([metric], itemIds).getAvgAndMaxByTime(this.start, this.end);
                for (const itemId in mData) {
                    if (mData[itemId][metric]) {
                        if (!this.data[itemId]) {
                            this.data[itemId] = mData[itemId];
                        }
                        else if (!this.data[itemId][metric]) {
                            this.data[itemId][metric] = mData[itemId][metric];
                        }
                    }
                }
            }
        }
        else {
            const metricsWrapper = new MetricsWrapper_1.MetricsWrapper(this.metrics, itemIds);
            this.data = await metricsWrapper.getAvgAndMaxByTime(this.start, this.end);
            this.metricsMetadata = await MetricsWrapper_1.MetricsWrapper.getMetrics(this.hwType, this.subsystem);
        }
    }
    async fetchMetricsData() {
        if (Object.keys(this.data).length === 0) {
            await this.loadMetricsData();
        }
    }
    getData() {
        return this.data;
    }
    async genTable() {
        await this.loadMetricsData();
        if (this.items.length === 0) {
            return {
                type: pageResponse_constants_1.BOX_TYPE.TextBox,
                text: 'No items found'
            };
        }
        const itemSubsystem = MetricsWrapper_1.MetricsWrapper.getSubsystemDefinition(this.items[0].hw_type_entity.class, this.items[0].hw_type, this.items[0].subsystem);
        let table = new Table_1.Table()
            .addHeader({ value: itemSubsystem && itemSubsystem.label ? itemSubsystem.label : this.items[0].subsystem });
        if (this.parentsSubsystem) {
            const parentSubsystemsDb = await (0, dataSource_1.getDataSource)().getRepository(subsystem_entity_1.Subsystem).find({
                where: {
                    subsystem: (0, typeorm_1.In)(this.parentsSubsystem),
                    hw_type: !this._class ? this.hwType : undefined,
                    hw_type_entity: this._class ? {
                        class: this._class
                    } : undefined
                }, relations: this._class ? {
                    hw_type_entity: true
                } : undefined
            });
            for (const subsystem of this.parentsSubsystem) {
                const subsys = parentSubsystemsDb.find(sub => sub.subsystem === subsystem);
                table.addHeader({ value: subsys.label });
            }
            let firstMetric = true;
            for (const metric of this.metrics) {
                const definition = this.metricsMetadata.find(row => row.metric === metric);
                table.addHeader({
                    value: definition ? definition.label : metric,
                    number: true,
                    unit: definition ? definition.unit : undefined,
                    shortcut: definition ? definition.shortcut : undefined,
                    decimal: definition ? definition.defaultPrecision : undefined,
                    prefix: definition ? definition.defaultPrefix : undefined,
                    sort: firstMetric ? 'desc' : undefined
                });
                if (firstMetric) {
                    firstMetric = false;
                }
            }
        }
        for (const item of this.items) {
            const row = [{
                    order: 0,
                    value: item.label,
                    url: URL_1.URL.generate(item.hw_type, item.subsystem, item.item_id)
                }];
            let i = 1;
            for (const parentSubsystem of this.parentsSubsystem) {
                const parent = item.parents.find(parent => parent._parent.subsystem === parentSubsystem);
                row.push({
                    order: i,
                    value: parent ? parent._parent.label : '',
                    url: parent ? URL_1.URL.generate(parent._parent.hw_type, parent._parent.subsystem, parent._parent.item_id) : undefined
                });
                i++;
            }
            for (const metric of this.metrics) {
                row.push({
                    order: i,
                    value: this.data[item.item_id] && this.data[item.item_id][metric]
                        ? this.data[item.item_id][metric][this.type]
                        : null,
                });
                i++;
            }
            table.addRow(row);
        }
        table.sortByIndex(this.parentsSubsystem.length + 1, true);
        table.setBoxExport(this.subsystem + "-top");
        return table.generateBox();
    }
    static generateRequestBox(hw_type, subsystem, metrics, parentsSubsystem, pageLabel) {
        return {
            type: pageResponse_constants_1.BOX_TYPE.CustomTopBox,
            hw_type: hw_type,
            subsystem: subsystem,
            metrics: metrics,
            parentsSubsystem: parentsSubsystem
                ? parentsSubsystem
                : undefined,
            pageLabel: pageLabel
                ? pageLabel
                : undefined
        };
    }
}
exports.CustomTop = CustomTop;
//# sourceMappingURL=CustomTop.js.map