"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Prediction = void 0;
const PM2_1 = require("../configuration/PM2");
const pageResponse_constants_1 = require("../../common/constants/pageResponse.constants");
const objectItem_entity_1 = require("../../entity/objectItem.entity");
const MetricsWrapper_1 = require("./MetricsWrapper");
const dataSource_1 = require("../../config/dataSource");
const typeorm_1 = require("typeorm");
const request_builder_1 = require("../../request/request.builder");
const hostcfgData_types_1 = require("../../common/hostcfgs/hostcfgData.types");
class Prediction {
    static async predict(item, command, frequency, action) {
        if (!await this.isAnalyzerActive()) {
            return null;
        }
        const interval = frequency ? frequency : this.getInterval(command);
        if (!action && command == 'predict_capacity') {
            action = 'load_prediction';
        }
        let request;
        if (Array.isArray(item)) {
            request = [];
            for (const requestItem of item) {
                const table = `metrics_${requestItem.hwType}_${requestItem.subsystem}_${interval}`;
                request.push({
                    command: command,
                    frequency: interval,
                    schema: 'metrics',
                    table: table,
                    column: requestItem.metric,
                    id: requestItem.itemId,
                    max: requestItem.max
                });
            }
        }
        else {
            const table = `metrics_${item.hwType}_${item.subsystem}_${interval}`;
            request = {
                command: command,
                frequency: interval,
                schema: 'metrics',
                table: table,
                column: item.metric,
                id: item.itemId,
                max: item.max
            };
        }
        try {
            const data = await this.requestData(request, action);
            return data;
        }
        catch (e) {
            console.error("[Analyzer] [Request] Error: " + e);
            return {};
        }
    }
    static async getCapacityTableData(items, metric, maxMetric) {
        const data = await this.getCapacityTableRequest(items, metric, maxMetric);
        for (const itemId in data) {
            if (+data[itemId]['time_to_less'] === 365) {
                data[itemId]['time_to_less'] = "365+";
            }
            if (+data[itemId]['time_to_full'] === 365) {
                data[itemId]['time_to_full'] = "365+";
            }
        }
        return data;
    }
    static async reloadPoolCapacityTable() {
        const pools = await (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem).find({
            where: {
                subsystem: (0, typeorm_1.In)(['pool', 'site']),
                hw_type_entity: {
                    class: 'storage'
                }
            }, relations: {
                hw_type_entity: true
            }
        });
        await this.reloadCapacityTableItems(pools, 'capacity_used', 'capacity_total');
    }
    static async reloadCapacityTableItems(items, metric, maxMetric) {
        return await this.getCapacityTableRequest(items, metric, maxMetric, true);
    }
    static async getCapacityTableRequest(items, metric, maxMetric, reload = false) {
        const structure = {};
        let request = [];
        const metricsWrapper = await (new MetricsWrapper_1.MetricsWrapper([maxMetric], items.map(item => item.item_id)))
            .getLastValuesFromTable();
        for (const item of items) {
            if (!structure[item.hw_type]) {
                structure[item.hw_type] = {
                    [item.subsystem]: [item]
                };
            }
            else if (!structure[item.hw_type][item.subsystem]) {
                structure[item.hw_type][item.subsystem] = [item];
            }
            else {
                structure[item.hw_type][item.subsystem].push(item);
            }
        }
        for (const hwType in structure) {
            for (const subsystem in structure[hwType]) {
                const table = `metrics_${hwType}_${subsystem}_1d`;
                for (const item of structure[hwType][subsystem]) {
                    const max = metricsWrapper[item.item_id] && metricsWrapper[item.item_id][maxMetric] !== undefined
                        ? metricsWrapper[item.item_id][maxMetric]
                        : undefined;
                    if (max) {
                        request.push({
                            command: 'predict_capacity',
                            frequency: '1d',
                            schema: 'metrics',
                            table: table,
                            column: metric,
                            id: item.item_id,
                            max: max
                        });
                    }
                }
            }
        }
        if (request.length >= 1) {
            try {
                const command = reload ? 'add_requests_and_calculate_all' : 'load_times_to_fill';
                const data = await this.requestData(request, command);
                return data;
            }
            catch (e) {
                console.error("[Analyzer] [Request] Error: " + e);
                return {};
            }
        }
        else {
            return {};
        }
    }
    static async isAnalyzerActive() {
        const service = await PM2_1.PM2.describeService('analyzer');
        if (service && service.active) {
            return true;
        }
        else {
            return false;
        }
    }
    static getInterval(command) {
        return command === 'predict_cpu' || command === 'predict_vm'
            ? '5m'
            : '1d';
    }
    static async requestData(data, action = 'calculate_now') {
        try {
            const response = await this.processRequestData(data, action);
            const parsed = this.parseJSON(response);
            return parsed;
        }
        catch (e) {
            throw new Error(e);
        }
    }
    static async processRequestData(data, action) {
        return (await request_builder_1.RequestBuilder.simpleRequest({
            protocol: hostcfgData_types_1.PROTO.HTTP,
            url: Prediction.ANALYZER_ENDPOINT + "?action=" + action,
            body: JSON.stringify({
                "request": data
            }),
            timeout: 20000,
            method: request_builder_1.RequestMethod.POST
        }, false)).body;
    }
    static parseJSON(json) {
        let parsed;
        try {
            parsed = JSON.parse(json);
        }
        catch (e) {
            throw new Error("Invalid JSON");
        }
        return parsed;
    }
    static async generatePredictionBox(data) {
        let intervalText;
        let minDays;
        switch (data.frequency) {
            case '1d':
                intervalText = 'Yearly predictions';
                minDays = 7;
                break;
            case '6h':
                intervalText = 'Yearly predictions';
                minDays = 7;
                break;
            case '5m':
                intervalText = 'Monthly predictions';
                minDays = 1;
                break;
            default:
                intervalText = "Predictions";
                break;
        }
        if (!await this.isAnalyzerActive()) {
            return {
                type: pageResponse_constants_1.BOX_TYPE.HtmlTextBox,
                html: `${intervalText} cannot be displayed. Analyzer is not active. <a href="https://xormon.com/Enhanced-Predictions-troubleshooting.php">Troubleshooting</a>`
            };
        }
        if (!await this.showPrediction(data.interval, data.item_id) || data.max === undefined || data.max === null) {
            return {
                type: pageResponse_constants_1.BOX_TYPE.HtmlTextBox,
                html: `${intervalText} cannot be displayed. Not enough data to calculate. At least ${minDays} day${minDays !== 1 ? 's' : ''} of data is needed.`
            };
        }
        const metadata = [
            {
                label: 'Description',
                key: 'description',
                number: false
            },
            data.command === 'predict_cpu' || data.command === 'predict_vm'
                ? {
                    label: 'Days to 70%',
                    key: 'days_to_70_percent_fill',
                    number: true
                }
                : {
                    label: 'Days to 90%',
                    key: 'days_to_90_percent_fill',
                    number: true
                },
            {
                label: 'Days to 100%',
                key: 'days_to_100_percent_fill',
                number: true
            }
        ];
        return {
            type: pageResponse_constants_1.BOX_TYPE.PredictionGraphBox,
            graphLabel: data.graphLabel ? data.graphLabel : `Prediction (${data.interval})`,
            item_id: data.item_id,
            other_metrics: data.other_metrics,
            metric: data.metric,
            command: data.command,
            frequency: data.frequency,
            interval: data.interval,
            max: data.max,
            lowerMax: data.command === 'predict_cpu' || data.command === 'predict_vm'
                ? Math.round(((data.max / 100) * 70) * 10) / 10
                : Math.round(((data.max / 100) * 90) * 10) / 10,
            metadata: metadata,
            help: {
                text: 'How does it work?',
                url: 'https://xormon.com/Enhanced-Predictions.php'
            }
        };
    }
    static filterTimeseries(timeseries) {
        const underZeroIndex = timeseries.findIndex(datapoint => datapoint < 0);
        if (underZeroIndex === -1) {
            return timeseries;
        }
        else {
            return timeseries.splice(0, underZeroIndex !== 0 ? underZeroIndex - 1 : underZeroIndex);
        }
    }
    static async showPrediction(interval, itemId) {
        let time;
        switch (interval) {
            case 'month':
                time = 1 * 24 * 60 * 60;
                break;
            case 'year':
                time = 7 * 24 * 60 * 60;
                break;
            default:
                return false;
        }
        const item = await (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem).findOneBy({ item_id: itemId });
        if (Date.now() - new Date(item.createdAt).getTime() > (time * 1000)) {
            return true;
        }
        return false;
    }
    static getCapacityPredictionTypes() {
        return ['trend', 'naive', 'ultra_clean', 'activity_checked', 'once_clean_trend', 'clean_trend', 'mean_splits', 'short_trend'];
    }
    static getCpuPredictionTypes() {
        return ['trend', 'short_trend', 'mean_splits'];
    }
    static isDefaultDisplayed(type, extended = false) {
        return extended
            ? ['trend', 'short_trend', 'medium_trend', 'long_trend', 'trend_peaks', 'long_ext_daily_averages'].includes(type)
            : ['trend', 'short_trend', 'medium_trend', 'long_trend', 'trend_peaks'].includes(type);
    }
}
exports.Prediction = Prediction;
Prediction.ANALYZER_ENDPOINT = "http://localhost:3020/analyzer";
Prediction.TO_SKIP = ['original'];
Prediction.analyzerTextBox = {
    'not-active': {
        type: pageResponse_constants_1.BOX_TYPE.HtmlTextBox,
        html: `Predictions cannot be displayed. Analyzer is not active. <br><a href="https://xormon.com/Enhanced-Predictions-troubleshooting.php">Troubleshooting</a>`
    },
    'no-data': {
        type: pageResponse_constants_1.BOX_TYPE.TextBox,
        text: "Predictions cannot be displayed. No data available."
    }
};
//# sourceMappingURL=Prediction.js.map