"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Vega = exports.TimeDataPoint = void 0;
const Color_1 = require("../metrics/Color");
const MetricTools_1 = require("./MetricTools");
class TimeDataPoint {
    constructor(time, value, item) {
        this.time = time;
        this.value = value;
        this.item = item;
    }
}
exports.TimeDataPoint = TimeDataPoint;
class Vega {
    static makeYAxisTitle(metricSpecification) {
        return MetricTools_1.MetricTools.makeLabel(metricSpecification, false);
    }
    static findTopTick(value, levelBelow) {
        let fitsTimes = Math.ceil(value / levelBelow);
        return fitsTimes * levelBelow;
    }
    static findBottomTick(value, levelBelow) {
        let fitsTimes = Math.floor(value / levelBelow);
        return fitsTimes * levelBelow;
    }
    static findLevelBelow(value) {
        let absolute = Math.abs(value);
        let levelAbove = 1024;
        while (absolute > levelAbove) {
            levelAbove *= 1024;
        }
        let levelBelow = levelAbove / 1024;
        return levelBelow;
    }
    static makeStepBase(range) {
        if (range == 1) {
            return 0.1;
        }
        if (range == 2) {
            return 0.2;
        }
        if (range < 6) {
            return 0.5;
        }
        if (range < 11) {
            return 1;
        }
        if (range < 21) {
            return 2;
        }
        if (range < 51) {
            return 5;
        }
        if (range < 101) {
            return 10;
        }
        if (range < 201) {
            return 20;
        }
        if (range < 501) {
            return 50;
        }
        return 100;
    }
    static makeBinaryTickValues(max, min = 0, includeZero = true) {
        let level = 1;
        if (includeZero) {
            if (max < 0) {
                max = 0;
            }
            if (min > 0) {
                min = 0;
            }
        }
        if (max === 0 && min === 0) {
            return {
                ticks: [0, 1],
                level: 1,
                invertedPrecision: 1,
            };
        }
        if (Math.sign(max) == 1 && Math.sign(min) == -1) {
            level = Math.min(Vega.findLevelBelow(max), Vega.findLevelBelow(min));
        }
        else if (Math.sign(max) == 1) {
            level = Vega.findLevelBelow(max);
        }
        else {
            level = Vega.findLevelBelow(min);
        }
        const topTick = Vega.findTopTick(max, level);
        const bottomTick = Vega.findBottomTick(min, level);
        if (topTick == bottomTick) {
            return {
                ticks: [topTick],
                level: level,
                invertedPrecision: 1,
            };
        }
        const range = (topTick - bottomTick) / level;
        const stepBase = Vega.makeStepBase(range);
        const first = (stepBase <= 1) ? bottomTick : Math.floor(bottomTick / stepBase) * stepBase;
        let tick = first;
        const ticks = [tick];
        const step = stepBase * level;
        while (tick <= topTick) {
            tick += step;
            ticks.push(tick);
        }
        return {
            ticks: ticks,
            level: level,
            invertedPrecision: (stepBase < 1) ? 10 : 1,
        };
    }
    static checkMaxMin(max, min) {
        if (!isFinite(max)) {
            console.warn(`Vega:checkMaxMin Warning: max (${max}) is not a finite number`);
            return false;
        }
        if (!isFinite(min)) {
            console.warn(`Vega:checkMaxMin Warning: min (${min}) is not a finite number`);
            return false;
        }
        if (max < min) {
            console.warn(`Vega:checkMaxMin Warning: min (${min}) is greater than max (${max})`);
            return false;
        }
        if (max === 0 && min === 0) {
            console.warn(`Vega:checkMaxMin Warning: max and min are both 0`);
            return false;
        }
        return true;
    }
    static makeBinaryAxis(max, min = 0, includeZero = true) {
        if (Vega.checkMaxMin(max, min) === false) {
            return false;
        }
        const { ticks, level, invertedPrecision } = Vega.makeBinaryTickValues(max, 0, includeZero);
        const prefix = MetricTools_1.MetricTools.makeBinaryPrefix(level);
        const format = (invertedPrecision == 10) ? '.1f' : ',d';
        return {
            values: ticks,
            labelExpr: `format(round(datum.value / ${level / invertedPrecision}) / ${invertedPrecision}, "${format}") + "${prefix}"`,
        };
    }
    static makeVegaSchemaBase(width = 460, height) {
        const schema = {
            "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
            "title": undefined,
            "data": {
                "values": [],
            },
            "layer": [
                {
                    "mark": {
                        "type": "line",
                        "strokeWidth": 1,
                    },
                    "encoding": {
                        "x": {
                            "field": "time",
                            "type": "temporal",
                            "timeUnit": "utcmonthdatehoursminutes",
                            "title": null,
                            "axis": {
                                "labelFontSize": 6,
                                "labelAngle": undefined
                            },
                        },
                        "y": {
                            "scale": {
                                "padding": 0
                            },
                            "field": "value",
                            "type": "quantitative",
                            "axis": {
                                "title": "",
                                "titleFontWeight": "normal",
                                "titleFontSize": 7,
                                "labelFontSize": 6,
                            },
                        },
                        "color": {
                            "field": "item",
                            "title": "Item",
                            "type": "nominal",
                            "scale": {
                                "domain": [],
                                "range": [],
                            },
                            "legend": {
                                "direction": "vertical",
                                "orient": "bottom",
                                "labelExpr": "",
                                "labelLimit": 0,
                            }
                        }
                    },
                },
            ],
            "autosize": {
                "type": "fit-x",
                "contains": "padding",
            },
            "width": width,
        };
        if (height) {
            schema['height'] = height;
            schema.autosize.type = "fit";
        }
        return schema;
    }
    static makeYAxisLabelFormat(max, min, usePrefix = false) {
        max = (max > 0) ? max : 0;
        min = (min < 0) ? min : 0;
        if (usePrefix && (Math.max(Math.abs(max), Math.abs(min))) > 1000) {
            return '~s';
        }
        const range = max - min;
        if (range < 7) {
            return '.1f';
        }
        return ',.15~g';
    }
    static makeAlertLimitLine(alertLimit) {
        return {
            "data": {
                "values": [{}]
            },
            "encoding": {
                "y": { "datum": alertLimit }
            },
            "layer": [{
                    "mark": "rule"
                }, {
                    "mark": {
                        "type": "text",
                        "align": "right",
                        "baseline": "bottom",
                        "dx": -2,
                        "dy": -2,
                        "x": "width",
                        "text": ""
                    }
                }]
        };
    }
    static makeImgVegaData(values, metricSpecification, itemLabels, stacked, max, width, height, alertLimit, includeLegend, graphTitle, alertingOptions) {
        const min = 0;
        const schema = Vega.makeVegaSchemaBase(width, height);
        if (!alertingOptions) {
            schema.layer[0].encoding.color.legend.direction = undefined;
            schema.layer[0].encoding.color.legend.orient = undefined;
        }
        schema.title = graphTitle ? graphTitle : undefined;
        schema.layer[0].encoding.color.title = alertingOptions ? alertingOptions.legendTitle : undefined;
        schema.data.values = values;
        schema.layer[0].encoding.y.axis.title = Vega.makeYAxisTitle(metricSpecification);
        schema.layer[0].encoding.color.scale.domain = itemLabels.map(item => item.uuid);
        schema.layer[0].encoding.color.scale.range.push(...Color_1.Color.colors);
        const msUnit = metricSpecification && 'unit' in metricSpecification && metricSpecification.unit ? metricSpecification.unit : undefined;
        switch (MetricTools_1.MetricTools.getUnitPrefixBase(msUnit, 'graphYAxis')) {
            case MetricTools_1.UNIT_PREFIX_BASE.noprefix:
                schema.layer[0].encoding.y.axis['format'] = Vega.makeYAxisLabelFormat(max, min, false);
                break;
            case MetricTools_1.UNIT_PREFIX_BASE.base10:
                schema.layer[0].encoding.y.axis['format'] = Vega.makeYAxisLabelFormat(max, min, true);
                break;
            case MetricTools_1.UNIT_PREFIX_BASE.base2:
                const binaryAxis = Vega.makeBinaryAxis(max, min, true);
                if (binaryAxis) {
                    Object.assign(schema.layer[0].encoding.y.axis, binaryAxis);
                }
                else {
                    schema.layer[0].encoding.y.axis['format'] = Vega.makeYAxisLabelFormat(max, min, false);
                }
        }
        if (stacked) {
            schema.layer[0].mark = 'area';
            schema.layer[0].encoding.y['aggregate'] = 'sum';
        }
        if (!includeLegend) {
            schema.layer[0].encoding.color.legend = null;
        }
        else {
            let string = "";
            for (const itemIlabel of itemLabels) {
                itemIlabel.label = itemIlabel.label.includes('[]') ? itemIlabel.label.replace('[]', '') : itemIlabel.label;
                string += `datum.label == '${itemIlabel.uuid}' ? '${itemIlabel.label}' : `;
            }
            string += `datum.label`;
            schema.layer[0].encoding.color.legend.labelExpr = string;
        }
        if (alertLimit !== undefined) {
            schema.layer.push(Vega.makeAlertLimitLine(alertLimit));
            schema.layer[0].encoding.x.axis.labelAngle = -25;
            schema.layer[0].encoding.color.scale.domain.push(`${alertingOptions ? alertingOptions.limitText : "Limit"}`);
            schema.layer[0].encoding.color.scale.range[schema.layer[0].encoding.color.scale.domain.length - 1] = "#000";
        }
        return schema;
    }
    static createImageBuffer(vegaData, format = 'png') {
        const vega = require('vega');
        const lite = require('vega-lite');
        return new Promise(function (resolve, reject) {
            let vegaspec = lite.compile(vegaData).spec;
            const view = new vega.View(vega.parse(vegaspec), {
                loader: vega.loader(),
                logLevel: vega.Warn,
                renderer: 'none'
            });
            if (format === 'png') {
                view.toCanvas()
                    .then(canvas => {
                    resolve(canvas.toBuffer());
                })
                    .catch(function (err) { reject(err); });
            }
            else {
                view.toSVG()
                    .then(resolve)
                    .catch(reject);
            }
        });
    }
}
exports.Vega = Vega;
//# sourceMappingURL=Vega.js.map