use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Test;
use AmazonWebServices;
use Xormon;
use POSIX;
use Digest::MD5 qw(md5_hex);

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3501;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

# waiting for a new client connection
my $client;
while ($client = $server->accept()) {
    
    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {

        # get information about a newly connected client
        my $client_address = $client->peerhost();
        my $client_port = $client->peerport();
        Xormon::log("Connection from $client_address:$client_port");

        # received data template 
        #
        # { 
        #   "jwt"  : "<backendToken>",
        #   "task" : "<taskName>",
        #   "data" : {
        #     <configuration_from_hostcfg>
        #   }
        # }

        # read up to 2048 characters from the connected client
        my $data = "";
        $client->recv($data, 2048);

        my $decoded = from_json($data);
        my $cfg = $decoded->{data};

        Xormon::log("Received task: ".$decoded->{task});
        my $response = '';
        if ($decoded->{task} eq "fetch") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'aws', $cfg->{hostcfg_id}, $cfg->{hostalias});
            
            # target destinations check
            if (defined $decoded->{'destinations'}) {
                $xormon->setDestinations($decoded->{'destinations'});
            }

            process($xormon, $cfg);

        } elsif ($decoded->{task} eq "test") {

            $response = apitest($cfg);

            $client->send($response);
            $client->shutdown(SHUT_WR);

        } elsif ($decoded->{task} eq "regions") {

            $response = regions($cfg);

            $client->send($response);
            $client->shutdown(SHUT_WR);

        } else {
            $response = "undefined task";

            $client->send($response);
            $client->shutdown(SHUT_WR);
        }

        exit();

    } else {
        $client->autoflush(1);    
    }
}

sub process {
    my $xormon = shift;
    my $cfg = shift;

    $xormon->startJob('fetch', $cfg->{hostalias}, $$);

    my $perfData = api2json($xormon, $cfg);

    my $rc = $xormon->saveJSON2($perfData);

    my $status = "successful";
    if ($rc eq "401") {
        $status = "unauthorized";
    } elsif ($rc ne "201") {
        $status = "error";
    }

    $xormon->endJob('fetch', $cfg->{hostalias}, $status, $$);

    Xormon::log("Task [fetch] for ".$cfg->{hostalias}." is done");
}

sub api2json {
    my $xormon = shift;
	my $data = shift;

    my $aws = AmazonWebServices->new( $data->{interval}, $data->{hostalias}, $data->{hostcfg_id} );
    $aws->set_aws_access_key_id($data->{aws_access_key_id});
    $aws->set_aws_secret_access_key($data->{aws_secret_access_key});

    for ( @{$data->{regions}} ) {
        my $region = $_;
        $aws->add_region($region);
    }

    my %conf = $aws->get_configuration();
    if (%conf) {
        if ($conf{architecture}) {
            #save architecture
            $xormon->saveArchitecture(\@{$conf{architecture}});
        }
        if ($conf{configuration}) {
            #save configuration
            $xormon->saveConfJSON(\%{$conf{configuration}});
        }
    }

    my $service_data = $xormon->getData();
    my $last = (defined $service_data->{timestamp}) ? $service_data->{timestamp} : time()-1200;

    my %perf = $aws->get_metrics($last);

    #save last read
    my %service_data_new = ('timestamp' => time()-300);
    $xormon->saveData(\%service_data_new);

    #return ();
    return \%perf;
}

sub apitest {
    my $data = shift;

    my $test = Test->new();

    my $aws = AmazonWebServices->new( $data->{interval}, $data->{hostalias}, $data->{hostcfg_id} );
    
    my $version =  `aws --version`;

    if($version eq ""){
        $test->addSection("AWS CLI not found.", 0, "AWS CLI test failed");
        return $test->genResult();
    }else{
        $test->addSection("AWS CLI found", 1, "AWS CLI test successful");
    }

    my $all_regions;
    eval {
      $aws->set_aws_access_key_id($data->{aws_access_key_id});
      $aws->set_aws_secret_access_key($data->{aws_secret_access_key});

      $all_regions = $aws->get_all_regions();
    };
    if ($@) {
        $test->addSection("Data test", 0, $@);
        return $test->genResult();
    }

    if ( ref($all_regions) eq 'ARRAY' ) {
        $test->addSection("Data test", 1);
    } else {
        $test->addSection("Data test", 0, "Bad Credentials");
    }

    return $test->genResult();
}

sub regions {
    my $data = shift;

    
    my $aws = AmazonWebServices->new( $data->{interval}, $data->{hostalias}, $data->{hostcfg_id} );
    
    my @empty;
    my $all_regions;
    eval {
      $aws->set_aws_access_key_id($data->{aws_access_key_id});
      $aws->set_aws_secret_access_key($data->{aws_secret_access_key});

      $all_regions = $aws->get_all_regions();
    };
    if ($@) {
        return encode_json(\@empty);
    }

    if ( ref($all_regions) eq 'ARRAY' ) {
        return encode_json($all_regions);
    } else {
        return encode_json(\@empty);
    }
}


$server->close();