use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                
use lib "$FindBin::Bin/"; # use lib in this folder
use Xormon;
use Cloudstack;
use POSIX;

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3104;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

my $client;

 # waiting for a new client connection
while ($client = $server->accept()) {

    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {
   
        # get information about a newly connected client
        my $client_address = $client->peerhost();
        my $client_port = $client->peerport();
        Xormon::log("Connection from $client_address:$client_port");

        # received data template 
        #
        # { 
        #   "jwt"  : "<backendToken>",
        #   "task" : "<taskName>"
        #   "data" : {
        #     <configuration_from_hostcfg>
        #   }
        # }

        # read up to 2048 characters from the connected client
        my $data = "";
        $client->recv($data, 2048);

        my $decoded = from_json($data);

        my ($xormon, $cfg);
        if (defined $decoded->{task} && defined $decoded->{jwt}) {
            $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'cloudstack', $cfg->{hostcfg_id}, $cfg->{hostalias});
            $cfg = $decoded->{data};
        } else {
            $client->send("Missing data");
            $client->shutdown(SHUT_WR);
            exit();
        }
        
        Xormon::log("Received task: ".$decoded->{task});
        my $response = '';
        if ($decoded->{task} eq "fetch") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            $xormon->startJob('fetch', $cfg->{hostalias}, $$);

            ############################### AGENT SECTION ###############################

            # get data from technology
            my $dataToBackend = api2json($xormon, $cfg, 'perf');

            # send data to backend
            my $rc = $xormon->saveJSON($dataToBackend);

            ############################### END AGENT SECTION ###############################

            # status code
            my $status;
            if ($rc eq "201") {
                $status = "successful";
            } elsif ($rc eq "401") {
                $status = "unauthorized";
            } else {
                $status = "error";
            }

            $xormon->endJob($decoded->{task}, $cfg->{hostalias}, $status, $$);

            Xormon::log("Task [$decoded->{task}] for ".$cfg->{hostalias}." is done");

        } elsif ($decoded->{task} eq "test") {

            $response = apitest($cfg);

            $client->send($response);
            $client->shutdown(SHUT_WR);

        } elsif ($decoded->{task} eq "conf") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            $xormon->startJob($decoded->{task}, $cfg->{hostalias}, $$);

            my $rc = api2json($xormon, $cfg, $decoded->{task});

             # status code
            my $status;
            if ($rc eq "201") {
                $status = "successful";
            } elsif ($rc eq "401") {
                $status = "unauthorized";
            } else {
                $status = "error";
            }

            $xormon->endJob($decoded->{task}, $cfg->{hostalias}, $status, $$);

            Xormon::log("Task [conf] for ".$cfg->{hostalias}." is done");

        } else {
            $response = "undefined task";

            $client->send($response);
            $client->shutdown(SHUT_WR);

        }

        exit();

    } else {
        $client->autoflush(1);    
    }
}

sub api2json {
    my $xormon = shift;
    my $data = shift;
    my $type = shift;

    if (!defined $type) {
        $type = 'perf';
    }

    my $cloudstack = Cloudstack->new($data->{hostalias}, $data->{protocol}, $data->{host}.":".$data->{api_port});
    my $session = $cloudstack->auth($data->{username},Xormon::unobscure_password($data->{password}));

    if ($type eq 'perf') {

        #metrics data
        my $time = time();
        my $hostsMetrics     = $cloudstack->getHostsMetrics();
        my $instancesMetrics = $cloudstack->getInstancesMetrics();
        my $volumesMetrics   = $cloudstack->getVolumesMetrics();
        my $primaryStorages  = $cloudstack->getPrimaryStorages();

        my %metricsData;
        foreach my $host (@{$hostsMetrics->{listhostsmetricsresponse}{host}}) {
            if (defined $host->{cpuusedghz}) {
            $metricsData{$host->{id}}{$time}{networkkbswrite} = $host->{networkkbswrite};
            $metricsData{$host->{id}}{$time}{networkkbsread} = $host->{networkkbsread};
            $metricsData{$host->{id}}{$time}{memoryallocated} = $host->{memoryallocated};
            $metricsData{$host->{id}}{$time}{memoryused} = $host->{memoryused};
            $metricsData{$host->{id}}{$time}{memorytotal} = $host->{memorytotal};
            $metricsData{$host->{id}}{$time}{cpuusedghz} = ceil(Cloudstack::removeChars($host->{cpuusedghz})*100);
            $metricsData{$host->{id}}{$time}{cputotalghz} = ceil(Cloudstack::removeChars($host->{cputotalghz})*100);
            $metricsData{$host->{id}}{$time}{cpuused} = Cloudstack::removeChars($host->{cpuused});
            $metricsData{$host->{id}}{$time}{cpunumber} = $host->{cpunumber};
            }
        }

        foreach my $instance (@{$instancesMetrics->{listvirtualmachinesmetricsresponse}{virtualmachine}}) {
            if (defined $instance->{cpuused}) {
                $metricsData{$instance->{id}}{$time}{diskkbswrite} = $instance->{diskkbswrite};
                $metricsData{$instance->{id}}{$time}{diskkbsread} = $instance->{diskkbsread};
                $metricsData{$instance->{id}}{$time}{diskiopstotal} = $instance->{diskiopstotal};
                $metricsData{$instance->{id}}{$time}{diskiowrite} = $instance->{diskiowrite};
                $metricsData{$instance->{id}}{$time}{diskioread} = $instance->{diskioread};
                $metricsData{$instance->{id}}{$time}{memoryintfreekbs} = ($instance->{memory}*1000 gt $instance->{memoryintfreekbs}) ? $instance->{memoryintfreekbs} : $instance->{memory} * 1000 ;
                $metricsData{$instance->{id}}{$time}{memory} = $instance->{memory};
                $metricsData{$instance->{id}}{$time}{cpuused} = (defined Cloudstack::removeChars($instance->{cpuused})) ? ceil(Cloudstack::removeChars($instance->{cpuused})) : ();
                $metricsData{$instance->{id}}{$time}{cpuspeed} = $instance->{cpuspeed};
                $metricsData{$instance->{id}}{$time}{networkread} = ceil(Cloudstack::removeChars($instance->{networkread}));
                $metricsData{$instance->{id}}{$time}{networkwrite} = ceil(Cloudstack::removeChars($instance->{networkwrite}));
                $metricsData{$instance->{id}}{$time}{networkkbsread} = $instance->{networkkbsread};
                $metricsData{$instance->{id}}{$time}{networkkbswrite} = $instance->{networkkbswrite};
            }
        }

        foreach my $volume (@{$volumesMetrics->{listvolumesmetricsresponse}{volume}}) {
            if (defined $volume->{diskiopstotal}) {
                $metricsData{$volume->{id}}{$time}{diskiowrite} = $volume->{diskiowrite};
                $metricsData{$volume->{id}}{$time}{diskiopstotal} = $volume->{diskiopstotal};
                $metricsData{$volume->{id}}{$time}{virtualsize} = $volume->{virtualsize};
                $metricsData{$volume->{id}}{$time}{size} = $volume->{size};
                $metricsData{$volume->{id}}{$time}{physicalsize} = $volume->{physicalsize};
                $metricsData{$volume->{id}}{$time}{diskkbswrite} = $volume->{diskkbswrite};
                $metricsData{$volume->{id}}{$time}{diskioread} = $volume->{diskioread};
                $metricsData{$volume->{id}}{$time}{diskkbsread} = $volume->{diskkbsread};
                $metricsData{$volume->{id}}{$time}{utilization} = ceil(Cloudstack::removeChars($volume->{utilization}));
            }
        }

        foreach my $primaryStorage (@{$primaryStorages->{liststoragepoolsmetricsresponse}{storagepool}}) {
            $metricsData{$primaryStorage->{id}}{$time}{disksizetotalgb} = ceil(Cloudstack::parseChars($primaryStorage->{disksizetotalgb}, 0)*1024);
            $metricsData{$primaryStorage->{id}}{$time}{overprovisioning} = ceil(Cloudstack::parseChars($primaryStorage->{disksizetotalgb}, 1)*1024);
            $metricsData{$primaryStorage->{id}}{$time}{disksizeusedgb} = Cloudstack::removeChars($primaryStorage->{disksizeusedgb})*1024;
            $metricsData{$primaryStorage->{id}}{$time}{disksizeallocatedgb} = Cloudstack::removeChars($primaryStorage->{disksizeallocatedgb})*1024;
            $metricsData{$primaryStorage->{id}}{$time}{disksizeunallocatedgb} = Cloudstack::removeChars($primaryStorage->{disksizeunallocatedgb})*1024;
        }

        return \%metricsData;

    } elsif ($type eq 'conf') {

        my $hosts             = $cloudstack->getHosts();
        my $instances         = $cloudstack->getInstances();
        my $volumes           = $cloudstack->getVolumes();
        my $alerts            = $cloudstack->getAlerts();
        my $events            = $cloudstack->getEvents();
        my $primaryStorages   = $cloudstack->getPrimaryStorages();
        my $secondaryStorages = $cloudstack->getSecondaryStorages();
        my $systemVMs         = $cloudstack->getSystemVMs();

        my %conf;
        my %perf;
        my @architecture;
        $conf{$data->{uuid}}{name} = $data->{hostalias};

        #architecture
        my %cloud_arch = (
            "item_id" => $data->{uuid},
            "label" => $data->{hostalias},
            "hw_type" => "CLOUDSTACK",
            "subsystem" => "CLOUD",
            "hostcfg_id" => $data->{hostcfg_id},
            "parents" => []
        );
        push(@architecture, \%cloud_arch);

        foreach my $host (@{$hosts->{listhostsresponse}{host}}) {
            if (defined $host->{hypervisor}) {
            
                #conf
                $conf{$host->{id}}{name} = $host->{name};
                $conf{$host->{id}}{cpunumber} = $host->{cpunumber};
                $conf{$host->{id}}{state} = $host->{state};
                $conf{$host->{id}}{cpusockets} = $host->{cpusockets};
                $conf{$host->{id}}{clustername} = $host->{clustername};
                $conf{$host->{id}}{ip} = $host->{ipaddress};
                $conf{$host->{id}}{hypervisor} = $host->{hypervisor};

                #architecture
                my %host_arch = (
                    "item_id" => $host->{id},
                    "label" => $host->{name},
                    "hw_type" => "CLOUDSTACK",
                    "subsystem" => "HOST",
                    "hostcfg_id" => $data->{hostcfg_id},
                    "parents" => [$data->{uuid}]
                );
                push(@architecture, \%host_arch);
            }  
        }

        foreach my $volume (@{$volumes->{listvolumesresponse}{volume}}) {
        
            #conf
            $conf{$volume->{id}}{name} = $volume->{name};
            $conf{$volume->{id}}{state} = $volume->{state};
            $conf{$volume->{id}}{size} = $volume->{size};
            $conf{$volume->{id}}{zonename} = $volume->{zonename};

            #architecture
            my %volume_arch = (
                "item_id" => $volume->{id},
                "label" => $volume->{name},
                "hw_type" => "CLOUDSTACK",
                "subsystem" => "VOLUME",
                "hostcfg_id" => $data->{hostcfg_id},
                "parents" => [$data->{uuid}]
            );
            push(@architecture, \%volume_arch);
        }

        foreach my $instance (@{$instances->{listvirtualmachinesresponse}{virtualmachine}}) {
        
            #conf
            $conf{$instance->{id}}{name} = $instance->{name};
            $conf{$instance->{id}}{memory} = $instance->{memory};
            $conf{$instance->{id}}{hypervisor} = $instance->{hypervisor};
            $conf{$instance->{id}}{cpunumber} = $instance->{cpunumber};
            $conf{$instance->{id}}{templatename} = $instance->{templatename};
            $conf{$instance->{id}}{state} = $instance->{state};

            #architecture
            my %instance_arch = (
                "item_id" => $instance->{id},
                "label" => $instance->{name},
                "hw_type" => "CLOUDSTACK",
                "subsystem" => "INSTANCE",
                "hostcfg_id" => $data->{hostcfg_id},
                "parents" => [$data->{uuid}]
            );
            push(@architecture, \%instance_arch);
        }

        foreach my $primaryStorage (@{$primaryStorages->{liststoragepoolsmetricsresponse}{storagepool}}) {
            
            #conf
            $conf{$primaryStorage->{id}}{name} = $primaryStorage->{name};
            $conf{$primaryStorage->{id}}{type} = $primaryStorage->{type};
            $conf{$primaryStorage->{id}}{state} = $primaryStorage->{state};
            $conf{$primaryStorage->{id}}{scope} = $primaryStorage->{scope};
            $conf{$primaryStorage->{id}}{disksizetotalgb} = $primaryStorage->{disksizetotalgb};
            $conf{$primaryStorage->{id}}{disksizeusedgb} = $primaryStorage->{disksizeusedgb};

            #architecture
            my %primaryStorage_arch = (
                "item_id" => $primaryStorage->{id},
                "label" => $primaryStorage->{name},
                "hw_type" => "CLOUDSTACK",
                "subsystem" => "PRIMARY_STORAGE",
                "hostcfg_id" => $data->{hostcfg_id},
                "parents" => [$data->{uuid}]
            );
            push(@architecture, \%primaryStorage_arch);
        }

        foreach my $secondaryStorage (@{$secondaryStorages->{listimagestoresresponse}{imagestore}}) {

            #conf
            $conf{$secondaryStorage->{id}}{name} = $secondaryStorage->{name};
            $conf{$secondaryStorage->{id}}{protocol} = $secondaryStorage->{protocol};
            $conf{$secondaryStorage->{id}}{disksizetotal} = $secondaryStorage->{disksizetotal};
            $conf{$secondaryStorage->{id}}{disksizeused} = $secondaryStorage->{disksizeused};

            #architecture
            my %secondaryStorage_arch = (
                "item_id" => $secondaryStorage->{id},
                "label" => $secondaryStorage->{name},
                "hw_type" => "CLOUDSTACK",
                "subsystem" => "SECONDARY_STORAGE",
                "hostcfg_id" => $data->{hostcfg_id},
                "parents" => [$data->{uuid}]
            );
            push(@architecture, \%secondaryStorage_arch);
        }

        foreach my $systemVM (@{$systemVMs->{listsystemvmsresponse}{systemvm}}) {
            
            #conf
            $conf{$systemVM->{id}}{name} = $systemVM->{name};
            $conf{$systemVM->{id}}{systemvmtype} = $systemVM->{systemvmtype};
            $conf{$systemVM->{id}}{hypervisor} = $systemVM->{hypervisor};
            $conf{$systemVM->{id}}{privateip} = $systemVM->{privateip};
            $conf{$systemVM->{id}}{state} = $systemVM->{state};
            $conf{$systemVM->{id}}{agentstate} = $systemVM->{agentstate};

            #architecture
            my %systemVM_arch = (
                "item_id" => $systemVM->{id},
                "label" => $systemVM->{name},
                "hw_type" => "CLOUDSTACK",
                "subsystem" => "SECONDARY_STORAGE",
                "hostcfg_id" => $data->{hostcfg_id},
                "parents" => [$data->{uuid}]
            );
            push(@architecture, \%systemVM_arch);
        }

        #save architecture
        $xormon->saveArchitecture(\@architecture);

        #save last read
        my %service_data_new = ('configuration' => time());
        $xormon->saveData($data->{hostalias}, \%service_data_new);

        #save configuration
        return $xormon->saveConfJSON(\%conf);

    }

}

sub apitest {
    my $data = shift;

    my $cloudstack = Cloudstack->new($data->{name}, $data->{protocol}, $data->{host}.":".$data->{api_port});
    my $session = $cloudstack->auth($data->{username},Xormon::unobscure_password($data->{password}));

    my $hosts             = $cloudstack->getHosts();
    my $instances         = $cloudstack->getInstances();

    my @rows;
    my $hosts_count    = scalar @{$hosts->{listhostsresponse}{host}};
    my $instance_count = scalar @{$instances->{listvirtualmachinesresponse}{virtualmachine}};

    if ($hosts_count >= 1) {
        push(@rows, "Number of hosts: $hosts_count");
        push(@rows, "Number of instances: $instance_count");
        return Xormon::conn_test(1, \@rows);
    } else {
        push(@rows, "Can not list hosts");
        return Xormon::conn_test(0, \@rows);
    }
    
}

$server->close();