package LanOthers;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Data::Dumper;
use Xormon;
use JSON qw(encode_json decode_json);
use SNMP_lib;
use Math::BigInt;
use XormonLogger;

my $xormon;
my $cfg;
my $wrkdir;
my $logger;

my %data;
my %conf;
my $sess;

my $cfg_time           = 1;
my $act_time           = "";
my $sample_rate        = 300;
my $snmp_request_sleep = 1; # use one second as a default sleep between each snmp request
my $device_data_in;
my $device_data_out = {};

my $switch_name;
my $switch_uuid;

my $mib_compatibility;
my $conf_interval = 6 * 60 * 60; # default 6 hours

# NOTE:
# - bin/lan_discovery.sh  --> looks like not used anymore

sub collect {
  $xormon   = shift;
  $cfg      = shift;
  $wrkdir   = shift;
  $act_time = time();

  #
  # test main variables
  #
  $switch_uuid = ( exists $cfg->{'hostcfg_id'} )  ? $cfg->{'hostcfg_id'}  : Xormon::error("Cannot continue, 'hostcfg_id' is not configured!")  && return "error";
  $switch_name = ( exists $cfg->{'hostalias'} )   ? $cfg->{'hostalias'}   : Xormon::error("Cannot continue, 'hostalias' is not configured!")   && return "error";
  $logger      = XormonLogger->new( 'lanothers', $cfg->{hostcfg_id}, $cfg->{hostalias} );

  #
  # additional variables
  #
  if ( exists $cfg->{'snmp_request_sleep'} && Xormon::isdigit($cfg->{'snmp_request_sleep'}) ) {
    $snmp_request_sleep = $cfg->{'snmp_request_sleep'};
    $logger->info("snmp_request_sleep=$snmp_request_sleep");
  }
  if ( exists $cfg->{'conf_interval'} && Xormon::isdigit($cfg->{'conf_interval'}) && $cfg->{'conf_interval'} > 0 && $cfg->{'conf_interval'} < 25 ) {
    $logger->info("conf_interval=$cfg->{'conf_interval'}");
    $conf_interval = $cfg->{'conf_interval'} * 60 * 60; # from hours to seconds
  }

  #
  # get microservice data for this device
  #
  $device_data_in  = $xormon->getData();
  $device_data_out = {};

  #
  # check if it is time to collect configuration (only once per hour)
  #
  if ( exists $device_data_in->{last_conf_timestamp} && time() - $device_data_in->{last_conf_timestamp} < $conf_interval ) {
    $cfg_time = 0;
    $device_data_out->{last_conf_timestamp} = $device_data_in->{last_conf_timestamp};
  }
  if ( exists $device_data_in->{mib_compatibility} ) {
    $mib_compatibility = $device_data_in->{mib_compatibility};
    $device_data_out->{mib_compatibility} = $device_data_in->{mib_compatibility};
  }

  #
  # initialize SNMP
  #
  #print Dumper $cfg;
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('lan');
  $sess = SNMP_lib::new_session();

  #
  # check MIB compatibility (only once per an hour)
  #
  if ( $cfg_time ) {
    $logger->info("check_mib_compatibility: START");
    $mib_compatibility = SNMP_lib::check_mib_compatibility( "lanothers" );
    $logger->info("check_mib_compatibility: END");
    $device_data_out->{mib_compatibility} = $mib_compatibility;
  }

  #
  # connection status
  #
  my ( $conn_status, $reason ) = SNMP_lib::collect_lan( \%data, $sess, 'configuration', 'sysName', '', 1 ); # 1 -> conn test
  if ( $conn_status ) {
    $xormon->isConnected(1);
  }
  else {
    $xormon->isConnected(0, $reason);
  }

  #
  # perf data
  #
  collect_perf();
  my $perf_out = prepare_perf();
  my $rc       = $xormon->saveJSON2($perf_out);
  my $status   = "error";
  if ( defined $rc ) {
    $logger->info("save perf: $rc");

    if ( $rc eq "201" ) {
      $status = "successful";
    }
    elsif ( $rc eq "401" ) {
      $status = "unauthorized";
      last;
    }
    else {
      $status = "error";
    }
  }
  else {
    $logger->info("save perf: UN");
  }

  #
  # perf data status
  #
  if ( $status eq "successful" ) {
    $xormon->hasData(1);
  }
  else {
    $xormon->hasData(0, "no available stats: $status");
  }

  #
  # health status
  #
  health_status();

  #
  # conf data
  #
  if ( $cfg_time ) {
    collect_conf();
    send_conf();

    $device_data_out->{last_conf_timestamp} = time();
  }

  #
  # save microservice data for this device
  #
  $xormon->saveData(\%{ $device_data_out });

  return $status;
}

sub health_status {
  my @hs             = ();
  my @status_reason  = ();
  my $status_summary = "ok";

  #
  # switch status (modules)
  #
  # some switches have health statuses for their modules (can be either CPUs, FEXes, switches in stack etc.)
  # these are visible in GUI under 'Health status' -> bookmark 'Modules'
  my @failed_statuses = (
    'failed',            'missing',               'mismatchWithParent',     'mismatchConfig', 'diagFailed', 'dormant',
    'outOfServiceAdmin', 'outOfServiceEnvTemp',   'poweredDown',            'poweredUp',      'powerDenied',
    'powerCycled',       'okButPowerOverWarning', 'okButPowerOverCritical', 'syncInProgress',
    'upgrading',         'okButAuthFailed',       'mdr'
  );
  if ( exists $data{switch_status} && scalar keys %{ $data{switch_status} } ) {
    foreach my $phys_index ( keys %{ $data{switch_status} } ) {
      my $modul_name   = $data{physical_entries}{$phys_index}{entPhysicalDescr};
      my $modul_serial = $data{physical_entries}{$phys_index}{entPhysicalSerialNum};
      my $admin_status = $data{switch_status}{$phys_index}{cefcModuleAdminStatus};
      my $oper_status  = $data{switch_status}{$phys_index}{cefcModuleOperStatus};

      if ( defined $modul_name && defined $modul_serial && defined $admin_status && defined $oper_status ) {
        foreach my $status (@failed_statuses) {
          if ( $oper_status eq $status ) {
            $status_summary = "error";
            push(@status_reason, "The operational status of the switch $switch_name is $oper_status.")
          }
        }
      }
    }
  }

  #
  # Ports health status
  #
  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    my $port_id      = $conf{ports}{$index}{port_id};
    my $port_type    = $conf{ports}{$index}{port_type};
    my $admin_status = "";
    my $oper_status  = "";

    unless ( defined $port_id && defined $port_type && $port_type eq 'physical' ) { next; }

    # item_id
    my $id = lc( Xormon::create_hash_key($switch_uuid.$port_type.$port_id) );

    if ( defined $data{ports}{$index}{ifAdminStatus} ) {
      $admin_status = $data{ports}{$index}{ifAdminStatus};
    }
    if ( defined $data{ports}{$index}{ifOperStatus} ) {
      $oper_status = $data{ports}{$index}{ifOperStatus};
    }

    if ( defined $oper_status && $oper_status eq "lowerLayerDown" ) {
      push(@hs, { "item_id" => $id, "status" => "error", "data" => "Port $port_id: The operational status is $oper_status and the admin status is $admin_status." });
      $status_summary = "error";
      push(@status_reason, "Port $port_id: The operational status is $oper_status and the admin status is $admin_status.")
    }
    else {
      push(@hs, { "item_id" => $id, "status" => "ok", "data" => "Port $port_id: The operational status is $oper_status and the admin status is $admin_status." });
    }
  }

  #
  # add switch status summary (including red states of the ports)
  #
  push(@hs, { "item_id" => $switch_uuid, "status" => $status_summary, "data" => join("\n", @status_reason) });

  # send statuses
  my $rc = $xormon->saveStatus(\@hs);
  $logger->info("save health status: $rc");
  #print Dumper @hs;

  return 1;
}

sub snmp_connection_test {
  $cfg = shift;

  #
  # initialize SNMP
  #
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('san');
  #$sess = SNMP_lib::new_session( { 'Timeout' => 3000000 } ); # use 3 secs, default '1000000', micro-seconds before retry
  if ( defined $cfg->{'vfId'} && $cfg->{'vfId'} ne '' ) {
    $sess = SNMP_lib::new_session( { 'Timeout' => 1000000, Context => "VF:$cfg->{'vfId'}" } ); # use only 1 sec, don't know why, but it always takes 6 times longer, so timeout is 6 seconds in this case
  }
  else {
    $sess = SNMP_lib::new_session( { 'Timeout' => 1000000 } ); # use only 1 sec, don't know why, but it always takes 6 times longer, so timeout is 6 seconds in this case
  }

  my ( $status, $reason ) = SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysName', '', 1 ); # 1 -> conn test

  if ( $status ) {
    return 1, $data{configuration}{sysName}, SNMP_lib::build_snmpwalk_cmd($sess);
  }
  else {
    return 0, $reason, SNMP_lib::build_snmpwalk_cmd($sess);
  }
}

sub prepare_perf {
  my %stats_last;
  my %stats_now;
  my %stats_out;

  #
  # load last stats (it's necessary for converting counters to gauge)
  #
  my $last_stats_file  = "$wrkdir/last_stats.json";
  my $convert_counters = 0;
  my $timediff         = "";
  my @add_timestamps   = ();
  if ( -f $last_stats_file ) {
    eval { %stats_last = %{ Xormon::read_json($last_stats_file) }; };
    if ( exists $stats_last{timestamp} ) {
      $timediff = $act_time - $stats_last{timestamp};
      #if ( $timediff < ( 2 * $sample_rate ) && $timediff >= ( $sample_rate - 60 ) ) { # convert counter only when timediff is between 4 and 10 minutes
      if ( $timediff < ( 12 * $sample_rate ) && $timediff >= ( $sample_rate - 60 ) ) { # convert counter only when timediff is between 4 and 10 minutes
        $convert_counters = 1;

        # if fetch task is longer than 2x300, then is necessary to save stats also for the previous 5 minutes samples
        if ( $timediff > ( 2 * $sample_rate ) ) {
          my $max = int( $timediff / $sample_rate );
          for ( my $i = 1; $i < $max; $i++ ) {
            my $prev_timestamp = $act_time - ( $i * $sample_rate);

            if ( $prev_timestamp > $stats_last{timestamp} && ( $prev_timestamp - $stats_last{timestamp}) > ( $sample_rate - 60 ) ) {
              push(@add_timestamps, $prev_timestamp);
            }
          }
          @add_timestamps = sort @add_timestamps;
        }
      }
      else {
        $logger->error( "Cannot convert counters to gauge. \$act_time=$act_time (".localtime($act_time)."), \$stats_last{timestamp}=$stats_last{timestamp} (".localtime($stats_last{timestamp})."), \$timediff=$timediff " . __FILE__ . ":" . __LINE__ );
      }
    }
    else {
      $logger->error( "Something wrong \$stats_last{timestamp} doesn't exist!" . __FILE__ . ":" . __LINE__ );
    }
  }

  #
  # port
  #
  my $ports_count = 0;
  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    my $var          = $data{ports}{$index};
    my $port_id      = "";
    my $port_alias   = "";
    my $port_type    = "";

    if ( defined $var->{ifName} ) {
      $port_id                        = port_name_format( $var->{ifName} );
      $conf{ports}{$index}{port_name} = $var->{ifName};
    }
    else { next; }    # ports without name are skipped

    if ( defined $var->{ifType} ) {
      $port_type = port_type( $var->{ifName}, $var->{ifType} );
      # physical
      # vlan
      # portchannel
      # stackport   --> skipped
      # tunnel      --> skipped

      unless ($port_type) { next; }    # ports without type are skipped
    }
    else { next; }                     # ports without type are skipped

    # item_id
    my $id = lc( Xormon::create_hash_key($switch_uuid.$port_type.$port_id) );

    if ( $port_type eq 'vlan' ) {
      $conf{ports}{$index}{port_subtype} = 'routed';
    }

    $conf{ports}{$index}{port_id}   = $port_id;
    $conf{ports}{$index}{port_type} = $port_type;

    if ( $port_type eq 'stackport' ) {

      # stackports don't provide any performance data so they must be skipped
      # skiiping must be placed here because their port type and id are needed to generate configuration
      next;
    }
    if ( $port_type eq 'tunnel' ) {

      # tunnels don't provide any performance data so they must be skipped
      # skiiping must be placed here because their port type and id are needed to generate configuration
      next;
    }

    #
    # counters
    #
    # set current counters (convert some metrics from hex2dec)
    $stats_now{timestamp} = $act_time;
    if ( exists $var->{ifHCOutOctets}  && Xormon::isdigit( $var->{ifHCOutOctets} ) )  { $stats_now{$id}{data_out}         = $var->{ifHCOutOctets}; }
    if ( exists $var->{ifHCInOctets}   && Xormon::isdigit( $var->{ifHCInOctets} ) )   { $stats_now{$id}{data_in}          = $var->{ifHCInOctets}; }
    if ( exists $var->{ifOutUcastPkts} && Xormon::isdigit( $var->{ifOutUcastPkts} ) ) { $stats_now{$id}{unicast_pkts_out} = $var->{ifOutUcastPkts}; }
    if ( exists $var->{ifInUcastPkts}  && Xormon::isdigit( $var->{ifInUcastPkts} ) )  { $stats_now{$id}{unicast_pkts_in}  = $var->{ifInUcastPkts}; }
    if ( exists $var->{ifOutErrors}    && Xormon::isdigit( $var->{ifOutErrors} ) )    { $stats_now{$id}{error_out}        = $var->{ifOutErrors}; }
    if ( exists $var->{ifInErrors}     && Xormon::isdigit( $var->{ifInErrors} ) )     { $stats_now{$id}{error_in}         = $var->{ifInErrors}; }

    # prepare stats to DB (convert counters to gauge)
    if ( $convert_counters && exists $stats_last{$id} ) {
      if ( exists $stats_now{$id}{data_out}         && exists $stats_last{$id}{data_out} )         { $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out}         = make_gauge($timediff,$stats_now{$id}{data_out},$stats_last{$id}{data_out}); }
      if ( exists $stats_now{$id}{data_in}          && exists $stats_last{$id}{data_in} )          { $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in}          = make_gauge($timediff,$stats_now{$id}{data_in},$stats_last{$id}{data_in}); }
      if ( exists $stats_now{$id}{unicast_pkts_out} && exists $stats_last{$id}{unicast_pkts_out} ) { $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_out} = make_gauge($timediff,$stats_now{$id}{unicast_pkts_out},$stats_last{$id}{unicast_pkts_out}); }
      if ( exists $stats_now{$id}{unicast_pkts_in}  && exists $stats_last{$id}{unicast_pkts_in} )  { $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_in}  = make_gauge($timediff,$stats_now{$id}{unicast_pkts_in},$stats_last{$id}{unicast_pkts_in}); }
      if ( exists $stats_now{$id}{error_out}        && exists $stats_last{$id}{error_out} )        { $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_out}        = make_gauge($timediff,$stats_now{$id}{error_out},$stats_last{$id}{error_out}); }
      if ( exists $stats_now{$id}{error_in}         && exists $stats_last{$id}{error_in} )         { $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_in}         = make_gauge($timediff,$stats_now{$id}{error_in},$stats_last{$id}{error_in}); }

      # port speed + utilization + check if max speed was ixceeded
      if ( exists $var->{ifHighSpeed} && Xormon::isdigit($var->{ifHighSpeed}) ) {
        # An estimate of the interface's current bandwidth in units of 1,000,000 bits per second.
        my $port_speed = ( $var->{ifHighSpeed} * 1000 * 1000 ) / 8; # from Mb/s to B/s

        if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in} && $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in} ne '' ) {

          if ( $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in} > $port_speed ) {
            $logger->error("The port $port_id:$id max speed has been exceeded! data_in max=$port_speed, current=$stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in}");
            delete($stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in});
          }
          elsif ( $port_speed > 0 ) {
            $stats_out{lanothers}{$port_type}{$id}{$act_time}{utilization_in} = Xormon::formatNumberByType("float4", ( $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in} / $port_speed ) * 100 );
          }
        }
        if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out} && $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out} ne '' ) {

          if ( $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out} > $port_speed ) {
            $logger->error("The port $port_id:$id max speed has been exceeded! data_out max=$port_speed, current=$stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out}");
            delete($stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out});
          }
          elsif ( $port_speed > 0 ) {
            $stats_out{lanothers}{$port_type}{$id}{$act_time}{utilization_out} = Xormon::formatNumberByType("float4", ( $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out} / $port_speed ) * 100 );
          }
        }
      }

      # if fetch task is longer than 2x300, then is necessary to save stats also for the previous 5 minutes samples
      if ( scalar(@add_timestamps) > 0 && exists $stats_out{lanothers}{$port_type}{$id}{$act_time} ) {
        foreach my $new_timestamp (@add_timestamps) {
          $stats_out{lanothers}{$port_type}{$id}{$new_timestamp} = $stats_out{lanothers}{$port_type}{$id}{$act_time};
        }
      }
    }

    $ports_count++;

    if ( $port_type eq "physical" ) {
      # aggregate stats, physical ports only
      if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out} && Xormon::isdigit($stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out}) ) {
        if ( exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{data_out} ) {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{data_out} += $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out};
        }
        else {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{data_out} = $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_out};
        }
      }
      if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in} && Xormon::isdigit($stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in}) ) {
        if ( exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{data_in} ) {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{data_in} += $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in};
        }
        else {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{data_in} = $stats_out{lanothers}{$port_type}{$id}{$act_time}{data_in};
        }
      }
      if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_out} && Xormon::isdigit($stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_out}) ) {
        if ( exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{unicast_pkts_out} ) {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{unicast_pkts_out} += $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_out};
        }
        else {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{unicast_pkts_out} = $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_out};
        }
      }
      if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_in} && Xormon::isdigit($stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_in}) ) {
        if ( exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{unicast_pkts_in} ) {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{unicast_pkts_in} += $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_in};
        }
        else {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{unicast_pkts_in} = $stats_out{lanothers}{$port_type}{$id}{$act_time}{unicast_pkts_in};
        }
      }
      if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_out} && Xormon::isdigit($stats_out{lanothers}{$port_type}{$id}{$act_time}{error_out}) ) {
        if ( exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{error_out} ) {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{error_out} += $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_out};
        }
        else {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{error_out} = $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_out};
        }
      }
      if ( exists $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_in} && Xormon::isdigit($stats_out{lanothers}{$port_type}{$id}{$act_time}{error_in}) ) {
        if ( exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{error_in} ) {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{error_in} += $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_in};
        }
        else {
          $stats_out{lanothers}{device}{$switch_uuid}{$act_time}{error_in} = $stats_out{lanothers}{$port_type}{$id}{$act_time}{error_in};
        }
      }
    }
  }

  if ( $ports_count == 0 ) {
    $logger->error( "No port found" . __FILE__ . ":" . __LINE__ );
  }

  if ( scalar(@add_timestamps) > 0 && exists $stats_out{lanothers}{device}{$switch_uuid}{$act_time} ) {
    foreach my $new_timestamp (@add_timestamps) {
      $logger->info("prepare perf for additional timestamp: $new_timestamp ".localtime($new_timestamp));
      $stats_out{lanothers}{device}{$switch_uuid}{$new_timestamp} = $stats_out{lanothers}{device}{$switch_uuid}{$act_time};
    }
    $logger->info("prepare perf for current timestamp: $act_time ".localtime($act_time));
  }

  # save current counters for the further usage
  eval { Xormon::save_as_json($last_stats_file,\%stats_now); };

  return \%stats_out;
}

sub send_conf {
  my @architecture;
  my %conf_act;

  # basic configuration per each port
  my $up_time = $data{configuration}{sysUpTime};

  #
  # mac addresses, to assign them to physical port they need to be mapped via bridge port
  #
  if ( keys %{ $data{mac_entries} } ) { macs(); }

  #
  # vlans
  #
  if ( keys %{ $data{vlans} } || keys %{ $data{vlan_member_ports} } ) { vlans(); }

  #
  # portchannel
  #
  if ( keys %{ $data{port_channels} } ) { port_channels(); }

  #
  # device (switch)
  #
  my %device_arch = (
    "item_id"    => $switch_uuid,
    "hostcfg_id" => $switch_uuid,
    "label"      => $switch_name,
    "hw_type"    => "lanothers",
    "subsystem"  => "device",
    "parents"    => []
  );
  push(@architecture, \%device_arch);
  # conf
  $conf_act{$switch_uuid}{'ip'}      = $cfg->{'snmpHost'};
  $conf_act{$switch_uuid}{'up_time'} = ( split ':', $up_time )[0];
  if ( defined $data{configuration}{sysLocation} ) { $conf_act{$switch_uuid}{'sysLocation'} = $data{configuration}{sysLocation}; }
  if ( defined $data{configuration}{sysName} )     { $conf_act{$switch_uuid}{'sysName'}     = $data{configuration}{sysName}; }
  my $switch_total_ports     = 0;
  my $switch_unused_ports    = 0;
  my $switch_total_isl_ports = 0;
  ## Port count, unused port count
  while ( my ( $index, $port_id ) = each %{ $conf{ports} } ) {
    my %port = %{ $conf{ports}{$index} };
    if ( defined $conf{ports}{$index}{port_type} && $conf{ports}{$index}{port_type} eq 'physical' ) {
      $switch_total_ports++;
    }
  }
  ## ISL count
  $switch_total_isl_ports = scalar keys %{ $conf{topology} };
  $conf_act{$switch_uuid}{'total_ports'}  = $switch_total_ports;
  $conf_act{$switch_uuid}{'isl_ports'}    = scalar keys %{ $conf{topology} };
  ## Serial number, os version, model
  if ( scalar keys %{ $data{physical_entries} } ) {
    my ( $switch_model, $switch_os_version, $switch_serial_number ) = switch_physical_info();
    $conf_act{$switch_uuid}{'model'}   = $switch_model;
    $conf_act{$switch_uuid}{'version'} = $switch_os_version;
    $conf_act{$switch_uuid}{'serial'}  = $switch_serial_number;
  }
  if ( exists $conf{vlans} && keys %{ $conf{vlans} } ) {
    my @vlan_ids   = ();
    my @vlan_names = ();
    foreach my $vlan_id ( keys %{ $conf{vlans} } ) {
      if ( exists $conf{vlans}{$vlan_id}{vlan_name} ) {
        push(@vlan_ids,   $vlan_id);
        push(@vlan_names, $conf{vlans}{$vlan_id}{vlan_name});
      }
    }
    $conf_act{$switch_uuid}{vlan_id_list}   = join(" ", @vlan_ids);
    $conf_act{$switch_uuid}{vlan_name_list} = join(" ", @vlan_names);
  }

  #
  # ports (physical, vlan, portchannel)
  #
  my $ports_count = 0;
  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    my $var          = $data{ports}{$index};
    my $var_cfg      = $conf{ports}{$index};
    my $port_id      = $conf{ports}{$index}{port_id};
    my $port_type    = $conf{ports}{$index}{port_type};
    my $port_alias   = "";
    #print Dumper $var;

    #unless ( defined $port_id && defined $port_type && $port_type eq 'physical' ) { next; }
    unless ( defined $port_id && defined $port_type ) { next; }

    # item_id
    my $id = lc( Xormon::create_hash_key($switch_uuid.$port_type.$port_id) );

    if ( $port_type eq 'stackport' ) {

      # stackports don't provide any performance data so they must be skipped
      # skiiping must be placed here because their port type and id are needed to generate configuration
      next;
    }
    if ( $port_type eq 'tunnel' ) {

      # tunnels don't provide any performance data so they must be skipped
      # skiiping must be placed here because their port type and id are needed to generate configuration
      next;
    }

    # Architecture
    my %arch = (
      "item_id"    => $id,
      "label"      => $port_id,
      "hw_type"    => "lanothers",
      "subsystem"  => "$port_type",
      "hostcfg_id" => $switch_uuid,
      "parents"    => [$switch_uuid]
    );
    push(@architecture, \%arch);
    # conf
    if ( exists $var->{'ifAlias'} ) {
      $conf_act{$id}{'alias'} = $var->{'ifAlias'};
      $arch{label} = "$port_id [$var->{'ifAlias'}]";
    }
    if ( exists $var->{'ifHighSpeed'} )              { $conf_act{$id}{'speed'}                = ( $var->{'ifHighSpeed'} / 1000 )." Gbps"; } # from Mb/s to Gb/s
    if ( exists $var->{'ifAdminStatus'} )            { $conf_act{$id}{'admin_status'}         = $var->{'ifAdminStatus'}; }
    if ( exists $var->{'ifOperStatus'} )             { $conf_act{$id}{'operational_status'}   = $var->{'ifOperStatus'}; }
    if ( exists $var->{'state'} )                    { $conf_act{$id}{'working_state'}        = $var->{'state'}; }
    if ( exists $var_cfg->{'port_subtype'} )         { $conf_act{$id}{'subtype'}              = $var_cfg->{'port_subtype'}; }
    if ( exists $var_cfg->{'port_channel_members'} ) { $conf_act{$id}{'port_channel_members'} = $var_cfg->{'port_channel_members'}; }
    if ( exists $var_cfg->{'port_channel_indexes'} ) { $conf_act{$id}{'port_channel_indexes'} = $var_cfg->{'port_channel_indexes'}; }
    if ( exists $var->{'ifOperStatus'} )  {
      $conf_act{$id}{'operational_status'} = $var->{'ifOperStatus'};
      if ( $var->{'ifOperStatus'} eq "up" ) {
        $conf_act{$id}{'working_state'} = "used";
      }
      else {
        $conf_act{$id}{'working_state'} = "unused";
      }
    }
    elsif ( defined $up_time && exists $var->{ifLastChange} ) {
      $conf_act{$id}{'working_state'} = happened_before( $up_time, $var->{ifLastChange} ) ? "used" : "unused";
    }
    $switch_unused_ports++ if ( $port_type eq 'physical' && exists $conf_act{$id}{'working_state'} && $conf_act{$id}{'working_state'} eq 'unused' );
    if ( exists $conf{topology}{$index} ) {
      my $topology = $conf{topology}{$index};

      if ( exists $topology->{remote_port} && exists $topology->{remote_system} && exists $topology->{address_type} ) {
        $conf_act{$id}{remote_port}   = $topology->{remote_port};
        $conf_act{$id}{remote_system} = $topology->{remote_system};
        $conf_act{$id}{address_type}  = $topology->{address_type};
        $conf_act{$id}{isl}           = "$topology->{remote_system}, $topology->{remote_port}";
      }
    }
    if ( exists $var_cfg->{vlan_list} && scalar(@{ $var_cfg->{vlan_list} }) ) {
      $conf_act{$id}{vlan_id_list} = join(" ", @{$var_cfg->{vlan_list}});

      my @vlans = ();
      foreach my $vlan_id (@{ $var_cfg->{vlan_list} }) {
        if ( exists $conf{vlans}{$vlan_id}{vlan_name} ) {
          push(@vlans, $conf{vlans}{$vlan_id}{vlan_name});
        }
      }
      if ( scalar(@vlans) ) { $conf_act{$id}{vlan_name_list} = join(" ", @vlans); }
    }
    if ( exists $var_cfg->{macs} && scalar(@{ $var_cfg->{macs} }) ) { $conf_act{$id}{macs} = join(" ", @{$var_cfg->{macs}}); }
  }
  $conf_act{$switch_uuid}{'unused_ports'} = $switch_unused_ports;

  #
  # VLAN cfg
  #
  #if ( exists $conf{vlans}{$vlan_id}{vlan_name} ) {
  if ( exists $conf{vlans} && keys %{ $conf{vlans} } ) {
    foreach my $index ( sort keys %{ $conf{vlans} } ) {
      my $var = $conf{vlans}{$index};
      if ( exists $var->{vlan_name} ) {
        my $id = lc( Xormon::create_hash_key($switch_uuid."vlancfg".$index) );

        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{vlan_name},
          "hw_type"    => "lanothers",
          "subsystem"  => "vlancfg",
          "hostcfg_id" => $switch_uuid,
          "parents"    => [$switch_uuid]
        );
        push(@architecture, \%arch);
        # conf
        $conf_act{$id}{index} = $index;
        if ( exists $var->{unrouted_interface} ) { $conf_act{$id}{unrouted_interface} = $var->{unrouted_interface}; }
        if ( exists $var->{unrouted_ifalias} )   { $conf_act{$id}{unrouted_ifalias}   = $var->{unrouted_ifalias}; }
        if ( exists $var->{unrouted_cfgalias} )  { $conf_act{$id}{unrouted_cfgalias}  = $var->{unrouted_cfgalias}; }
        if ( exists $var->{routed_interface} )   { $conf_act{$id}{routed_interface}   = $var->{routed_interface}; }
        if ( exists $var->{routed_ifalias} )     { $conf_act{$id}{routed_ifalias}     = $var->{routed_ifalias}; }
        if ( exists $var->{routed_cfgalias} )    { $conf_act{$id}{routed_cfgalias}    = $var->{routed_cfgalias}; }
      }
    }
  }


  #print Dumper \%data;
  #print Dumper \%conf;
  #print "#################### Architecture\n";
  #print Dumper @architecture;
  #print "#################### CONF\n";
  #print Dumper \%conf_act;

  #save architecture
  my $rc_arch = $xormon->saveArchitecture(\@architecture);
  $logger->info("save architecture: $rc_arch");

  #save configuration
  my $rc_conf = $xormon->saveConfJSON(\%conf_act);
  $logger->info("save conf: $rc_conf");

  return 1;
}

sub collect_perf {

  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifAlias' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifType' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifOperStatus' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifAdminStatus' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifHighSpeed' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  # exclude characters like ":" from port ids and from port aliases
  if ( exists $data{ports} ) {
    foreach ( keys %{ $data{ports} } ) {
      if ( exists $data{ports}{$_}{ifName} )  { $data{ports}{$_}{ifName}  = remove_unauthorized_characters( $data{ports}{$_}{ifName} ); }
      if ( exists $data{ports}{$_}{ifAlias} ) { $data{ports}{$_}{ifAlias} = remove_unauthorized_characters( $data{ports}{$_}{ifAlias} ); }
    }
  }

  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifHCOutOctets' );  # bytes transmitted - data out
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifHCInOctets' );   # bytes recived - data in
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifOutUcastPkts' ); # Unicast Packets transmitted - Ucast_pkts_out
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifInUcastPkts' );  # Unicast Packets recived - Ucast_pkts_in
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifOutErrors' );    # transmitted Packets with Errors - errors_out
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifInErrors' );     # recived Packets with Errors - errors_i
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  return 1;
}

sub collect_conf {
  ## switch configurations
  SNMP_lib::collect_lan( \%data, $sess, 'configuration',    'sysLocation' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'configuration',    'sysUpTime' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'physical_entries', 'entPhysicalClass' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'physical_entries', 'entPhysicalName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'physical_entries', 'entPhysicalSerialNum' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'physical_entries', 'entPhysicalModelName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'physical_entries', 'entPhysicalSoftwareRev' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  ## port configurations
  SNMP_lib::collect_lan( \%data, $sess, 'ports', 'ifLastChange' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  ## macs
  SNMP_lib::collect_lan( \%data, $sess, 'mac_entries', 'dot1dTpFdbAddress', undef, undef, undef, \&mac_format );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'mac_entries', 'dot1dTpFdbPort' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_lan( \%data, $sess, 'mac_index',   'dot1dBasePortIfIndex' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  # KZ: some cisco 2950SX switches with older OS version don't support OID dot3adAggPortSelectedAggID
  if ( $mib_compatibility->{"IEEE8023-LAG-MIB"} ) {
    SNMP_lib::collect_lan( \%data, $sess, 'port_channels', 'dot3adAggPortSelectedAggID' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  }

  ## vlans
  if ( $mib_compatibility->{"Q-BRIDGE-MIB"} ) {
    SNMP_lib::collect_lan( \%data, $sess, 'vlans', 'dot1qVlanFdbId',              undef, undef, 1 );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_lan( \%data, $sess, 'vlans', 'dot1qVlanCurrentEgressPorts', undef, undef, 1 );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_lan( \%data, $sess, 'vlans', 'dot1qVlanStaticName' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  }

  ## lldp
  if ( $mib_compatibility->{"LLDP-MIB"} && !$mib_compatibility->{"CISCO-CDP-MIB"} ) {
    SNMP_lib::collect_lan( \%data, $sess, 'lldp', 'lldpRemChassisIdSubtype' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_lan( \%data, $sess, 'lldp', 'lldpRemChassisId' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_lan( \%data, $sess, 'lldp', 'lldpRemPortId' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_lan( \%data, $sess, 'lldp', 'lldpRemSysName' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_lan( \%data, $sess, 'lldp', 'lldpRemSysCapEnabled', \&parse_capabilities_lldp );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  }

  return 1;
}

sub hex2dec {
  my $num = shift;

  $num =~ s/\s+//g;
  my $hex = "0x$num";
  my $dec = Math::BigInt->new($hex);
  my $int = "$dec"; # convert object to string (number)

  # example: convert object to string (number)
  #use Math::BigInt;
  #sub bigint { Math::BigInt->new(shift); }
  #$x = Math::BigInt->bstr("1234")       # string "1234"
  #$x = "$x";                            # same as bstr()

  return $int;
}

sub make_gauge {
  my $timediff     = shift;
  my $counter_now  = shift;
  my $counter_last = shift;

  if ( Xormon::isdigit($timediff) && Xormon::isdigit($counter_now) && Xormon::isdigit($counter_last) && $timediff > 0 ) {
    if ( $counter_now > $counter_last ) {
      #return Xormon::formatNumberByType("float4",int( ( $counter_now - $counter_last ) / $timediff )); # --> this is not necessary, value is already formated as an integer
      return int( ( $counter_now - $counter_last ) / $timediff );
    }
    elsif ( $counter_now < $counter_last ) {
      # probably counter overflow
      return "";
    }
    else {
      return 0;
    }
  }

  return "";
}

sub remove_unauthorized_characters {
  my $name = shift;

  $name =~ s/://g;

  #$name =~ s/\///g;
  $name =~ s/\\//g;
  $name =~ s/"//g;
  $name =~ s/'//g;
  $name =~ s/\t//g;   # remove tabulator characters

  if ( exists $cfg->{UNAUTHORIZED_CHARACTERS} && $cfg->{UNAUTHORIZED_CHARACTERS} ne '' ) {
    my $string = $ENV{UNAUTHORIZED_CHARACTERS};
    my @chars  = split( "", $string );

    foreach my $char (@chars) {
      $name =~ s/$char//g;
    }
  }

  return $name;
}

sub parse_capabilities_lldp {

  # parses hex-string into capabilities of device
  my $input        = shift;
  my @capabilities = (
    'other',             'repeater', 'bridge', 'wlanAccessPoint', 'router', 'telephone',
    'docsisCableDevice', 'stationOnly'
  );
  my @output = ();

  $input =~ s/ //g;
  unless ( $input =~ /^[0-9a-fA-F]*$/ ) { return \@output; }    # check if string is in correct format

  my $dec  = hex($input);
  my $mask = ( 16**length $input ) >> 1;
  for ( my $i = 0; $i < scalar @capabilities; $i++ ) {
    if ( $dec & $mask ) {
      push @output, $capabilities[$i];
    }

    $mask >>= 1;
  }

  #print "#### parse_capabilities_lldp:".@output;
  return \@output;
}

sub mac_format {

  # format for readable mac addresses (always with same length)
  # example '0 50 56 8f fd a8' -> '00:50:56:8f:fd:a8'
  my $input = shift;
  chomp $input;
  $input =~ s/ /:/;

  if ( $input =~ /^(([0-9a-fA-f]{1,2})(:|$)){6}/ ) {
    #print "#### mac_format:".join ":", map { length $_ > 1 ? $_ : "0$_" } split /:/, $input;
    return join ":", map { length $_ > 1 ? $_ : "0$_" } split /:/, $input;
  }
  else {    # in case input value is corrupted
    #print "#### mac_format:".$input;
    return $input;
  }
}

sub ip_decimal_format {

  # converts hex-string ip adress to human readable format
  # example of input: '0A 16 63 37'
  # output: '10.22.99.55'
  my $input = shift;
  return $input =~ /([0-9a-fA-F]{2} ?){4}/
    ? join ".", map { hex $_ } split ' ', $input
    : $input;
}

sub parse_capabilities_cdp {

  # parses hex-string into capabilities of device
  my $input        = shift;
  my @capabilities = (
    'Router',   'TB Bridge',  'SR Bridge', 'Switch', 'Host', 'IGMP conditional filtering',
    'Repeater', 'VoIP Phone', 'Remotely-Managed Device'
  );
  my @output = ();

  $input =~ s/ //g;
  unless ( $input =~ /^[0-9a-fA-F]*$/ ) { return \@output; }    # check if string is in correct format
  $input = hex($input);

  for ( my $i = 0; $i < scalar @capabilities; $i++ ) {
    if ( $input & 1 << $i ) {
      push @output, $capabilities[$i];
    }
  }
  return \@output;
}

sub port_name_format {

  # substitues potential troublesome characters in port name (otherwise '/' and whitespaces would end
  #   up in name of RRDs, legends under graphs etc)
  my $name = shift;
  $name =~ s/\//-/g;
  $name =~ s/\./-/g;
  $name =~ s/\s+/_/g;

  return $name;
}

sub port_type {

  # depending on port name and type determines type for PORTS.cfg
  # virtual ports can be distinguished only by their name (this was done for CISCO so other
  #   vendors might not work as expected)
  # physical types fastEther, fastEtherFX, gigabitEthernet are for backward compatibility
  my $port_name = shift;
  my $port_type = shift;

  if    ( $port_type =~ /^(ethernetCsmacd|fastEther|fastEtherFX|gigabitEthernet|ethernet-csmacd)$/ ) { return 'physical'; }
  elsif ( $port_type eq 'tunnel' )                                                                   { return 'tunnel'; }
  elsif ( $port_type eq 'propVirtual' || $port_type eq 'ieee8023adLag' || $port_type eq '161' || $port_type eq '135' ) {
    if    ( $port_name =~ /^(VL|Vl|vl)/ )           { return 'vlan'; }
    elsif ( $port_name =~ /^(PO|Po|port-channel)/ ) { return 'portchannel'; }
    elsif ( $port_name =~ /^(Stack|stack)/ )        { return 'stackport'; }
  }

  # KZ
  # add "ethernet-csmacd" different name for physical ports
  # add "161" -> ieee8023adLag (161),  -- IEEE 802.3ad Link Aggregate
  # add "135" -> l2vlan (135),         -- Layer 2 Virtual LAN using 802.1Q

  # KZ
  # Brocade LAN switch 6740 VDX
  # IF-MIB::ifName.671088641 = STRING: Port-channel 1
  # IF-MIB::ifType.671088641 = INTEGER: other(1)
  if ( ( $port_type eq "other" || $port_type eq "1" ) && $port_name =~ m/^Port-channel/ ) { return 'portchannel'; }

  # KZ
  # Huawei switches S6330 S6320 series
  # 40GE0/0/1 : other(1)                     --> physical
  # Eth-Trunk93 : ieee8023adLag(161) :       --> physical
  # Eth-Trunk12.308 : l2vlan(135) :          --> vlan
  # XGigabitEthernet1/0/13.202 : l2vlan(135) --> vlan
  if ( $port_name =~ m/^40GE/             && ( $port_type eq "other"         || $port_type eq "1" ) )   { return 'physical'; }
  if ( $port_name =~ m/^Eth-Trunk/        && ( $port_type eq "ieee8023adLag" || $port_type eq "161" ) ) { return 'physical'; }
  if ( $port_name =~ m/^Eth-Trunk/        && ( $port_type eq "l2vlan"        || $port_type eq "135" ) ) { return 'vlan'; }
  if ( $port_name =~ m/^XGigabitEthernet/ && ( $port_type eq "l2vlan"        || $port_type eq "135" ) ) { return 'vlan'; }

  # KZ
  # add new ports: ifName -> Bundle-Ether
  #
  # Catalyst 6513 uses Cisco IOS
  # Cisco CRS uses Cisco IOS-XR, and has different naming for port-channel (Bundle-Ether)
  #
  # IF-MIB::ifName.212 = STRING: Bundle-Ether111    IF-MIB::ifType.212 = INTEGER: propVirtual(53)
  # IF-MIB::ifName.233 = STRING: Bundle-Ether9050   IF-MIB::ifType.233 = INTEGER: ieee8023adLag(161)
  # IF-MIB::ifName.323 = STRING: Bundle-Ether63.25  IF-MIB::ifType.323 = INTEGER: l2vlan(135)
  #
  # propVirtual(53),     -- proprietary virtual/internal      --> vlan
  # ieee8023adLag (161), -- IEEE 802.3ad Link Aggregate       --> port-channel
  # l2vlan (135),        -- Layer 2 Virtual LAN using 802.1Q  --> vlan
  if ( $port_name =~ m/^Bundle-Ether/ ) {
    if ( $port_type eq "propVirtual" || $port_type eq "l2vlan" ) {
      return 'vlan';
    }
    elsif ($port_type eq "ieee8023adLag" ) {
      return 'portchannel';
    }
  }

  return '';
}

sub happened_before {

  # compares two time values in from: 255:20:29:20.00 (this format is used in sysUpTime and ifLastChange)
  my @a = split( /:|\./, shift );
  my @b = split( /:|\./, shift );

  for ( my $i = 0; $i < scalar @a; $i++ ) {
    if ( $a[$i] < $b[$i] ) { return 1; }
    if ( $a[$i] > $b[$i] ) { return 0; }
  }

  return 0;
}

sub parse_portlist {

  # Parses list of ifIndexes of ports which belong to VLAN of given portlist
  # Example of input:
  # '00 00 00 00 00 00 00 00 5C 04 00 1F 40 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
  #  00 00 00 00 00 00 00 00 00 00 00 00 00 00 '
  my $portlist = shift;
  my $i        = 1;
  my @parts    = split ' ', $portlist;
  my @list_of_ports;

  for my $part (@parts) {
    my $dec     = hex($part);
    my @indexes = $i .. $i + 8;

    push @list_of_ports, map { $dec & ( 128 >> $_ ) ? $indexes[$_] : () } 0 .. 7;
    $i += 8;
  }
  return \@list_of_ports;
}

sub vlans {
  # add to each physical port list of it's configured VLANs

  ## non-indexed vlans (qBridgeMIB)
  while ( my ( $key, $vlan ) = each %{ $data{vlans} } ) {
    my %vlan            = %{ $data{vlans}{$key} };
    my $vlan_name       = $vlan{dot1qVlanStaticName};
    my $portlist_egress = $vlan{dot1qVlanCurrentEgressPorts};

    unless ( defined $portlist_egress ) {
      next;
    }

    if ( defined $vlan_name ) {
      $conf{vlans}{$key}{vlan_name} = $vlan_name;
    }

    # find set of ports transmitting for this VLAN
    for my $index ( @{ parse_portlist($portlist_egress) } ) {
      my $port_name = $conf{ports}{$index}{port_name};
      my $vlan_id   = $vlan{dot1qVlanFdbId};

      # port name is not needed, but assures that port with index exists
      if ( defined $port_name && defined $vlan_id ) {
        push @{ $conf{ports}{$index}{vlan_list} }, $vlan_id;
      }
    }
  }

  return 1;
}

sub macs {
  foreach my $key ( keys %{ $data{mac_entries} } ) {
    my $mac_address = $data{mac_entries}{$key}{dot1dTpFdbAddress};
    my $bridge_port = $data{mac_entries}{$key}{dot1dTpFdbPort};
    unless ( defined $mac_address && defined $bridge_port ) { next; }

    my $port_index = $data{mac_index}{$bridge_port}{dot1dBasePortIfIndex};
    unless ( defined $port_index ) { next; }

    my $port_name = $data{ports}{$port_index}{ifName};
    unless ( defined $port_name ) { next; }

    push @{ $conf{ports}{$port_index}{macs} }, $mac_address;
  }

  return 1;
}

sub port_channels {
  # maps ports to port channel depending on used technology (PAGP or LACP)
  foreach my $index ( keys %{ $data{port_channels} } ) {
    my $port_channel;
    my $agg_index  = $data{port_channels}{$index}{dot3adAggPortSelectedAggID};    # LACP index
    my $pagp_index = $data{port_channels}{$index}{pagpGroupIfIndex};              # PAGP index

    if ( defined $agg_index && $agg_index ne $index && $agg_index ne '0' && defined $data{ports}{$index} ) {
      $port_channel = $agg_index;
    }
    elsif ( defined $pagp_index && $pagp_index ne $index && $pagp_index ne '0' && defined $data{ports}{$index} ) {
      $port_channel = $pagp_index;
    }
    else {
      next;
    }

    # add member to port channel
    my %member_port = %{ $data{ports}{$index} };
    if ( defined $member_port{ifName} ) {
      push @{ $conf{ports}{$port_channel}{port_channel_members} }, $member_port{ifName};
      push @{ $conf{ports}{$port_channel}{port_channel_indexes} }, $index;
    }
  }

  return 1;
}

sub switch_physical_info {

  # loop through entPhysicalTable and tries to find model, serial and os version
  # problems arise on switches with VSS and other virtualization technologies because
  my $switch_model         = "";
  my $switch_os_version    = "";
  my $switch_serial_number = "";
  my @sorted_keys          = sort keys %{ $data{physical_entries} };
  my @classes              = ( 'chassis', 'module' );

  # first try to find 'chassis' entries and only if there is no suitable entry then look among modules.
  # suitable entry = it has both of model and serial. In some cases OS can be defined only for entry with
  # module class, then use model and serial from chassis and OS from module.
LOOP: foreach my $tested_class (@classes) {
    foreach my $phys_entry_index (@sorted_keys) {
      my %phys_entry = %{ $data{physical_entries}{$phys_entry_index} };
      my $class      = defined $phys_entry{entPhysicalClass} ? $phys_entry{entPhysicalClass} : '';

      unless ( $class eq 'chassis' || $class eq 'module' ) { next; }

      my $serial = defined $phys_entry{entPhysicalSerialNum}   ? $phys_entry{entPhysicalSerialNum}   : '';
      my $model  = defined $phys_entry{entPhysicalModelName}   ? $phys_entry{entPhysicalModelName}   : '';
      my $os     = defined $phys_entry{entPhysicalSoftwareRev} ? $phys_entry{entPhysicalSoftwareRev} : '';

      $serial =~ s/\s+$//;
      $model  =~ s/\s+$//;
      $os     =~ s/\s+$//;

      if ( $class eq $tested_class && $serial && $model && $switch_model eq '' && $switch_serial_number eq '' ) {
        $switch_serial_number = $serial;
        $switch_model         = $model;

        if ($os) {
          $switch_os_version = $os;
          last LOOP;
        }
      }
      elsif ( $class eq 'module' && $switch_model ne '' && $switch_serial_number ne '' && $os ) {

        # case where only os is missing
        $switch_os_version = $os;
        last LOOP;
      }
    }
  }

  return ( $switch_model, $switch_os_version, $switch_serial_number );
}

1;
