use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Test;
use Proxmox;
use Xormon;
use POSIX;
use Digest::MD5 qw(md5_hex);
use XormonLogger;

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3103;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

# waiting for a new client connection
my $client;
while ($client = $server->accept()) {
    
    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {

        # get information about a newly connected client
        my $client_address = $client->peerhost();
        my $client_port = $client->peerport();
        Xormon::log("Connection from $client_address:$client_port");

        # received data template 
        #
        # { 
        #   "jwt"  : "<backendToken>",
        #   "task" : "<taskName>",
        #   "data" : {
        #     <configuration_from_hostcfg>
        #   }
        # }

        # read up to 2048 characters from the connected client
        my $data = "";
        $client->recv($data, 2048);

        my $decoded = from_json($data);
        my $cfg = $decoded->{data};

        Xormon::log("Received task: ".$decoded->{task});
        my $response = '';
        if ($decoded->{task} eq "fetch") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'proxmox', $cfg->{hostcfg_id}, $cfg->{hostalias});

            # target destinations check
            if (defined $decoded->{'destinations'}) {
                $xormon->setDestinations($decoded->{'destinations'});
            }

            process($xormon, $cfg);

        } elsif ($decoded->{task} eq "test") {

            apitest($client, $cfg);

        } else {
            $response = "undefined task";

            $client->send($response);
            $client->shutdown(SHUT_WR);
        }

        exit();

    } else {
        $client->autoflush(1);    
    }
}

sub process {
    my $xormon = shift;
    my $cfg = shift;

    $xormon->startJob('fetch', $cfg->{hostalias}, $$);

    my $perfData = api2json($xormon, $cfg);

    #foreach my $subsystem (keys %{$perfData->{'proxmox'}}) {
    #    foreach my $item (keys %{$perfData->{'proxmox'}{$subsystem}}) {
    #        foreach my $time (keys %{$perfData->{'proxmox'}{$subsystem}{$item}}) {
    #            foreach my $metric (keys %{$perfData->{'proxmox'}{$subsystem}{$item}{$time}}) {
    #                $perfData->{'proxmox'}{$subsystem}{$item}{$time}{$metric} = ceil($perfData->{'proxmox'}{$subsystem}{$item}{$time}{$metric});
    #            }
    #        }
    #    }
    #}

    my $rc = $xormon->saveJSON2($perfData);

    my $status = "successful";
    if ($rc eq "401") {
        $status = "unauthorized";
    } elsif ($rc ne "201") {
        $status = "error";
    }

    $xormon->endJob('fetch', $cfg->{hostalias}, $status, $$);

    Xormon::log("Task [fetch] for ".$cfg->{hostalias}." is done");
}

sub api2json {
    my $xormon = shift;
	my $data = shift;

    my $connected = Test::connTest($data->{host}, $data->{port});
    if ($connected == 1) {
        $xormon->isConnected(1);
    } else {
        $xormon->isConnected(0, "TCP connection to " . $data->{host} . ":" . $data->{port} . " has failed!");
        return ();
    }

    my $proxmox = Proxmox->new($data->{protocol}, $data->{host}, $data->{port}, $data->{backup_host});
    my $auth = $proxmox->authCredentials($data->{username}, $data->{password}, $data->{domain}, $data->{hostalias});

    my %perf;
    my @architecture;
    my @status;
    my @metrics;
    my %conf;
    my %tmp;
    my $service_data = $xormon->getData();
    my $last = (defined $service_data->{timestamp}) ? $service_data->{timestamp} : time()-1200;

    #cluster
    my $cluster = $proxmox->getClusterName();
    my $cluster_id = md5_hex( $cluster );

    #architecture
    my %cluster_arch = (
        "item_id" => $cluster_id,
        "label" => $cluster,
        "hw_type" => "proxmox",
        "subsystem" => "cluster",
        "parents" => [],
        "hostcfg_id" => $data->{hostcfg_id}
    );
    push(@architecture, \%cluster_arch);

    #storages (for wwn)
    my %node_wwn;
    my $stors = $proxmox->getStorages();
    for ( @{ $stors->{data} } ) {
        my $stor = $_;

        if (defined $stor->{vgname} && defined $stor->{base} && defined $stor->{nodes}) {
            my @nodes = split(',', $stor->{nodes});
            my @wwn_split = split('-', $stor->{base});
            my $wwn = $wwn_split[1];
            for (@nodes) {
                my $node = $_;
                if (exists $node_wwn{$node}[0]) {
                    push(@{$node_wwn{$node}}, substr($wwn, 1));
                } else {
                    $node_wwn{$node}[0] = substr($wwn, 1);
                }
            }
        }
    }

    #nodes
    my $nodes = $proxmox->getNodes();

    # node returns the current data
    if (defined $nodes) {
        $xormon->hasData(1);
    } else {
        $xormon->hasData(0, "Unable to get nodes data");
    }

    my $nodesC = scalar @{$nodes->{data}};
    Xormon::log("[".$data->{hostalias}."] Found $nodesC nodes");

    for (@{$nodes->{data}}) {
        my $node = $_;

        my $node_id = md5_hex( $cluster."-".$node->{node} );

        Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] start");

        #conf
        $conf{$node_id}{cluster} = $cluster;
        $conf{$node_id}{name} = $node->{node};
        $conf{$node_id}{status} = $node->{status};
        if (defined $node->{uptime}) {
            $conf{$node_id}{uptime} = time()-$node->{uptime};
        }

        $conf{$node_id}{maxcpu}  = $node->{maxcpu};
        $conf{$node_id}{maxmem}  = $node->{maxmem};
        $conf{$node_id}{maxdisk} = $node->{maxdisk};

        #health status
        #my %node_status = (
        #    "item_id" => $node_id,
        #    "status" => "ok"
        #);
        #push(@status, \%node_status);

        #wwn
        my @wwns;
        my $disks = $proxmox->getNodeDisks($node->{node});
        if (defined $disks) {
            for (@{$disks->{data}}) {
                my $disk = $_;
                if ($disk->{serial} ne "unknown" && length($disk->{serial}) > 8) {
                    push(@wwns, uc($disk->{serial}));
                }
            }
            if (exists $node_wwn{$node->{node}}[0]) {
                for (@{$node_wwn{$node->{node}}}) {
                    my $wwn = $_;
                    push(@wwns, uc($wwn));
                }
            }
        } else {
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] disks returned an empty response, skipping...");
        }

        #architecture
        my %node_arch = (
            "item_id" => $node_id,
            "label" => $node->{node},
            "hw_type" => "proxmox",
            "subsystem" => "node",
            "parents" => [$cluster_id],
            "hostcfg_id" => $data->{hostcfg_id},
            "wwns" => \@wwns
        );
        push(@architecture, \%node_arch);

        #perf data
        my $node_metrics = $proxmox->getNodeMetrics($node->{node});
        if (defined $node_metrics) {
            for (@{$node_metrics->{data}}) {
                my $node_metric = $_;

                if ($node_metric->{time} < $last || !defined $node_metric->{cpu}) { next }

                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{swaptotal} = $node_metric->{swaptotal};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{memused} = $node_metric->{memused};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{iowait} = $node_metric->{iowait}*100; # from 0.00-1 to 0-100 for % 
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{swapused} = $node_metric->{swapused};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{swapfree} = $node_metric->{swaptotal}-$node_metric->{swapused};
                #$perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{loadavg} = $node_metric->{loadavg};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{maxcpu} = $node_metric->{maxcpu};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{roottotal} = $node_metric->{roottotal};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{rootused} = $node_metric->{rootused};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{rootfree} = $node_metric->{roottotal}-$node_metric->{rootused};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{cpu_utilization_percent} = $node_metric->{cpu}*100; # from 0.00-1 to 0-100 for % 
                #$perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{cpu_utilization_percent_max} = $node_metric->{cpu}*100; # from 0.00-1 to 0-100 for % 
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{cpu_utilization_cores} = $node_metric->{cpu}*$node_metric->{maxcpu};
                #$perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{cpu_utilization_cores_max} = $node_metric->{cpu}*$node_metric->{maxcpu};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{memtotal} = $node_metric->{memtotal};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{memfree} = $node_metric->{memtotal} -  $node_metric->{memused};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{netout} = $node_metric->{netout};
                $perf{'proxmox'}{'node'}{$node_id}{$node_metric->{time}}{netin} = $node_metric->{netin};

            }
        } else {
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] node metrics returned an empty response, skipping...");
        }

        #vm under node
        my $vms = $proxmox->getNodeVMs($node->{node});

        if (defined $vms) {
            my $vmsC = scalar @{$vms->{data}};
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] Found $vmsC VMs");

            for (@{$vms->{data}}) {
                my $vm = $_;

                my $vm_id = md5_hex( $cluster."-".$vm->{vmid} );

                #conf
                $conf{$vm_id}{cluster} = $cluster;
                $conf{$vm_id}{name} = $vm->{name};
                $conf{$vm_id}{status} = $vm->{status};
                $conf{$vm_id}{uptime} = time()-$vm->{uptime};

                $conf{$vm_id}{maxmem}  = $vm->{maxmem};
                $conf{$vm_id}{maxdisk} = $vm->{maxdisk};
                $conf{$vm_id}{cpus}    = $vm->{cpus};

                #health status
                #my %vm_status = (
                #    "item_id" => $vm_id,
                #    "status" => "warning"
                #);
                #push(@status, \%vm_status);

                #architecture
                my %vm_arch = (
                    "item_id" => $vm_id,
                    "label" => $vm->{name},
                    "hw_type" => "proxmox",
                    "subsystem" => "vm",
                    "parents" => [$node_id, $cluster_id],
                    "hostcfg_id" => $data->{hostcfg_id}
                );
                push(@architecture, \%vm_arch);

                #perf data
                my $vm_metrics = $proxmox->getVMMetrics($node->{node}, $vm->{vmid});
                for (@{$vm_metrics->{data}}) {
                    my $vm_metric = $_;

                    if ($vm_metric->{time} < $last || !defined $vm_metric->{cpu} ) { next; }

                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{maxmem} = $vm_metric->{maxmem};
                    #$perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{disk} = $vm_metric->{disk};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{diskread} = $vm_metric->{diskread};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{maxdisk} = $vm_metric->{maxdisk};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{diskwrite} = $vm_metric->{diskwrite};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{netin} = $vm_metric->{netin};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{memfree} = $vm_metric->{maxmem}-$vm_metric->{mem};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{mem} = $vm_metric->{mem};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{maxcpu} = $vm_metric->{maxcpu};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{cpu_utilization_percent} = $vm_metric->{cpu}*100; # from 0.00-1 to 0-100 for % 
                    #$perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{cpu_utilization_percent_max} = $vm_metric->{cpu}*100; # from 0.00-1 to 0-100 for % 
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{cpu_utilization_cores} = $vm_metric->{maxcpu} * $vm_metric->{cpu};
                    #$perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{cpu_utilization_cores_max} = $vm_metric->{maxcpu} * $vm_metric->{cpu};
                    $perf{'proxmox'}{'vm'}{$vm_id}{$vm_metric->{time}}{netout} = $vm_metric->{netout};
                }
            }
        } else {
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] vms returned an empty response, skipping...");
        }

        #lxc under node
        my $lxcs = $proxmox->getNodeLXCs($node->{node});

        if (defined $lxcs) {
            my $lxcsC = scalar @{$lxcs->{data}};
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] Found $lxcsC LXCs");

            for (@{$lxcs->{data}}) {
                my $lxc = $_;

                my $lxc_id = md5_hex( $cluster."-".$lxc->{vmid} );

                #conf
                $conf{$lxc_id}{cluster} = $cluster;
                $conf{$lxc_id}{name} = $lxc->{name};
                $conf{$lxc_id}{status} = $lxc->{status};
                $conf{$lxc_id}{uptime} = time()-$lxc->{uptime};

                #health status
                #my %lxc_status = (
                #    "item_id" => $lxc_id,
                #    "status" => "ok"
                #);
                #push(@status, \%lxc_status);

                #architecture
                my %lxc_arch = (
                    "item_id" => $lxc_id,
                    "label" => $lxc->{name},
                    "hw_type" => "proxmox",
                    "subsystem" => "lxc",
                    "parents" => [$node_id, $cluster_id],
                    "hostcfg_id" => $data->{hostcfg_id}
                );
                push(@architecture, \%lxc_arch);

                #perf data
                my $lxc_metrics = $proxmox->getLXCMetrics($node->{node}, $lxc->{vmid});
                for (@{$lxc_metrics->{data}}) {
                    my $lxc_metric = $_;

                    if ($lxc_metric->{time} < $last || !defined $lxc_metric->{cpu}) { next; }

                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{maxmem} = $lxc_metric->{maxmem};
                    #$perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{disk} = $lxc_metric->{disk};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{diskread} = $lxc_metric->{diskread};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{maxdisk} = $lxc_metric->{maxdisk};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{diskwrite} = $lxc_metric->{diskwrite};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{netin} = $lxc_metric->{netin};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{mem} = $lxc_metric->{mem};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{memfree} = $lxc_metric->{maxmem} - $lxc_metric->{mem};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{maxcpu} = $lxc_metric->{maxcpu};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{cpu} = $lxc_metric->{cpu}*100; # from 0.00-1 to 0-100 for % 
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{cpucores} = $lxc_metric->{maxcpu} * $lxc_metric->{cpu};
                    $perf{'proxmox'}{'lxc'}{$lxc_id}{$lxc_metric->{time}}{netout} = $lxc_metric->{netout};
                }
            }
        } else {
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] lxc returned an empty response, skipping...");
        }

        #storage under node
        my $storages = $proxmox->getNodeStorages($node->{node});

        if (defined $storages) {
            my $storagesC = scalar @{$storages->{data}};
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] Found $storagesC storages");

            for (@{$storages->{data}}) {
                my $storage = $_;

                my $storage_id = md5_hex( $cluster."-".$node->{node}."-".$storage->{storage} );

                #conf
                $conf{$storage_id}{cluster} = $cluster;
                $conf{$storage_id}{name} = $storage->{storage}." (".$node->{node}.")";
                $conf{$storage_id}{active} = $storage->{active};
                $conf{$storage_id}{shared} = $storage->{shared};
                $conf{$storage_id}{type} = $storage->{type};
                $conf{$storage_id}{enabled} = $storage->{enabled};

                #architecture
                my %storage_arch = (
                    "item_id" => $storage_id,
                    "label" => $storage->{storage} . " (" . $node->{node} . ")",
                    "hw_type" => "proxmox",
                    "subsystem" => "storage",
                    "parents" => [$node_id, $cluster_id],
                    "hostcfg_id" => $data->{hostcfg_id}
                );
                push(@architecture, \%storage_arch);

                #perf data
                my $storage_metrics = $proxmox->getStorageMetrics($node->{node}, $storage->{storage});
                for (@{$storage_metrics->{data}}) {
                    my $storage_metric = $_;

                    if ($storage_metric->{time} < $last || !defined $storage_metric->{used}) { next }

                    $perf{'proxmox'}{'storage'}{$storage_id}{$storage_metric->{time}}{used} = $storage_metric->{used};
                    $perf{'proxmox'}{'storage'}{$storage_id}{$storage_metric->{time}}{total} = $storage_metric->{total};
                    $perf{'proxmox'}{'storage'}{$storage_id}{$storage_metric->{time}}{free} = $storage_metric->{total} - $storage_metric->{used};
                }
            }
        } else {
            Xormon::log("[".$data->{hostalias}."] [".$node->{node}."] storages returned an empty response, skipping...");
        }

    }
    
    #save last read
    my %service_data_new = ('timestamp' => time()-300);
    $xormon->saveData(\%service_data_new);

    #save architecture
    $xormon->saveArchitecture(\@architecture);

    #save configuration
    $xormon->saveConfJSON(\%conf);

    #save health status
    #$xormon->saveStatus(\@status);

    #save metrics def
    #$xormon->registerMetrics(\@metrics);

    return \%perf;
}

sub apitest {
    my $client = shift;
    my $data = shift;

    my $test = Test->new($client);

    my $connected = Test::connTest($data->{host}, $data->{port});
    if ($connected == 1) {
        $test->addSection("TCP connection test", 1);
    } else {
        $test->addSection("TCP connection test", 0, "TCP connection to " . $data->{host} . ":" . $data->{port} . " has failed! Open it on the firewall.");
        $test->stop();
    }

    my $proxmox = Proxmox->new($data->{protocol}, $data->{host}, $data->{port}, $data->{backup_host});

    my $authTest = $proxmox->authTest($data->{username}, $data->{password}, $data->{domain}, $data->{name});
    if ($authTest == 1) {
        $test->addSection("Auth test", 1);
    } else {
        $test->addSection("Auth test", 0, "Wrong credentials");
        $test->stop();
        return;
    }

    my $nodes;
    eval {
    my $auth = $proxmox->authCredentials($data->{username}, $data->{password}, $data->{domain}, $data->{name});
    $nodes = $proxmox->getNodes();
    };
    if ($@) {
        $test->addSection("Data test", 0, $@);
        $test->stop();
    }

    if (defined $nodes->{data}) {
        my $nodesC = scalar @{$nodes->{data}};
        $test->addSection("Data test", 1, "Number of nodes: $nodesC");
    } else {
        $test->addSection("Data test", 0, "Can not list nodes");
    }

    return $test->stop();
}


$server->close();