package SanBrocade;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Data::Dumper;
use Xormon;
use JSON qw(encode_json decode_json);
use SNMP_lib;
use Math::BigInt;
use XormonLogger;

my $xormon;
my $cfg;
my $wrkdir;
my $logger;

my %data;
my $sess;

my $cfg_time           = 1;
my $act_time           = "";
my $sample_rate        = 300;
my $snmp_request_sleep = 1; # use one second as a default sleep between each snmp request
my $device_data_in;
my $device_data_out = {};

my $switch_name;
my $switch_uuid;

sub collect {
  $xormon   = shift;
  $cfg      = shift;
  $wrkdir   = shift;
  $act_time = time();

  #
  # test main variables
  #
  $switch_uuid = ( exists $cfg->{'hostcfg_id'} )  ? $cfg->{'hostcfg_id'}  : Xormon::error("Cannot continue, 'hostcfg_id' is not configured!")  && return "error";
  $switch_name = ( exists $cfg->{'hostalias'} )   ? $cfg->{'hostalias'}   : Xormon::error("Cannot continue, 'hostalias' is not configured!")   && return "error";
  $logger      = XormonLogger->new( 'sanbrcd', $cfg->{hostcfg_id}, $cfg->{hostalias} );

  #
  # additional variables
  #
  if ( exists $cfg->{'snmp_request_sleep'} && Xormon::isdigit($cfg->{'snmp_request_sleep'}) ) {
    $snmp_request_sleep = $cfg->{'snmp_request_sleep'};
    $logger->info("snmp_request_sleep=$snmp_request_sleep");
  }

  #
  # get microservice data for this device
  #
  $device_data_in  = $xormon->getData();
  $device_data_out = {};

  #
  # check if it is time to collect configuration (only once per hour)
  #
  if ( exists $device_data_in->{last_conf_timestamp} && time() - $device_data_in->{last_conf_timestamp} < 3600 ) {
    $cfg_time = 0;
    $device_data_out->{last_conf_timestamp} = $device_data_in->{last_conf_timestamp};
  }
  if ( exists $device_data_in->{last_sfp_stats} ) {
    $device_data_out->{last_sfp_stats} = $device_data_in->{last_sfp_stats};
  }

  #
  # initialize SNMP
  #
  #print Dumper $cfg;
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('san');
  if ( defined $cfg->{'vfId'} && $cfg->{'vfId'} ne '' ) {
    $sess = SNMP_lib::new_session( { Context => "VF:$cfg->{'vfId'}" } );
  }
  else {
    $sess = SNMP_lib::new_session();
  }

  #
  # connection status
  #
  my ( $conn_status, $reason ) = SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysName', '', 1 ); # 1 -> conn test
  if ( $conn_status ) {
    $xormon->isConnected(1);
  }
  else {
    $xormon->isConnected(0, $reason);
  }

  #
  # perf data
  #
  collect_perf();
  my $perf_out = prepare_perf();
  my $rc       = $xormon->saveJSON2($perf_out);
  my $status   = "error";
  if ( defined $rc ) {
    $logger->info("save perf: $rc");

    if ( $rc eq "201" ) {
      $status = "successful";
    }
    elsif ( $rc eq "401" ) {
      $status = "unauthorized";
      last;
    }
    else {
      $status = "error";
    }
  }
  else {
    $logger->info("save perf: UN");
  }

  #
  # perf data status
  #
  if ( $status eq "successful" ) {
    $xormon->hasData(1);
  }
  else {
    $xormon->hasData(0, "no available stats: $status");
  }

  #
  # health status
  #
  health_status();

  #
  # conf data
  #
  if ( $cfg_time ) {
    collect_conf();
    send_conf();

    $device_data_out->{last_conf_timestamp} = time();
  }

  #
  # events
  #
  if ( $cfg_time ) {
    collect_events(); # collect events only once per hour
  }
  send_events(); # every MS run, because there is possibility to change presented severity

  #
  # save microservice data for this device
  #
  $xormon->saveData(\%{ $device_data_out });


  return $status;
}

sub send_events {
  my $events_file = "$wrkdir/events.json";

  if ( -f $events_file ) {
    my %events = %{ Xormon::read_json($events_file) };

    # save events to DB
    my @eve_data = ();
    foreach my $event_id ( sort { $b <=> $a } keys( %events ) ) {
      my $var = $events{$event_id};
      if ( exists $var->{swEventLevel} && exists $var->{swEventDescr} && exists $var->{swEventTimeInfo} ) {
        # critical (1)
        # error (2)
        # warning (3)
        # informational (4)
        # debug (5)

        if ( exists $cfg->{events} ) {
          if ( $cfg->{events} eq "all" ) {
            push(@eve_data, { "id" => $event_id, "severity" => $var->{swEventLevel}, "time" => $var->{swEventTimeInfo} , "description" => $var->{swEventDescr} } );
          }
          else { # show defined, must be in this format events=critical,error,warning
            foreach my $def_severity ( split(/,/, $cfg->{events}) ) {
              $def_severity =~ s/^\s+|\s+$//g;

              if ( $def_severity eq $var->{swEventLevel} ) {
                push(@eve_data, { "id" => $event_id, "severity" => $var->{swEventLevel}, "time" => $var->{swEventTimeInfo} , "description" => $var->{swEventDescr} } );
                last;
              }
            }
          }
        }
        else { # default, show only critical and error events
          if ( $var->{swEventLevel} eq "critical" || $var->{swEventLevel} eq "error" ) {
            push(@eve_data, { "id" => $event_id, "severity" => $var->{swEventLevel}, "time" => $var->{swEventTimeInfo} , "description" => $var->{swEventDescr} } );
          }
        }
      }
    }
    my $rc_hs_events = $xormon->saveConfJSON( { $cfg->{'hostcfg_id'} => { "events" => encode_json(\@eve_data) } }, 1 );
    $logger->info("save health status events: $rc_hs_events");
  }

  return 1;
}

sub collect_events {
  my $collect_events_full = 0;
  my $last_event_file     = "$wrkdir/last_event.json";
  my $events_file         = "$wrkdir/events.json";

  if ( -f "$last_event_file-tmp" && Xormon::file_time_diff("$last_event_file-tmp") < 3599 ) { return 1; }    # still in progress, collect events after 1 hour

  open( my $file, ">$last_event_file-tmp" ) || $logger->error( "Couldn't open file $last_event_file-tmp $!" . __FILE__ . ":" . __LINE__ ) && exit;
  close($file);

  SNMP_lib::collect_san( \%data, $sess, 'events', 'swEventTimeInfo' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  #
  # test last event (do not collect events if there are no any new events)
  #
  my $last_event_id_act = "";
  foreach my $event_id ( sort { $b <=> $a } keys( %{ $data{events} } ) ) {
    $last_event_id_act = $event_id;
    last;
  }
  if ( -f $last_event_file ) {
    open( my $leve, "<$last_event_file" ) || $logger->error( "Couldn't open file $last_event_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    my ($last_event_id) = <$leve>;
    close($leve);

    if ( defined $last_event_id && defined $last_event_id_act && $last_event_id_act ne $last_event_id ) { $collect_events_full = 1; }
  }
  else {
    $collect_events_full = 1;
  }
  if ( !-f $events_file ) { $collect_events_full = 1; }

  #
  # collect rest of data for events
  #
  if ( $collect_events_full == 1 ) {

    # save last event id
    open( LEVE, ">$last_event_file" ) || error( "Couldn't open file $last_event_file $!" . __FILE__ . ":" . __LINE__ ) && next;
    print LEVE "$last_event_id_act";
    close(LEVE);

    SNMP_lib::collect_san( \%data, $sess, 'events', 'swEventLevel' );
    SNMP_lib::collect_san( \%data, $sess, 'events', 'swEventDescr' );

    # save all events to file
    eval { Xormon::save_as_json($events_file,\%{ $data{events} }); };
  }

  if ( -f "$last_event_file-tmp" ) { unlink("$last_event_file-tmp"); }

  #print "\$collect_events_full=$collect_events_full\n";
  #print Dumper %{ $data{events} };

  return 1;
}

sub health_status {
  my @hs             = ();
  my @status_reason  = ();
  my $status_summary = "unknown";

  #
  # switch status
  #
  if ( exists $data{'configuration'}{'connUnitStatus'} && $data{'configuration'}{'connUnitStatus'} ne '' ) {
    my $switch_status = $data{'configuration'}{'connUnitStatus'};

    if ( $switch_status eq "ok" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "ok", "data" => "The operational status of the switch $switch_name is Healthy/ok." });
      $status_summary = "ok";
      push(@status_reason, "The operational status of the switch $switch_name is Healthy/ok.")
    }
    elsif ( $switch_status eq "unused" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "ok", "data" => "The operational status of the switch $switch_name is Unmonitored." });
      $status_summary = "ok";
      push(@status_reason, "The operational status of the switch $switch_name is Unmonitored.")
    }
    elsif ( $switch_status eq "unknown" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "unknown", "data" => "The operational status of the switch $switch_name is Unknown." });
      $status_summary = "unknown";
      push(@status_reason, "The operational status of the switch $switch_name is Unknown.")
    }
    elsif ( $switch_status eq "warning" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "warning", "data" => "The operational status of the switch $switch_name is Marginal/Warning." });
      $status_summary = "warning";
      push(@status_reason, "The operational status of the switch $switch_name is Marginal/Warning.")
    }
    elsif ( $switch_status eq "failed" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "error", "data" => "The operational status of the switch $switch_name is Down/Failed." });
      $status_summary = "error";
      push(@status_reason, "The operational status of the switch $switch_name is Down/Failed.")
    }
  }
  #else {
  #  push(@hs, { "item_id" => $switch_uuid, "status" => "unknown" });
  #}

  #
  # sanport statuses
  #
  if ( exists $data{ports} ) {
    foreach ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
      my $var = $data{ports}{$_};

      if ( exists $var->{swFCPortSpecifier} ) {
        $var->{swFCPortSpecifier} =~ s/\//-/g;

        my $id = lc( Xormon::create_hash_key($switch_uuid."sanport".$var->{swFCPortSpecifier}) );

        my $oper_status = ( exists $var->{'swFCPortOpStatus'} ) ? $var->{'swFCPortOpStatus'} : "";
        my $phys_status = ( exists $var->{'swFCPortPhyState'} ) ? $var->{'swFCPortPhyState'} : "";

        # translate new states
        # validating (10) - The module is being validated.
        # invalidModule (11) - The module is invalid.
        # noSigDet (14) - No Signal is detected in the port.
        # unknown (255) - The module is unknown.
        if ( Xormon::isdigit($phys_status) ) {
          if ( $phys_status eq "10" )  { $phys_status = "validating"; }
          if ( $phys_status eq "11" )  { $phys_status = "invalidModule"; }
          if ( $phys_status eq "14" )  { $phys_status = "noSigDet"; }
          if ( $phys_status eq "255" ) { $phys_status = "unknown"; }
        }

        # only ok or error
        if ( $oper_status eq "faulty" || $phys_status eq "laserFault" || $phys_status eq "portFault" || $phys_status eq "diagFault" ) {
          push(@hs, { "item_id" => $id, "status" => "error", "data" => "Port $var->{swFCPortSpecifier}: The operational status is $oper_status and the physical status is $phys_status." });
          $status_summary = "error";
          push(@status_reason, "Port $var->{swFCPortSpecifier}: The operational status is $oper_status and the physical status is $phys_status.")
        }
        else {
          push(@hs, { "item_id" => $id, "status" => "ok", "data" => "Port $var->{swFCPortSpecifier}: The operational status is $oper_status and the physical status is $phys_status." });
        }
      }
    }
  }

  #
  # add last events
  #
  if ( $status_summary eq "error" ) {
    # check if there is some related event if the status is NOK
    my @last_events = check_last_event();
    if ( scalar(@last_events) > 0 ) {
      push( @status_reason, "Last events:\n" );
      push( @status_reason, @last_events );
    }
  }

  #
  # add switch status summary (including red states of the ports)
  #
  push(@hs, { "item_id" => $switch_uuid, "status" => $status_summary, "data" => join("\n", @status_reason) });

  # send statuses
  my $rc = $xormon->saveStatus(\@hs);
  $logger->info("save health status: $rc");
  #print Dumper @hs;
  #print "HS : send statuses : $rc\n";

  return 1;
}

sub check_last_event {
  my $events_file = "$wrkdir/events.json";

  my @events_out = ();
  if ( -f $events_file ) {
    my %events = %{ Xormon::read_json($events_file) };

    foreach my $event_id ( sort { $b <=> $a } keys( %{ $events{events} } )  ) {
      my $var = $events{events}{$event_id};
      if ( exists $var->{swEventLevel} && exists $var->{swEventDescr} && exists $var->{swEventTimeInfo} && ( $var->{swEventLevel} eq "critical" || $var->{swEventLevel} eq "error" ) ) {

        unless ( $var->{swEventTimeInfo} =~ m/^[0-9][0-9][0-9][0-9]\/[0-9][0-9]\/[0-9][0-9]-[0-9][0-9]:[0-9][0-9]:[0-9][0-9]$/ ) { last; }

        my ( $date_only, undef ) = split( /-/, $var->{swEventTimeInfo} );
        my ( undef, $eve_month, $eve_day ) = split( /\//, $date_only );

        my ( undef, undef, undef, $day_act, $month_act, undef ) = localtime();

        $eve_month =~ s/^0//;
        $eve_day   =~ s/^0//;

        $month_act += 1;

        if ( $eve_month eq $month_act && $eve_day eq $day_act ) {
          push( @events_out, "event_id: $event_id, date/time: $var->{swEventTimeInfo}, severity: $var->{swEventLevel}, description: $var->{swEventDescr}\n" );
        }
        else {
          last;    # skip older events, we just need only today events
        }
      }
    }
  }

  return @events_out;
}

sub snmp_connection_test {
  $cfg = shift;

  #
  # initialize SNMP
  #
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('san');
  #$sess = SNMP_lib::new_session( { 'Timeout' => 3000000 } ); # use 3 secs, default '1000000', micro-seconds before retry
  if ( defined $cfg->{'vfId'} && $cfg->{'vfId'} ne '' ) {
    $sess = SNMP_lib::new_session( { 'Timeout' => 1000000, Context => "VF:$cfg->{'vfId'}" } ); # use only 1 sec, don't know why, but it always takes 6 times longer, so timeout is 6 seconds in this case
  }
  else {
    $sess = SNMP_lib::new_session( { 'Timeout' => 1000000 } ); # use only 1 sec, don't know why, but it always takes 6 times longer, so timeout is 6 seconds in this case
  }

  my ( $status, $reason ) = SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysName', '', 1 ); # 1 -> conn test

  if ( $status ) {
    return 1, $data{configuration}{sysName}, SNMP_lib::build_snmpwalk_cmd($sess);
  }
  else {
    return 0, $reason, SNMP_lib::build_snmpwalk_cmd($sess);
  }
}

sub prepare_perf {
  my %stats_last;
  my %stats_now;
  my %stats_out;

  #
  # switch stats
  #
  if ( exists $data{'configuration'}{'swCpuUsage'} ) { $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{cpu_system}          = Xormon::formatNumberByType("float4",$data{'configuration'}{'swCpuUsage'}); }
  if ( exists $data{'configuration'}{'swMemUsage'} ) { $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{memory_used_percent} = Xormon::formatNumberByType("float4",$data{'configuration'}{'swMemUsage'}); }

  #
  # load last stats (it's necessary for converting counters to gauge)
  #
  my $last_stats_file  = "$wrkdir/last_stats.json";
  my $convert_counters = 0;
  my $timediff         = "";
  my @add_timestamps   = ();
  if ( -f $last_stats_file ) {
    eval { %stats_last = %{ Xormon::read_json($last_stats_file) }; };
    if ( exists $stats_last{timestamp} ) {
      $timediff = $act_time - $stats_last{timestamp};
      #if ( $timediff < ( 2 * $sample_rate ) && $timediff >= ( $sample_rate - 60 ) ) { # convert counter only when timediff is between 4 and 10 minutes
      if ( $timediff < ( 12 * $sample_rate ) && $timediff >= ( $sample_rate - 60 ) ) { # convert counter only when timediff is between 4 and 10 minutes
        $convert_counters = 1;

        # if fetch task is longer than 2x300, then is necessary to save stats also for the previous 5 minutes samples
        if ( $timediff > ( 2 * $sample_rate ) ) {
          my $max = int( $timediff / $sample_rate );
          for ( my $i = 1; $i < $max; $i++ ) {
            my $prev_timestamp = $act_time - ( $i * $sample_rate);

            if ( $prev_timestamp > $stats_last{timestamp} && ( $prev_timestamp - $stats_last{timestamp}) > ( $sample_rate - 60 ) ) {
              push(@add_timestamps, $prev_timestamp);
            }
          }
          @add_timestamps = sort @add_timestamps;
        }
      }
      else {
        $logger->error( "Cannot convert counters to gauge. \$act_time=$act_time (".localtime($act_time)."), \$stats_last{timestamp}=$stats_last{timestamp} (".localtime($stats_last{timestamp})."), \$timediff=$timediff " . __FILE__ . ":" . __LINE__ );
      }
    }
    else {
      $logger->error( "Something wrong \$stats_last{timestamp} doesn't exist!" . __FILE__ . ":" . __LINE__ );
    }
  }

  #
  # ports stats
  #
  my $ports_count = 0;
  if ( exists $data{ports} ) {
    foreach ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
      my $var = $data{ports}{$_};

      if ( exists $var->{swFCPortSpecifier} ) {
        $var->{swFCPortSpecifier} =~ s/\//-/g;

        my $id = lc( Xormon::create_hash_key($switch_uuid."sanport".$var->{swFCPortSpecifier}) );

        #
        # counters
        #
        # set current counters (convert some metrics from hex2dec)
        $stats_now{timestamp} = $act_time;
        if ( exists $var->{connUnitPortStatCountTxElements} )            { $stats_now{$id}{data_out}   = hex2dec($var->{connUnitPortStatCountTxElements}); }
        if ( exists $var->{connUnitPortStatCountRxElements} )            { $stats_now{$id}{data_in}    = hex2dec($var->{connUnitPortStatCountRxElements}); }
        if ( exists $var->{connUnitPortStatCountTxObjects} )             { $stats_now{$id}{frames_out} = hex2dec($var->{connUnitPortStatCountTxObjects}); }
        if ( exists $var->{connUnitPortStatCountRxObjects} )             { $stats_now{$id}{frames_in}  = hex2dec($var->{connUnitPortStatCountRxObjects}); }
        if ( exists $var->{connUnitPortStatCountBBCreditZero} )          { $stats_now{$id}{bb_credits} = hex2dec($var->{connUnitPortStatCountBBCreditZero}); }
        if ( exists $var->{connUnitPortStatCountInvalidCRC} )            { $stats_now{$id}{crc_errors} = hex2dec($var->{connUnitPortStatCountInvalidCRC}); }
        if ( exists $var->{connUnitPortStatCountLinkFailures} )          { $stats_now{$id}{link_fail}  = hex2dec($var->{connUnitPortStatCountLinkFailures}); }
        if ( exists $var->{connUnitPortStatCountLossofSynchronization} ) { $stats_now{$id}{loss_sync}  = hex2dec($var->{connUnitPortStatCountLossofSynchronization}); }
        if ( exists $var->{swConnUnitPCSErrorCounter} )                  { $stats_now{$id}{pcs_errors} = hex2dec($var->{swConnUnitPCSErrorCounter}); }
        if ( exists $var->{swFCPortRxEncOutFrs} )                        { $stats_now{$id}{enc_out}    = $var->{swFCPortRxEncOutFrs}; }
        if ( exists $var->{swFCPortC3Discards} )                         { $stats_now{$id}{disc_c3}    = $var->{swFCPortC3Discards}; }

        # if operational status is offline, then frames in/out must be NaN.
        if ( exists $var->{swFCPortOpStatus} && $var->{swFCPortOpStatus} eq "offline" ) {
          if ( exists $stats_now{$id}{frames_out} ) { $stats_now{$id}{frames_out} = ""; }
          if ( exists $stats_now{$id}{frames_in} )  { $stats_now{$id}{frames_in}  = ""; }

          # if port is disabled, then CRC errors must be NaN.
          if ( exists $var->{swFCPortPhyState} && $var->{swFCPortPhyState} eq "inSync" ) {
            if ( exists $stats_now{$id}{crc_errors} ) { $stats_now{$id}{crc_errors} = ""; }
          }
        }


        # prepare stats to DB (convert counters to gauge)
        if ( $convert_counters && exists $stats_last{$id} ) {
          if ( exists $stats_now{$id}{data_out}   && exists $stats_last{$id}{data_out} )   { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out}   = make_gauge($timediff,$stats_now{$id}{data_out},$stats_last{$id}{data_out}); }
          if ( exists $stats_now{$id}{data_in}    && exists $stats_last{$id}{data_in} )    { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in}    = make_gauge($timediff,$stats_now{$id}{data_in},$stats_last{$id}{data_in}); }
          if ( exists $stats_now{$id}{frames_out} && exists $stats_last{$id}{frames_out} ) { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_out} = make_gauge($timediff,$stats_now{$id}{frames_out},$stats_last{$id}{frames_out}); }
          if ( exists $stats_now{$id}{frames_in}  && exists $stats_last{$id}{frames_in} )  { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_in}  = make_gauge($timediff,$stats_now{$id}{frames_in},$stats_last{$id}{frames_in}); }
          if ( exists $stats_now{$id}{bb_credits} && exists $stats_last{$id}{bb_credits} ) { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} = make_gauge($timediff,$stats_now{$id}{bb_credits},$stats_last{$id}{bb_credits}); }
          if ( exists $stats_now{$id}{crc_errors} && exists $stats_last{$id}{crc_errors} ) { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{crc_errors} = make_gauge($timediff,$stats_now{$id}{crc_errors},$stats_last{$id}{crc_errors}); }
          if ( exists $stats_now{$id}{link_fail}  && exists $stats_last{$id}{link_fail} )  { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{link_fail}  = make_gauge($timediff,$stats_now{$id}{link_fail},$stats_last{$id}{link_fail}); }
          if ( exists $stats_now{$id}{loss_sync}  && exists $stats_last{$id}{loss_sync} )  { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{loss_sync}  = make_gauge($timediff,$stats_now{$id}{loss_sync},$stats_last{$id}{loss_sync}); }
          if ( exists $stats_now{$id}{pcs_errors} && exists $stats_last{$id}{pcs_errors} ) { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{pcs_errors} = make_gauge($timediff,$stats_now{$id}{pcs_errors},$stats_last{$id}{pcs_errors}); }
          if ( exists $stats_now{$id}{enc_out}    && exists $stats_last{$id}{enc_out} )    { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{enc_out}    = make_gauge($timediff,$stats_now{$id}{enc_out},$stats_last{$id}{enc_out}); }
          if ( exists $stats_now{$id}{disc_c3}    && exists $stats_last{$id}{disc_c3} )    { $stats_out{sanbrcd}{sanport}{$id}{$act_time}{disc_c3}    = make_gauge($timediff,$stats_now{$id}{disc_c3},$stats_last{$id}{disc_c3}); }

          # port speed + utilization + check if max speed was ixceeded
          if ( exists $var->{connUnitPortSpeed} && Xormon::isdigit($var->{connUnitPortSpeed}) ) { # INTEGER: 500000 (The speed of the port, in kilobytes per second.)
            #my $port_speed = Xormon::returnBytes($var->{connUnitPortSpeed}, "KB"); # 1024
            my $port_speed = $var->{connUnitPortSpeed} * 1000; # from KB/s to B/s
                                                               # must be *1000, ( 500 000 KB / ( 1000*1000 ) ) = 0.5 GB * 8 = 4 Gbps )

            if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in} && $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in} ne '' ) {

              if ( $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in} > $port_speed ) {
                $logger->error("The port $var->{swFCPortSpecifier}:$id max speed has been exceeded! data_in max=$port_speed, current=$stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in}");
                delete($stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in});
              }
              elsif ( $port_speed > 0 ) {
                $stats_out{sanbrcd}{sanport}{$id}{$act_time}{utilization_in} = Xormon::formatNumberByType("float4", ( $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in} / $port_speed ) * 100 );
              }
            }
            if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out} && $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out} ne '' ) {

              if ( $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out} > $port_speed ) {
                $logger->error("The port $var->{swFCPortSpecifier}:$id max speed has been exceeded! data_out max=$port_speed, current=$stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out}");
                delete($stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out});
              }
              elsif ( $port_speed > 0 ) {
                $stats_out{sanbrcd}{sanport}{$id}{$act_time}{utilization_out} = Xormon::formatNumberByType("float4", ( $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out} / $port_speed ) * 100 );
              }
            }
          }

          # missing credits in % of time
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits}) ) {
            #  $val = 0.0000025;
            #  $cmd .= " DEF:value${i}=\\\"$rrd\\\":$item_rrd:AVERAGE";
            #  $cmd .= " CDEF:valmb1_${i}=value${i},$val,*";
            #  $cmd .= " CDEF:valmb2_${i}=valmb1_${i},60,/";
            #  $cmd .= " CDEF:valmb3_${i}=valmb2_${i},100,*";
            #  $cmd .= " CDEF:valmb${i}=valmb3_${i},1000,GT,UNKN,valmb3_${i},IF";
            #  $cmd .= " $gtype:valmb${i}$color[$col_indx]:\\\"$port_space\\\"";
            #my $value = ( $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} * 0.00025 ) / 300; # replace 60 by 300, bcs step is 5 minutes now instead of 1 minute in stor
            my $value = ( $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} * 0.00025 ) / $timediff;
            if ( $value > 1000 ) {
              $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} = "";
            }
            else {
              #$stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} = $value;
              $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} = sprintf( '%.2f', $value );
            }
          }

          # aggregate stats
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{data_out} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{data_out} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{data_out} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_out};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{data_in} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{data_in} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{data_in} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{data_in};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_out} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_out}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{frames_out} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{frames_out} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_out};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{frames_out} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_out};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_in} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_in}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{frames_in} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{frames_in} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_in};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{frames_in} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{frames_in};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{bb_credits} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{bb_credits} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{bb_credits} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{bb_credits};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{crc_errors} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{crc_errors}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{crc_errors} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{crc_errors} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{crc_errors};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{crc_errors} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{crc_errors};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{link_fail} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{link_fail}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{link_fail} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{link_fail} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{link_fail};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{link_fail} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{link_fail};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{loss_sync} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{loss_sync}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{loss_sync} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{loss_sync} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{loss_sync};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{loss_sync} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{loss_sync};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{pcs_errors} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{pcs_errors}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{pcs_errors} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{pcs_errors} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{pcs_errors};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{pcs_errors} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{pcs_errors};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{enc_out} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{enc_out}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{enc_out} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{enc_out} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{enc_out};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{enc_out} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{enc_out};
            }
          }
          if ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{disc_c3} && Xormon::isdigit($stats_out{sanbrcd}{sanport}{$id}{$act_time}{disc_c3}) ) {
            if ( exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{disc_c3} ) {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{disc_c3} += $stats_out{sanbrcd}{sanport}{$id}{$act_time}{disc_c3};
            }
            else {
              $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time}{disc_c3} = $stats_out{sanbrcd}{sanport}{$id}{$act_time}{disc_c3};
            }
          }
        }


        #
        # gauge
        #
        if ( exists $var->{swSfpTxPower} ) {
          $var->{swSfpTxPower} =~ s/\s+dBm$//;
          if ( Xormon::isdigit($var->{swSfpTxPower}) ) {
            $stats_out{sanbrcd}{sanport}{$id}{$act_time}{sfp_power_out} = $var->{swSfpTxPower};
            $device_data_out->{last_sfp_stats}{timestamp}               = $act_time;
            $device_data_out->{last_sfp_stats}{$id}{sfp_power_out}      = $var->{swSfpTxPower};
            if ( exists $cfg->{'debug'} && $cfg->{'debug'} eq "sfp_metrics" ) { $logger->info("$id new sfp_power_out $device_data_out->{last_sfp_stats}{$id}{sfp_power_out}"); }
          }
        }
        if ( exists $var->{swSfpRxPower} ) {
          $var->{swSfpRxPower} =~ s/\s+dBm$//;
          if ( Xormon::isdigit($var->{swSfpRxPower}) ) {
            $stats_out{sanbrcd}{sanport}{$id}{$act_time}{sfp_power_in} = $var->{swSfpRxPower};
            $device_data_out->{last_sfp_stats}{timestamp}              = $act_time;
            $device_data_out->{last_sfp_stats}{$id}{sfp_power_in}      = $var->{swSfpRxPower};

            if ( exists $cfg->{'debug'} && $cfg->{'debug'} eq "sfp_metrics" ) {
              $logger->info("$id new sfp_power_in $device_data_out->{last_sfp_stats}{$id}{sfp_power_in}");
            }
          }
        }
        unless ( exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{sfp_power_out} && exists $stats_out{sanbrcd}{sanport}{$id}{$act_time}{sfp_power_in} ) {
          if ( exists $device_data_out->{last_sfp_stats} && exists $device_data_out->{last_sfp_stats}{timestamp} && ( $act_time - $device_data_out->{last_sfp_stats}{timestamp} ) < 7200 ) { # max 2 hours back
            if ( exists $device_data_out->{last_sfp_stats}{$id}{sfp_power_out} ) {
              $stats_out{sanbrcd}{sanport}{$id}{$act_time}{sfp_power_out} = $device_data_out->{last_sfp_stats}{$id}{sfp_power_out};

              if ( exists $cfg->{'debug'} && $cfg->{'debug'} eq "sfp_metrics" ) {
                $logger->info("$id old sfp_power_out $device_data_out->{last_sfp_stats}{$id}{sfp_power_out}");
              }
            }
            if ( exists $device_data_out->{last_sfp_stats}{$id}{sfp_power_in} ) {
              $stats_out{sanbrcd}{sanport}{$id}{$act_time}{sfp_power_in}  = $device_data_out->{last_sfp_stats}{$id}{sfp_power_in};

              if ( exists $cfg->{'debug'} && $cfg->{'debug'} eq "sfp_metrics" ) {
                $logger->info("$id old sfp_power_in $device_data_out->{last_sfp_stats}{$id}{sfp_power_in}");
              }
            }
          }
        }

        # if fetch task is longer than 2x300, then is necessary to save stats also for the previous 5 minutes samples
        if ( scalar(@add_timestamps) > 0 && exists $stats_out{sanbrcd}{sanport}{$id}{$act_time} ) {
          foreach my $new_timestamp (@add_timestamps) {
            $stats_out{sanbrcd}{sanport}{$id}{$new_timestamp} = $stats_out{sanbrcd}{sanport}{$id}{$act_time};
          }
        }

      }
      $ports_count++;
    }
  }

  if ( $ports_count == 0 ) {
    $logger->error( "No port found" . __FILE__ . ":" . __LINE__ );
  }

  if ( scalar(@add_timestamps) > 0 && exists $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time} ) {
    foreach my $new_timestamp (@add_timestamps) {
      $logger->info("prepare perf for additional timestamp: $new_timestamp ".localtime($new_timestamp));
      $stats_out{sanbrcd}{device}{$switch_uuid}{$new_timestamp} = $stats_out{sanbrcd}{device}{$switch_uuid}{$act_time};
    }
    $logger->info("prepare perf for current timestamp: $act_time ".localtime($act_time));
  }

  # save current counters for the further usage
  eval { Xormon::save_as_json($last_stats_file,\%stats_now); };

  return \%stats_out;
}

sub send_conf {
  my @architecture;
  my %conf;

  #
  # device (switch)
  #
  my %device_arch = (
    "item_id"    => $switch_uuid,
    "hostcfg_id" => $switch_uuid,
    "label"      => $switch_name,
    "hw_type"    => "sanbrcd",
    "subsystem"  => "device",
    "parents"    => []
  );
  if ( (defined $cfg->{fabric_id}    && $cfg->{fabric_id}    ne '') && defined $cfg->{fabric_label} && $cfg->{fabric_label} ne '') {
    my %fabric_arch = (
      "item_id"    => $cfg->{fabric_id},
      "label"      => $cfg->{fabric_label},
      "hw_type"    => "sanbrcd",
      "subsystem"  => "fabric",
      "parents"    => []
    );
    push(@architecture, \%fabric_arch);
  }
  if ( defined $cfg->{fabric_id}    && $cfg->{fabric_id}    ne '' ) { push(@{ $device_arch{parents} }, $cfg->{fabric_id}); } # Fabric ID for grouping purposes
  if ( defined $cfg->{fabric_label} && $cfg->{fabric_label} ne '' ) { $conf{$switch_uuid}{'fabric_name'} = $cfg->{fabric_label}; }
  push(@architecture, \%device_arch);
  # conf
  if ( exists $data{configuration}{connUnitDomainId} )     { $conf{$switch_uuid}{'domain_id'}  = hex2dec($data{configuration}{connUnitDomainId}); }
  if ( exists $data{configuration}{connUnitSn} )           { $conf{$switch_uuid}{'serial'}     = $data{configuration}{connUnitSn}; }
  if ( exists $data{configuration}{entPhysicalSerialNum} ) { $conf{$switch_uuid}{'serial'}     = $data{configuration}{entPhysicalSerialNum}; }
  if ( exists $data{'switch-type'}{2}{connUnitRevsRevId} ) { $conf{$switch_uuid}{'version'}    = $data{'switch-type'}{2}{connUnitRevsRevId}; }
  if ( exists $data{configuration}{fcFeElementName} )      { $conf{$switch_uuid}{'wwn'}        = Xormon::formatWWN($data{configuration}{fcFeElementName}); }
  if ( exists $data{configuration}{systemURL} ) {
    $conf{$switch_uuid}{'ip'} = $data{configuration}{systemURL};
    $conf{$switch_uuid}{'ip'} =~ s/http://;
    $conf{$switch_uuid}{'ip'} =~ s/\///g;
  }
  if ( exists $data{configuration}{chassisName} && $data{configuration}{chassisName} !~ m/^..\s..\s..\s..\s..\s..\s..\s..$/ ) {
    # m/^..\s..\s..\s..\s..\s..\s..\s..$/ --> chassis name is not available, but snmp returns some trash which looks like WWN of something
    $conf{$switch_uuid}{'chassis_name'} = $data{configuration}{chassisName};
  }
  if ( exists $data{'switch-type'}{1}{connUnitRevsRevId} ) {
    my $type_line = translate_brcd_type( $data{'switch-type'}{1}{connUnitRevsRevId} );
    if ( defined $type_line && $type_line ne '' ) {
      my ( undef, $speed, $model ) = split( ",", $type_line );
      $conf{$switch_uuid}{'model'} = $model;
      $conf{$switch_uuid}{'speed'} = $speed;
    }
  }
  #if ( exists $data{configuration}{sysUpTime} ) {
  #  ( $conf{$switch_uuid}{'days_up'}, undef ) = split( ":", $data{configuration}{sysUpTime} );
  #}
  my $total_ports  = 0;
  my $free_ports   = 0;
  my $avail_ports  = 0;
  my $unused_ports = 0;
  my $isl_ports    = "-"; #TODO

  #
  # sanport
  #
  if ( exists $data{ports} ) {
    foreach ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
      my $var = $data{ports}{$_};

      if ( exists $var->{swFCPortSpecifier} ) {
        $var->{swFCPortSpecifier} =~ s/\//-/g;

        my $id = lc( Xormon::create_hash_key($switch_uuid."sanport".$var->{swFCPortSpecifier}) );

        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{swFCPortSpecifier},
          "hw_type"    => "sanbrcd",
          "subsystem"  => "sanport",
          "hostcfg_id" => $switch_uuid,
          "parents"    => [$switch_uuid]
        );
        # add host WWN to architecture for SAN topology purposes
        my $host_wwns    = "";
        my $host_wwn_idx = 0;
        if ( exists $var->{connUnitPortWwn} ) {
          my $remote_wwn_count = 0;
          foreach my $id_line ( sort { $a <=> $b } keys( %{ $data{ISL} } ) ) {
            my $isl_var = $data{ISL}{$id_line};
            if ( exists $isl_var->{connUnitLinkPortWwnX} && $isl_var->{connUnitLinkPortWwnX} eq $var->{connUnitPortWwn} && exists $isl_var->{connUnitLinkPortWwnY} && $isl_var->{connUnitLinkPortWwnY} ne '' ) {
              $host_wwn_idx++;
              my $host_wwn_act = Xormon::formatWWN($isl_var->{connUnitLinkPortWwnY});
              push(@{ $arch{target_wwns} }, $host_wwn_act);
              if ( $host_wwn_idx == 1 ) {
                $host_wwns = $host_wwn_act;
              }
              else {
                $host_wwns .= ", $host_wwn_act";
              }
            }
          }
        }
        if ( $host_wwn_idx > 0 ) { $conf{$id}{'host_wwn'} = $host_wwns; }
        # port WWN
        if ( exists $var->{'connUnitPortWwn'} ) {
          my $port_wwn = Xormon::formatWWN($var->{'connUnitPortWwn'});
          $conf{$id}{'wwn'} = $port_wwn;
          $arch{wwn}        = $port_wwn;


          ####################### fake for testing ISL
          ## Brocade01 port 8 <---> Brocade101 port 10
          #if ( $port_wwn eq "200800051E9049E7" ) {
          #  $arch{target_wwns}     = ["200A00051E802092"];
          #  $conf{$id}{'host_wwn'} = "200A00051E802092";
          #}
          #if ( $port_wwn eq "200A00051E802092" ) {
          #  $arch{target_wwns}     = ["200800051E9049E7"];
          #  $conf{$id}{'host_wwn'} = "200800051E9049E7";
          #}
          ####################### fake for testing ISL
        }
        # conf
        if ( exists $var->{'swPortCompression'} ) { $conf{$id}{'compression'}        = $var->{'swPortCompression'}; }
        if ( exists $var->{'swPortEncrypt'} )     { $conf{$id}{'encrypt'}            = $var->{'swPortEncrypt'}; }
        if ( exists $var->{'swFCPortOpStatus'} )  { $conf{$id}{'operational_status'} = $var->{'swFCPortOpStatus'}; }
        if ( exists $var->{'connUnitPortName'} )  {
          $conf{$id}{'alias'} = $var->{'connUnitPortName'};
          $conf{$id}{'alias'} =~ s/\\//g;
          $conf{$id}{'alias'} =~ s/\r//g;
          $conf{$id}{'alias'} =~ s/'//g;

          # add alias to the label
          $arch{label} = "$var->{swFCPortSpecifier} [$conf{$id}{'alias'}]";
        }
        if ( exists $var->{'swFCPortPhyState'} )  {
          $conf{$id}{'physical_status'} = $var->{'swFCPortPhyState'};

          # translate new states
          # validating (10)    - The module is being validated.
          # invalidModule (11) - The module is invalid.
          # noSigDet (14)      - No Signal is detected in the port.
          # unknown (255)      - The module is unknown.
          if ( Xormon::isdigit($var->{'swFCPortPhyState'}) ) {
            if ( $var->{'swFCPortPhyState'} eq "10" )  { $conf{$id}{'physical_status'} = "validating"; }
            if ( $var->{'swFCPortPhyState'} eq "11" )  { $conf{$id}{'physical_status'} = "invalidModule"; }
            if ( $var->{'swFCPortPhyState'} eq "14" )  { $conf{$id}{'physical_status'} = "noSigDet"; }
            if ( $var->{'swFCPortPhyState'} eq "255" ) { $conf{$id}{'physical_status'} = "unknown"; }
          }
        }
        if ( exists $var->{'connUnitPortSpeed'} && Xormon::isdigit($var->{'connUnitPortSpeed'}) ) {
          $conf{$id}{'speed'}    = ( $var->{'connUnitPortSpeed'} * 8 ) / 1000000 . " Gbps";
          $conf{$id}{'speedBps'} = $var->{'connUnitPortSpeed'} * 1000; # from KB/s to B/s
        }
        # switch ports summary
        $total_ports++;
        if ( $conf{$id}{'physical_status'} eq "noTransceiver" || $conf{$id}{'physical_status'} eq "noSigDet" ) {
          $unused_ports++;
          $conf{$id}{'working_state'} = "unused";
        }
        if ( $conf{$id}{'physical_status'} eq "noLight" && $conf{$id}{'operational_status'} eq "offline" ) {
          $free_ports++;
          $avail_ports++;
        }
        if ( $conf{$id}{'physical_status'} eq "inSync" ) {
          $avail_ports++;
        }
        if ( defined $cfg->{fabric_id} && $cfg->{fabric_id} ne '' ) { push(@{ $arch{parents} }, $cfg->{fabric_id}); } # Fabric ID for grouping purposes
        push(@architecture, \%arch);
      }
    }
  }

  # switch ports summary
  $conf{$switch_uuid}{'total_ports'}     = $total_ports;
  $conf{$switch_uuid}{'free_ports'}      = $free_ports;
  $conf{$switch_uuid}{'available_ports'} = $avail_ports;
  $conf{$switch_uuid}{'unused_ports'}    = $unused_ports;
  $conf{$switch_uuid}{'isl_ports'}       = $isl_ports;

  #print "#################### Architecture\n";
  #print Dumper @architecture;
  #print "#################### CONF\n";
  #print Dumper \%conf;

  #save architecture
  my $rc_arch = $xormon->saveArchitecture(\@architecture);
  $logger->info("save architecture: $rc_arch");

  #save configuration
  my $rc_conf = $xormon->saveConfJSON(\%conf);
  $logger->info("save conf: $rc_conf");

  return 1;
}

sub collect_perf {
  # switch stats
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'swCpuUsage' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'swMemUsage' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'connUnitStatus' );                                # status of this switch
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # ports stats
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountTxElements' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountRxElements' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountTxObjects' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountRxObjects' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountBBCreditZero' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountInvalidCRC' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'swFCPortRxEncOutFrs' );                           # enc_out
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'swFCPortC3Discards' );                            # disc_c3
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountLinkFailures' );             # link_fail
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortStatCountLossofSynchronization' );    # loss_sync
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'swConnUnitPCSErrorCounter' );                     # pcs_error
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # ports cfg
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'swFCPortPhyState' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'swFCPortOpStatus' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortName' );                              # alias from the switch
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortSpeed' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'connUnitPortWwn' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports',         'swFCPortSpecifier' );                             # physical id
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  if ( $cfg_time == 1 ) {
    # perf stats which are collecten only once per hour
    SNMP_lib::collect_san( \%data, $sess, 'ports', 'swSfpRxPower' );                                        # received optical power of the SFP
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
    SNMP_lib::collect_san( \%data, $sess, 'ports', 'swSfpTxPower' );                                        # transmitted optical power of the SFP
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  }

  return 1;
}

sub collect_conf {
  # switch cfg
  #SNMP_lib::collect_san( \%data, $sess, 'configuration', '1.3.6.1.2.1.75.1.1.1',     'fcFeFabricName' );  #TODO
  SNMP_lib::collect_san( \%data, $sess, 'configuration', '1.3.6.1.2.1.75.1.1.2',     'fcFeElementName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', '1.3.6.1.2.1.47.1.1.1.1.2', 'chassisName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  #SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysUpTime' );
  #if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'connUnitSn' );         # serial number
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'systemURL' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'connUnitDomainId' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'switch-type',   'connUnitRevsRevId' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # ports cfg
  SNMP_lib::collect_san( \%data, $sess, 'ISL',   'connUnitLinkPortWwnX' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ISL',   'connUnitLinkPortWwnY' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'swPortEncrypt' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'swPortCompression' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  # if serial number is "none" or WWN, use Factory serial number
  if ( !exists $data{'configuration'}{'connUnitSn'} || $data{'configuration'}{'connUnitSn'} eq '' || $data{'configuration'}{'connUnitSn'} =~ m/none/ || $data{'configuration'}{'connUnitSn'} =~ m/:/ ) {
    SNMP_lib::collect_san( \%data, $sess, 'configuration', '1.3.6.1.2.1.47.1.1.1.1.11', 'entPhysicalSerialNum' );
    if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  }

  #TODO
  #if ( $ENV{FISERV} ) { # customer modification
  #  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'fruSupplierSerialNum' ); # serial number
  #}


  return 1;
}

sub translate_brcd_type {
  my $type = shift;

  my @all_types = (
    "2,1 Gbps,Brocade 2010",
    "3,1 Gbps,Brocade 2400",
    "6,1 Gbps,Brocade 2800",
    "9,2 Gbps,Brocade 3800",
    "10,2 Gbps,Brocade 12000",
    "12,2 Gbps,Brocade 3900",
    "16,2 Gbps,Brocade 3200",
    "21,2 Gbps,Brocade 24000",
    "22,2 Gbps,Brocade 3016",
    "26,2 Gbps,Brocade 3850",
    "27,2 Gbps,Brocade 3250",
    "29,2 Gbps,Brocade 4012",
    "32,4 Gbps,Brocade 4100",
    "34,2 Gbps,Brocade 200E",
    "37,2 Gbps,Brocade 4020",
    "38,2 Gbps,Brocade AP7420",
    "42,4 Gbps,Brocade 48000",
    "43,4 Gbps,Brocade 4024",
    "44,4 Gbps,Brocade 4900",
    "45,2 Gbps,Brocade 4016",
    "46,4 Gbps,Brocade 7500",
    "50,UN,Brocade CP8",
    "51,2 Gbps,Brocade 4018",
    "58,4 Gbps,Brocade 5000",
    "61,2 Gbps,Brocade 4424",
    "62,8 Gbps,Brocade DCX",
    "64,8 Gbps,Brocade 5300",
    "66,8 Gbps,Brocade 5100",
    "67,8 Gbps,Brocade Encryption Switch",
    "70,8 Gbps,Brocade 5410",
    "71,8 Gbps,Brocade 300",
    "72,8 Gbps,Brocade 5480",
    "73,8 Gbps,Brocade 5470",
    "75,8 Gbps,Brocade M5424",
    "76,UN,Brocade 8000",
    "77,8 Gbps,Brocade DCX-4S",
    "83,8 Gbps,Brocade 7800",
    "86,UN,Brocade 450",
    "87,UN,Brocade 460",
    "92,UN,Brocade A-40FC",
    "96,16 Gbps,Brocade FC16-48",
    "97,16 Gbps,Brocade FC16-32",
    "98,16 Gbps,Brocade CR16-8",
    "99,16 Gbps,Brocade CR16-4",
    "109,16 Gbps,Brocade 6510",
    "117,16 Gbps,Brocade 6547",
    "118,16 Gbps,Brocade 6505",
    "120,16 Gbps,Brocade DCX8510-8",
    "121,16 Gbps,Brocade DCX8510-4",
    "124,UN,Brocade 430",
    "125,8 Gbps,Brocade 5431",
    "129,16 Gbps,Brocade 6548",
    "130,16 Gbps,Brocade M6505",
    "133,16 Gbps,Brocade 6520",
    "134,8 Gbps,Brocade 5432",
    "148,16 Gbps,Brocade 7840",
    "149,16 Gbps,Brocade 6546",
    "150,16 Gbps,Brocade 6545",
    "152,16 Gbps,Brocade 6549",
    "153,16 Gbps,Brocade FC16-64",
    "156,16 Gbps,Brocade 6543",
    "157,16 Gbps,Brocade 6558",
    "158,16 Gbps,Brocade 6559",
    "162,32 Gbps,Brocade G620",
    "165,32 Gbps,Brocade X6-4",
    "166,32 Gbps,Brocade X6-8",
    "167,16 Gbps,Brocade 6542",
    "170,32 Gbps,Brocade G610",
    "173,128 Gbps,Brocade G630",
    "175,UN,Brocade CPX6",
    "176,32 Gbps,Brocade CR32-4",
    "177,32 Gbps,Brocade CR32-8",
    "178,32 Gbps,Brocade FC32-48",
    "179,64 Gbps,Brocade X7-4",
    "180,64 Gbps,Brocade X7-8",
    "181,128 Gbps,Brocade G720",
    "183,128 Gbps,Brocade G620",
    "184,128 Gbps,Brocade G630",
    "186,16 Gbps,Brocade SX6",
  );

  # 178 has already been used
  # 178 Brocade 7810 Extension Switch Extension switch with configurable scaling of 12 ports and supports 32G, 16G, 8G, and 4G media types. It supports 1GbE and 10GbE Ethernet interfaces.

  $type =~ s/\.[0-9]*$//;
  my $type_line = "";
  ($type_line) = grep {/^$type,/} @all_types;

  return ($type_line);
}

sub hex2dec {
  my $num = shift;

  $num =~ s/\s+//g;
  my $hex = "0x$num";
  my $dec = Math::BigInt->new($hex);
  my $int = "$dec"; # convert object to string (number)

  # example: convert object to string (number)
  #use Math::BigInt;
  #sub bigint { Math::BigInt->new(shift); }
  #$x = Math::BigInt->bstr("1234")       # string "1234"
  #$x = "$x";                            # same as bstr()

  return $int;
}

sub make_gauge {
  my $timediff     = shift;
  my $counter_now  = shift;
  my $counter_last = shift;

  if ( Xormon::isdigit($timediff) && Xormon::isdigit($counter_now) && Xormon::isdigit($counter_last) && $timediff > 0 ) {
    if ( $counter_now > $counter_last ) {
      #return Xormon::formatNumberByType("float4",int( ( $counter_now - $counter_last ) / $timediff )); # --> this is not necessary, value is already formated as an integer
      return int( ( $counter_now - $counter_last ) / $timediff );
    }
    elsif ( $counter_now < $counter_last ) {
      # probably counter overflow
      return "";
    }
    else {
      return 0;
    }
  }

  return "";
}

1;
