package SANnav;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Data::Dumper;
use Xormon;
use JSON qw(encode_json decode_json);
use XormonLogger;

my $PYTHON = ( -f "../../venv-sannav/bin/python3" ) ? "../../venv-sannav/bin/python3" : "python3";

my $xormon;
my $cfg;
my $wrkdir;
my $kafkadir;
my $act_dir;
my $logger;

my %data;
my %perf_data;
my %conf_data;
my %directors;
my %ports_wwns;
my %errors;
my %isl_map;

my $sessionId;
my $kafka_ip;
my $sannav_name;
my $sannav_uuid;

my $device_data_in;
my $device_data_out = {};
my $cfg_time        = 1;
my $act_time        = "";
my $sample_rate     = 300;
my $connection_test = 0;
my @rows;

sub collect {
  $xormon    = shift;
  $cfg       = shift;
  $wrkdir    = shift;
  $kafkadir  = shift;
  $act_time  = time();
  my $status = "error";

  #
  # test main variables
  #
  $sannav_uuid = ( exists $cfg->{'hostcfg_id'} )  ? $cfg->{'hostcfg_id'}  : Xormon::error("Cannot continue, 'hostcfg_id' is not configured!")  && return $status;
  $sannav_name = ( exists $cfg->{'hostalias'} )   ? $cfg->{'hostalias'}   : Xormon::error("Cannot continue, 'hostalias' is not configured!")   && return $status;
  $logger      = XormonLogger->new( 'sannav', $cfg->{hostcfg_id}, $cfg->{hostalias} );
  unless ( exists $cfg->{'kafkaHost'} )   { $logger->error("Cannot continue, 'kafkaHost' is not configured!")   && return $status; }
  unless ( exists $cfg->{'apiHost'} )     { $logger->error("Cannot continue, 'apiHost' is not configured!")     && return $status; }
  unless ( exists $cfg->{'apiPort'} )     { $logger->error("Cannot continue, 'apiPort' is not configured!")     && return $status; }
  unless ( exists $cfg->{'apiUsername'} ) { $logger->error("Cannot continue, 'apiUsername' is not configured!") && return $status; }
  unless ( exists $cfg->{'apiPassword'} ) { $logger->error("Cannot continue, 'apiPassword' is not configured!") && return $status; }
  $kafka_ip = $cfg->{'kafkaHost'};

  #
  # collect perf data
  #
  unless ( get_stats() ) { return $status; }

  #
  # get microservice data for this device
  #
  $device_data_in  = $xormon->getData();
  $device_data_out = {};

  #
  # check if it is time to collect configuration (only once per hour)
  #
  if ( exists $device_data_in->{last_conf_timestamp} && time() - $device_data_in->{last_conf_timestamp} < 3600 ) {
    $cfg_time = 0;
    $device_data_out->{last_conf_timestamp} = $device_data_in->{last_conf_timestamp};
  }
  else {
    $device_data_out->{last_conf_timestamp} = time(); # conf_time=1
  }

  #
  # login to SANnav API (get sessionId), connection status
  #
  my $login = new_post_request("external-api/v1/login/","");
  if ( defined $login ) {
    $xormon->isConnected(1);
  }
  else {
    $xormon->isConnected(0, "Cannot login to SANnav API!");
    $logger->error("Cannot login to SANnav API! Exiting...") && return $status;
  }

  #
  # collect conf data
  #
  get_conf_data();

  #
  # send arch, conf and perf separately for each switch
  #
  send_conf("sannavmgmt");
  if ( exists $conf_data{switches} ) {
    foreach my $switch_name ( sort keys %{ $conf_data{switches} } ) {
      if ( exists $conf_data{switches}{$switch_name}{conf} && exists $conf_data{switches}{$switch_name}{ports} ) {
        #
        # send arch and conf (once per hour)
        #
        send_conf($switch_name);

        #
        # health status
        #
        health_status($switch_name);

        #
        # send perf data
        #
        my $perf_out = prepare_perf($switch_name);
        my $rc       = $xormon->saveJSON2($perf_out);
        if ( defined $rc ) {
          $logger->info("$switch_name: save perf: $rc");

          if ( $rc eq "201" ) {
            $status = "successful";
          }
          elsif ( $rc eq "401" ) {
            $status = "unauthorized";
            last;
          }
          else {
            $status = "error";
            last;
          }
        }
        else {
          $logger->info("$switch_name: save perf: UN");
        }
      }
    }
  }

  #
  # perf data status
  #
  if ( $status eq "successful" ) {
    $xormon->hasData(1);
  }
  else {
    $xormon->hasData(0, "no available stats: $status");
  }

  #
  # events, zones
  #
  if ( $cfg_time ) {
    #collect_events(); # collect events only once per hour
    #collect_zones(); # collect events only once per hour
  }
  #send_events(); # every MS run, because there is possibility to change presented severity

  #
  # logout from SANnav API
  #
  new_post_request("external-api/v1/logout/","");

  #
  # save microservice data for this device
  #
  $xormon->saveData(\%{ $device_data_out });

  return $status;
}

sub connection_test {
  my $client       = shift;
  $cfg             = shift;
  $wrkdir          = shift;
  $connection_test = 1;
  my $test         = Test->new($client);

  #
  # test main variables
  #
  my @variables = ("hostcfg_id", "hostalias", "apiHost", "apiPort", "apiUsername", "apiPassword", "kafkaHost", "registryPort", "brokerPort");
  foreach my $variable (@variables){
    unless ( exists $cfg->{$variable} ){
      $test->addSection("Device cfg", 0 , "Cannot continue, $variable is not configured!");
      return $test->stop();
    }
  }
  $test->addSection("Device cfg", 1 , "");

  $logger = XormonLogger->new( 'sannav', $cfg->{hostcfg_id}, $cfg->{hostalias} );

  #
  # test SANnav IP and port
  #
  my ( $sannav_code, $sannav_text ) = Xormon::conntest_tcp($cfg->{'apiHost'},$cfg->{'apiPort'});
  $test->addSection("SANnav API host network connection", $sannav_code, $sannav_text);
  if ( $sannav_code == 0 ) { return $test->stop(); }

  #
  # test KAFKA network connection (usually on the same machine as Xormon NG)
  #
  # schema registry
  my ( $kafka_registry_code, $kafka_registry_text ) = Xormon::conntest_tcp($cfg->{'kafkaHost'},$cfg->{'registryPort'});
  $test->addSection("Kafka Schema registry network connection", $kafka_registry_code, $kafka_registry_text);
  if ( $kafka_registry_code == 0 ) { return $test->stop(); }
  # broker
  my ( $kafka_broker_code, $kafka_broker_text ) = Xormon::conntest_tcp($cfg->{'kafkaHost'},$cfg->{'brokerPort'});
  $test->addSection("Kafka broker network connection", $kafka_broker_code, $kafka_broker_text);
  if ( $kafka_broker_code == 0 ) { return $test->stop(); }

  #
  # SANnav REST API
  #
  # login (get sessionId)
  my $login = new_post_request("external-api/v1/login/","");
  unless ( defined $login ) {
    push(@rows, "Cannot login to SANnav API!\n");
    push(@rows, "Connection failed!\n");
    $test->addSection("API authorization", 0 , \@rows);
    return $test->stop();
  }

  # switches cfg
  get_switches_conf();
  my %fabrics;
  my @out        = ();
  my $switch_idx = 0;
  my $fabric_idx = 0;
  if ( exists $conf_data{switches} ) {
    foreach my $switch_name ( sort keys %{ $conf_data{switches} } ) {
      if ( exists $conf_data{switches}{$switch_name}{conf} ) {
        $switch_idx++;
        if ( exists $conf_data{switches}{$switch_name}{conf}{fabricName} ) {
          my $fab_name = $conf_data{switches}{$switch_name}{conf}{fabricName};
          unless ( exists $fabrics{$fab_name} ) {
            $fabric_idx++;
            $fabrics{$fab_name} = $fab_name;
          }
        }
      }
    }
  }

  # logout from SANnav API
  new_post_request("external-api/v1/logout/","");

  if ( $switch_idx > 0 ) {
    push(@rows, "Switches : $switch_idx\n");
    push(@rows, "Fabrics  : $fabric_idx\n");

    $test->addSection("API authorization", 1 , \@rows);
    @rows = ();
  }
  else {
    push(@rows, "SAN switches have not been found!\n");
    push(@rows, "Connection failed!\n");
    $test->addSection("API authorization", 0 , \@rows);
    return $test->stop();
  }

  #
  # Python
  #
  my @output = `$PYTHON -V >$wrkdir/python3-test.txt 2>&1; cat $wrkdir/python3-test.txt; rm $wrkdir/python3-test.txt`;
  @rows = ("$PYTHON -V\n");
  if ( (grep {/^Python\s+3\./} @output) > 0 ) {
    push(@rows, @output);
    $test->addSection("Python", 1 , \@rows);
    @rows = ();
  }
  else {
    push(@rows, @output);
    $test->addSection("Python", 0 , \@rows);
    return $test->stop();
  }

  #
  # Kafka stats
  #
  my $last_stats_json = "$wrkdir/$cfg->{kafkaHost}-last-stats.json";
  if ( -f $last_stats_json ) {
    if ( Xormon::file_time_diff($last_stats_json) < 900 ) {
      $test->addSection("Kafka statistics", 1 , \@rows); # OK, stats found
    }
    else {
      push(@rows, "Current statistics are not available.\n");
      push(@rows, "act_time   : ".localtime()."\n");
      push(@rows, "last_stats : ".localtime( (stat($last_stats_json))[9] )."\n");
      #print "For more info check these logs:\n";
      #print "    $kafka_output_log\n";
      #print "    $kafka_error_log\n";
      #print "    $logsdir/kafkanb.log\n";
      #print "\nBasic Troubleshooting:\n";
      #print "    https://stor2rrd.com/SANnav-monitoring-scripts.php#troubleshooting\n";
      $test->addSection("Kafka statistics", 0 , \@rows);
      return $test->stop();
    }
  }
  else {
    if ( is_process_running("sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG") ) {
      # kafka consumer daemon is running, but it doesn't provide stats
      push(@rows, "Kafka consumer daemon is running (sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'}), but it doesn't provide stats.\n");
      push(@rows, "Try this test again after 10 minutes if this is the first connection test with the current SANnav configuration.\n");
      #print "For more info check these logs:\n";
      #print "    $kafka_output_log\n";
      #print "    $kafka_error_log\n";
      #print "    $logsdir/kafkanb.log\n";
      #print "\nBasic Troubleshooting:\n";
      #print "    https://stor2rrd.com/SANnav-monitoring-scripts.php#troubleshooting\n";
      $test->addSection("Kafka statistics", 0 , \@rows);
      return $test->stop();
    }
    else {
      push(@rows, "Kafka consumer daemon is not running (sannavperf.py $kafka_ip $cfg->{'hostalias'}).\n");
      push(@rows, "Try this connection test again after 10 minutes.\n");
      $test->addSection("Kafka statistics", 0 , \@rows);
      return $test->stop();
    }
  }

  return $test->stop();
}

sub collect_zones {
  if ( exists $conf_data{fabrics} ) {
    foreach my $fabric_name ( keys %{ $conf_data{fabrics} } ) {
      if ( exists $conf_data{fabrics}{$fabric_name}{conf}{id} ) {
        my $fabric_id = $conf_data{fabrics}{$fabric_name}{conf}{id};

        my $fetchZones = new_post_request("zoning/fabrics/fetchZones",{'fabricId'=>$fabric_id,'multiSortCriteria'=>{'noOfRecords'=>1000,'columnDetails'=>[{'sortColumn'=>'zonename','sortOrder'=>'ASC'}]}});
        unless ( $fetchZones ) { next; }
        if ( defined $fetchZones && ref($fetchZones) eq "ARRAY" ) {
          $conf_data{zones}{$fabric_name}{fetchZones} = $fetchZones;
        }

        my $fetchZoneAliasesPerFabric = new_post_request("zoning/fabrics/fetchZoneAliasesPerFabric",{'fabricId'=>$fabric_id,'multiSortCriteria'=>{'noOfRecords'=>1000,'columnDetails'=>[{'sortColumn'=>'zonename','sortOrder'=>'ASC'}]}});
        unless ( $fetchZoneAliasesPerFabric ) { next; }
        if ( defined $fetchZoneAliasesPerFabric && ref($fetchZoneAliasesPerFabric) eq "ARRAY" ) {
          $conf_data{zones}{$fabric_name}{fetchZoneAliasesPerFabric} = $fetchZoneAliasesPerFabric;
        }
      }
    }
  }
  #print Dumper %{ $conf_data{zones} };

  return 1;
}

sub collect_events {
  my $result = new_post_request("fm/events",{'severityGroup'=>'ALL','eventCategory'=>'ALL','timeScope'=>'OneMonth','filters'=>{},'fabricGuids'=>[],'pageSize'=>1000,'orderBy'=>'lastOccurrenceHostTime','startIndex'=>0,'sortOrder'=>'DESC','acknowledgedEvents'=>'false','orderBy'=>'lastOccurrenceHostTime'});
  print Dumper $result;

  #if ( defined $result && ref($result) eq "ARRAY" ) {
  #  foreach my $var ( @{ $result } ) {
  #    #print Dumper $var;
  #    if ( exists $var->{name} ) {
  #      $conf_data{fabrics}{ $var->{name} }{conf} = $var;
  #    }
  #  }
  #}

  return 1;
}

sub health_status {
  my $switch_name = shift;

  my @hs             = ();
  my @status_reason  = ();
  my $status_summary = "unknown";
  my $switch_var     = $conf_data{switches}{$switch_name}{conf};

  unless ( exists $switch_var->{guid} ) { $logger->error("$switch_name: Switch guid was not found! cannot continue... $!" . __FILE__ . ":" . __LINE__ ) && return 0; }

  #
  # switch status
  #
  if ( exists $switch_var->{status} && exists $switch_var->{reason} ) {
    if ( $switch_var->{status} eq "Healthy" ) {
      $status_summary = "ok";
    }
    else {
      $status_summary = "error";
    }
    push(@status_reason, "The operational status of the switch $switch_name is $switch_var->{status}. The status reason is $switch_var->{reason}.")
  }

  #
  # sanport statuses
  #
  foreach my $port_id ( sort { $a <=> $b } keys %{ $conf_data{switches}{$switch_name}{ports} } ) {
    my $port_var  = $conf_data{switches}{$switch_name}{ports}{$port_id};
    my $port_name = set_port_name($switch_name,$port_var);

    unless ( defined $port_name ) { next; }
    unless ( exists $port_var->{guid} ) { $logger->error("$switch_name:$port_id:$port_name Port guid was not found! Port skipped... $!" . __FILE__ . ":" . __LINE__ ) && next; }

    # Health status
    my $operational_status = ( exists $port_var->{state} )  ? $port_var->{state}  : "";
    my $physical_status    = ( exists $port_var->{status} ) ? $port_var->{status} : "";

    # GREEN:
    #########################################
    #In_Sync
    #The module is receiving light and in sync.

    #Online
    #The port is up and running.

    #Mod_Val
    #Module validation in process.

    # GREY:
    #########################################
    #No_Card
    #No interface card present.

    #No_Module
    #No module (GBIC or other) present.

    #No_SigDet
    #No signal is detected on the port. For 16Gb/s-capable interchassis link (ICL) ports, this state replaces the No_Light indicator. It indicates that a quad small form-factor pluggable (QSFP) has been installed but is not connected with a cable.

    # ORANGE:
    #########################################
    #No_Light
    #The module is not receiving light. This state is not applicable to 16Gb/s-capable interchassis link (ICL) ports.

    #No_Sync
    #The module is receiving light but is out of sync.

    #Lock_Ref
    #The port is locking to the reference signal.

    #Offline
    #A port connection is not established (for virtual ports only).

    #Disabled

    #Testing
    #The port is running diagnostics.

    #Transient
    #The port is disabled and re-enabled by MAPS to recover from the bottleneck condition caused by the target device.

    # RED:
    #########################################
    #Laser_Flt
    #The module is signaling a laser fault.

    #Port_Flt
    #The port is marked faulty.

    #Hard_Flt
    #The port is hard faulted.

    #Mod_Inv
    #Module speed mismatch or incompatible SFP.

    #Mod_Uns
    #Module unsupported.

    if ( $operational_status =~ m/Faulty/i || $physical_status =~ m/Laser_Flt/i || $physical_status =~ m/Port_Flt/i || $physical_status =~ m/Hard_Flt/i || $physical_status =~ m/Mod_Inv/i || $physical_status =~ m/Mod_Uns/i ) {
      push(@hs, { "item_id" => $port_var->{guid}, "status" => "error", "data" => "Port $port_name: The operational status is $operational_status and the physical status is $physical_status." });
      $status_summary = "error";
      push(@status_reason, "Port $port_name: The operational status is $operational_status and the physical status is $physical_status.")
    }
    else {
      push(@hs, { "item_id" => $port_var->{guid}, "status" => "ok", "data" => "Port $port_name: The operational status is $operational_status and the physical status is $physical_status." });
    }
  }

  #
  # add last events
  #
  #if ( $status_summary eq "error" ) {
  #  # check if there is some related event if the status is NOK
  #  my @last_events = check_last_event();
  #  if ( scalar(@last_events) > 0 ) {
  #    push( @status_reason, "Last events:\n" );
  #    push( @status_reason, @last_events );
  #  }
  #}

  #
  # add switch status summary (including red states of the ports)
  #
  push(@hs, { "item_id" => $switch_var->{guid}, "status" => $status_summary, "data" => join("\n", @status_reason) });

  # send statuses
  my $rc = $xormon->saveStatus(\@hs);
  $logger->info("$switch_name: save health status: $rc");
  #print Dumper @hs;
  #print "HS : send statuses : $rc\n";

  return 1;
}

sub prepare_perf {
  my $switch_name = shift;
  my %stats_out;

  my $switch_var = $conf_data{switches}{$switch_name}{conf};

  if ( exists $switch_var->{guid} ) {
    #
    # go through all intervals and check if there are stats available for the ports and for this switch
    #
    foreach my $timestamp ( sort { $a <=> $b } keys %perf_data ) {
      #print Dumper \%{ $perf_data{$timestamp} }; last;
      #print "$timestamp : ".localtime($timestamp)."\n";

      my $stats_found        = 0;
      my $switch_stats_found = 0;
      my $this_time_stats    = 0;

      #
      # switch stats
      #
      my $stats;
      if ( exists $switch_var->{wwn} && exists $perf_data{$timestamp}{ $switch_var->{wwn} } ) {
        $stats = $perf_data{$timestamp}{ $switch_var->{wwn} };
        $switch_stats_found++;
      }
      elsif ( exists $switch_var->{chassisSerialNumber} && exists $conf_data{chassis}{ $switch_var->{chassisSerialNumber} }{conf}{wwn} ) {
        # sometimes Kafka returns WWN of the chassis instead of the WWN of the switch, but it's called "switch_wwn", so it's necessary to test both WWNs (switch and chassis)
        # {"switch_wwn": "10:00:00:05:1E:DD:F0:EB", "statistics_collection_time": 1654594500240, "up_time": 2521591704, "cpu_utilization": 5.0, "memory_utilization": 44.0}
        # 10:00:00:05:1E:DD:F0:EB --> is not switch_wwn, but it is chassis WWN
        my $chassis_wwn = $conf_data{chassis}{ $switch_var->{chassisSerialNumber} }{conf}{wwn};
        if ( exists $perf_data{$timestamp}{$chassis_wwn} ) {
          $stats = $perf_data{$timestamp}{$chassis_wwn};
          $switch_stats_found++;
        }
      }

      if ( $switch_stats_found > 0 ) {
        $stats_found++;
        #print Dumper $stats;
        #'memory_utilization' => '24',
        #'statistics_collection_time' => '1724852100095',
        #'cpu_utilization' => '6',
        #'switch_wwn' => '10:00:C4:F5:7C:DA:A9:4F',
        #'up_time' => 2843255500
        if ( exists $stats->{statistics_collection_time} ) {
          my $stats_time = convert_collection_time($stats->{statistics_collection_time});
          if ( defined $stats_time ) {
            if ( exists $stats->{cpu_utilization} )    { $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{cpu_system}          = Xormon::formatNumberByType("float4",$stats->{cpu_utilization}); }
            if ( exists $stats->{memory_utilization} ) { $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{memory_used_percent} = Xormon::formatNumberByType("float4",$stats->{memory_utilization}); }

            $logger->info("$switch_name: prepare perf: add switch stats for timestamp: $stats_time ".localtime($stats_time));
          }
          else {
            $logger->error("Unsupported statistics_collection_time! Stats skipped... statistics_collection_time=\"$stats->{statistics_collection_time} $!" . __FILE__ . ":" . __LINE__ );
            $logger->error(Dumper $stats);
          }
        }
      }

      #
      # ports stats
      #
      foreach my $port_id ( sort { $a <=> $b } keys %{ $conf_data{switches}{$switch_name}{ports} } ) {
        my $port_var = $conf_data{switches}{$switch_name}{ports}{$port_id};

        if ( $port_var->{guid} && exists $port_var->{wwn} && exists $perf_data{$timestamp}{ $port_var->{wwn} } ) {
          $stats_found++;
          my $stats     = $perf_data{$timestamp}{ $port_var->{wwn} };
          my $port_name = set_port_name($switch_name,$port_var);

          unless ( defined $port_name ) { next; }

          #print Dumper $stats;
          #'switch_wwn' => '10:00:C4:F5:7C:DA:A9:40',
          #'loss_of_signal' => 4,
          #'media_rx_power' => '-3.70000004768372',
          #'credit_stall_state' => 0,
          #'invalid_transmission_words' => 1,
          #'tx_byte_rate' => 0,
          #'primitive_sequence_protocol_errors' => 0,
          #'class3_out_discards' => 0,
          #'in_link_resets' => 2,
          #'oversubscription_state' => 0,
          #'transmit_queue_latency' => 0,
          #'encoding_disparity_errors' => 0,
          #'port_quarantined' => 0,
          #'media_tx_power' => '-2.5',
          #'user_port_index' => 3,
          #'frame_too_long' => 0,
          #'invalid_ordered_sets' => 2,
          #'bad_eofs_received' => 0,
          #'class3_in_discards' => 0,
          #'statistics_collection_time' => '1724913600017',
          #'truncated_frames' => 0,
          #'credit_stall_update_time' => 0,
          #'media_current' => '6.9060001373291',
          #'media_voltage' => '3294.30004882812',
          #'fec_uncorrected' => 0,
          #'in_crc_errors' => 0,
          #'fec_corrected' => 0,
          #'class3_discards_other' => 1,
          #'single_credit_loss' => 0,
          #'multi_credit_loss' => 0,
          #'class3_out_unreachable' => 0,
          #'fabric_port_wwn' => '20:03:C4:F5:7C:DA:A9:40',
          #'out_link_resets' => 2,
          #'other_credit_loss' => 0,
          #'rx_byte_rate' => 0,
          #'media_temperature' => 50
          my $last_stats_file = "$wrkdir/$switch_name-$port_name-last-stats.json";
          if ( -f $last_stats_file ) {
            my $last_stats = Xormon::read_json($last_stats_file);

            if ( exists $stats->{statistics_collection_time} ) {
              my $stats_time = convert_collection_time($stats->{statistics_collection_time});
              if ( defined $stats_time ) {

                # data_out, data_in
                if ( exists $switch_var->{firmware} && $switch_var->{firmware} =~ m/^v9/ ) { # FW 9+ doesn't provide these stats in octets
                  my $bytes_tra = make_gauge($switch_name,"out_octets",$stats,$last_stats);
                  if ( Xorux_lib::isdigit($bytes_tra) ) {
                    $bytes_tra = $bytes_tra / 1000;
                  }
                  $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} = $bytes_tra;

                  my $bytes_rec = make_gauge($switch_name,"in_octets",$stats,$last_stats);
                  if ( Xorux_lib::isdigit($bytes_rec) ) {
                    $bytes_rec = $bytes_rec / 1000;
                  }
                  $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in} = $bytes_rec;
                }
                else {
                  $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} = make_gauge($switch_name,"out_octets",$stats,$last_stats);
                  $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in}  = make_gauge($switch_name,"in_octets",$stats,$last_stats);
                }

                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_out} = make_gauge($switch_name,"out_frames",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_in}  = make_gauge($switch_name,"in_frames",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{bb_credits} = make_gauge($switch_name,"bb_credit_zero",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{crc_errors} = make_gauge($switch_name,"crc_errors",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{enc_out}    = make_gauge($switch_name,"encoding_errors_outside_frame",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{disc_c3}    = make_gauge($switch_name,"class3_discards",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{link_fail}  = make_gauge($switch_name,"link_failures",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{loss_sync}  = make_gauge($switch_name,"loss_of_sync",$stats,$last_stats);
                $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{pcs_errors} = make_gauge($switch_name,"pcs_block_errors",$stats,$last_stats);

                # SFP stats
                if ( exists $stats->{media_tx_power} && Xormon::isdigit($stats->{media_tx_power}) ) {
                  $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{sfp_power_out} = sprintf( '%.2f', $stats->{media_tx_power});
                }
                if ( exists $stats->{media_rx_power} && Xormon::isdigit($stats->{media_rx_power}) ) {
                  $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{sfp_power_in} = sprintf( '%.2f', $stats->{media_rx_power});
                }

                # port speed + utilization + check if max speed was ixceeded
                if ( exists $port_var->{speed} && $port_var->{speed} =~ m/Gb\/s$/ ) {
                  my $port_speed = $port_var->{speed};
                  $port_speed    =~ s/\s+Gb\/s$//;
                  if ( Xormon::isdigit($port_speed) ) {
                    $port_speed /= 8;    # from Gb to GB
                    $port_speed *= 1000; # from GB to MB
                    $port_speed *= 1000; # from MB to kB
                    $port_speed *= 1000; # from kB to B

                    if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in} && $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in} ne '' ) {

                      if ( $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in} > $port_speed ) {
                        $logger->error("$switch_name: port $port_name max speed has been exceeded! data_in max=$port_speed, current=$stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in}");
                        delete($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in});
                      }
                      elsif ( $port_speed > 0 ) {
                        $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{utilization_in} = Xormon::formatNumberByType("float4", ( $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in} / $port_speed ) * 100 );
                      }
                    }
                    if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} && $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} ne '' ) {

                      if ( $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} > $port_speed ) {
                        $logger->error("$switch_name: port $port_name max speed has been exceeded! data_out max=$port_speed, current=$stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out}");
                        delete($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out});
                      }
                      elsif ( $port_speed > 0 ) {
                        $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{utilization_out} = Xormon::formatNumberByType("float4", ( $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} / $port_speed ) * 100 );
                      }
                    }
                  }
                }

                # aggregate stats
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{data_out} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{data_out} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{data_out} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_out};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{data_in} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{data_in} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{data_in} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{data_in};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_out} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_out}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{frames_out} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{frames_out} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_out};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{frames_out} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_out};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_in} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_in}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{frames_in} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{frames_in} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_in};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{frames_in} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{frames_in};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{bb_credits} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{bb_credits}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{bb_credits} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{bb_credits} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{bb_credits};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{bb_credits} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{bb_credits};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{crc_errors} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{crc_errors}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{crc_errors} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{crc_errors} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{crc_errors};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{crc_errors} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{crc_errors};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{link_fail} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{link_fail}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{link_fail} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{link_fail} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{link_fail};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{link_fail} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{link_fail};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{loss_sync} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{loss_sync}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{loss_sync} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{loss_sync} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{loss_sync};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{loss_sync} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{loss_sync};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{pcs_errors} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{pcs_errors}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{pcs_errors} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{pcs_errors} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{pcs_errors};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{pcs_errors} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{pcs_errors};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{enc_out} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{enc_out}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{enc_out} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{enc_out} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{enc_out};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{enc_out} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{enc_out};
                  }
                }
                if ( exists $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{disc_c3} && Xormon::isdigit($stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{disc_c3}) ) {
                  if ( exists $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{disc_c3} ) {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{disc_c3} += $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{disc_c3};
                  }
                  else {
                    $stats_out{sannav}{device}{$switch_var->{guid}}{$stats_time}{disc_c3} = $stats_out{sannav}{sanport}{$port_var->{guid}}{$stats_time}{disc_c3};
                  }
                }

                $this_time_stats++;
                if ( $this_time_stats == 1 ) { # report it only once
                  $logger->info("$switch_name: prepare perf: add ports stats for timestamp: $stats_time ".localtime($stats_time));
                }

              }
              else {
                $logger->error("Unsupported statistics_collection_time! Stats skipped... statistics_collection_time=\"$stats->{statistics_collection_time} $!" . __FILE__ . ":" . __LINE__ );
                $logger->error(Dumper $stats);
              }
            }
          }
          else {
            $logger->error("Cannot save stats, last stats file doesn't exist $last_stats_file $!" . __FILE__ . ":" . __LINE__ );
          }

          #
          # save current stats
          #
          Xormon::save_as_json( $last_stats_file, $stats );
        }
      }
    }

  }

  return \%stats_out;
}

sub make_gauge {
  my $switch_name = shift;
  my $metric      = shift;
  my $act_stats   = shift;
  my $last_stats  = shift;

  if ( exists $act_stats->{$metric} && Xormon::isdigit($act_stats->{$metric}) && exists $last_stats->{$metric} && Xormon::isdigit($last_stats->{$metric}) ) {

    if ( $act_stats->{$metric} == 0 ) {
      # it's useless to calculate difference if there is zero
      return 0;
    }
    elsif ( $act_stats->{$metric} < $last_stats->{$metric} ) {
      # counter overflow when value of the current counter is less than previous value
      return "";
    }
    elsif ( $act_stats->{$metric} == $last_stats->{$metric} ) {
      # counter didn't grouw up, values are the same
      return 0;
    }

    if ( exists $act_stats->{statistics_collection_time} && exists $last_stats->{statistics_collection_time} ) {
      my $interval = int($act_stats->{statistics_collection_time} / 1000) - int($last_stats->{statistics_collection_time} / 1000); # from ms to s

      if ( $interval <= ( 2 * $sample_rate ) && $interval > 0 ) {
        return sprintf( '%.2f', ($act_stats->{$metric} - $last_stats->{$metric}) / $interval );
      }
      elsif ( $interval == 0) {
        unless ( exists $errors{$switch_name} ) { # report it only once per agent run
          $logger->error( "$switch_name: Cannot convert COUNTER $metric to GAUGE. The interval is $interval seconds :" . __FILE__ . ":" . __LINE__ );
        }
        $errors{$switch_name} = 1;
      }
      else {
        unless ( exists $errors{$switch_name} ) { # report it only once per agent run
          $logger->error( "$switch_name: Cannot convert COUNTER $metric to GAUGE. The interval $interval is longer then ".2*$sample_rate." seconds :" . __FILE__ . ":" . __LINE__ );
        }
        $errors{$switch_name} = 1;
      }
    }
  }

  return "";
}

sub convert_collection_time {
  my $timestamp = shift;

  if ( $timestamp =~ m/^[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]$/ ) { # in seconds
    return int($timestamp / $sample_rate) * $sample_rate; # just to be sure, convert it to 5 minutes samples
  }
  elsif ( $timestamp =~ m/^[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]$/ ) { # in ms
    return int( int($timestamp / 1000) / $sample_rate) * $sample_rate; # just to be sure, convert it to 5 minutes samples
  }
  else {
    return undef;
  }
}

sub send_conf {
  my $switch_name = shift;

  unless ( $cfg_time ) { return 1; } # not this time

  my @architecture;
  my %conf;

  if ( $switch_name eq "sannavmgmt" ) {
    #
    # SANNAV mgmt
    #
    my %sannav_arch = (
      "item_id"    => $sannav_uuid,
      "hostcfg_id" => $sannav_uuid,
      "label"      => $sannav_name,
      "hw_type"    => "sannav",
      "subsystem"  => "mgmt",
      "parents"    => []
    );
    push(@architecture, \%sannav_arch);

    #
    # Fabrics
    #
    if ( exists $conf_data{fabrics} ) {
      foreach my $fabric_name ( sort keys %{ $conf_data{fabrics} } ) {
        if ( exists $conf_data{fabrics}{$fabric_name}{conf}{guid} ) {
          my %fabric_arch = (
            "item_id"    => $conf_data{fabrics}{$fabric_name}{conf}{guid},
            "label"      => $fabric_name,
            "hw_type"    => "sannav",
            "subsystem"  => "fabric",
            "parents"    => []
          );
          push(@architecture, \%fabric_arch);
        }
      }
    }
  }
  else {
    my $switch_var = $conf_data{switches}{$switch_name}{conf};
    my $switch_wwn = "";

    if ( exists $switch_var->{guid} ) {
      #
      # device (switch)
      #
      my %switch_arch = (
        "item_id"    => $switch_var->{guid},
        "hostcfg_id" => $sannav_uuid,
        "label"      => $switch_name,
        "hw_type"    => "sannav",
        "subsystem"  => "device",
        "parents"    => [$sannav_uuid]
      );
      if ( exists $switch_var->{fabricGuid} )                      { push(@{ $switch_arch{parents} }, $switch_var->{fabricGuid}); } # Fabric ID for grouping purposes
      if ( exists $switch_var->{wwn} && $switch_var->{wwn} ne '' ) {
        $conf{$switch_var->{guid}}{'wwn'} = Xormon::formatWWN($switch_var->{wwn});
        $switch_arch{wwn} = $conf{$switch_var->{guid}}{'wwn'};
        $switch_wwn       = $switch_var->{wwn};
      }
      push(@architecture, \%switch_arch);
      # conf
      if ( exists $switch_var->{fabricName} )         { $conf{$switch_var->{guid}}{'fabric_name'}    = $switch_var->{fabricName}; }
      if ( exists $switch_var->{model} )              { $conf{$switch_var->{guid}}{'model'}          = $switch_var->{model}; }
      if ( exists $switch_var->{serial} )             { $conf{$switch_var->{guid}}{'serial'}         = $switch_var->{serial}; }
      if ( exists $switch_var->{firmware} )           { $conf{$switch_var->{guid}}{'version'}        = $switch_var->{firmware}; }
      if ( exists $switch_var->{ipAddress} )          { $conf{$switch_var->{guid}}{'ip'}             = $switch_var->{ipAddress}; }
      if ( exists $switch_var->{physicalSwitchName} ) { $conf{$switch_var->{guid}}{'chassis_name'}   = $switch_var->{physicalSwitchName}; }
      if ( exists $switch_var->{upTime} )             { $conf{$switch_var->{guid}}{'days_up'}        = $switch_var->{upTime}; }
      if ( exists $switch_var->{status} )             { $conf{$switch_var->{guid}}{'status'}         = $switch_var->{status}; }
      if ( exists $switch_var->{health} )             { $conf{$switch_var->{guid}}{'health'}         = $switch_var->{health}; }
      if ( exists $switch_var->{state} )              { $conf{$switch_var->{guid}}{'state'}          = $switch_var->{state}; }
      if ( exists $switch_var->{reason} )             { $conf{$switch_var->{guid}}{'reason'}         = $switch_var->{reason}; }
      if ( exists $switch_var->{portCount} )          { $conf{$switch_var->{guid}}{'total_ports'}    = $switch_var->{portCount}; }
      if ( exists $switch_var->{licensedPortCount} )  { $conf{$switch_var->{guid}}{'licensed_ports'} = $switch_var->{licensedPortCount}; }
      if ( exists $switch_var->{switchType} ) {
        my $type_line = translate_brcd_type( $switch_var->{switchType} );
        if ( defined $type_line && $type_line ne '' ) {
          my ( undef, $speed, $model ) = split( ",", $type_line );
          $conf{$switch_var->{guid}}{'speed'} = $speed;
        }
      }
      if ( exists $switch_var->{chassisSerialNumber} && exists $conf_data{chassis}{ $switch_var->{chassisSerialNumber} }{conf}{name} ) {
        $conf{$switch_var->{guid}}{chassis_name} = $conf_data{chassis}{ $switch_var->{chassisSerialNumber} }{conf}{name};
      }
      if ( exists $switch_var->{virtualSwitchDetails}{domainId} ) {
        $conf{$switch_var->{guid}}{domain_id} = $switch_var->{virtualSwitchDetails}{domainId};
      }
      my $free_ports   = 0;
      my $avail_ports  = 0;
      my $unused_ports = 0;
      my $isl_ports    = 0;

      #
      # Ports
      #
      foreach my $port_id ( sort keys %{ $conf_data{switches}{$switch_name}{ports} } ) {
        my $port_var   = $conf_data{switches}{$switch_name}{ports}{$port_id};
        #print Dumper $port_var;
        my $port_name  = set_port_name($switch_name,$port_var);
        my $port_label = $port_name;
        my $port_wwn   = "";

        unless ( defined $port_name )       { next; }
        unless ( exists $port_var->{guid} ) { $logger->error("$switch_name:$port_id:$port_name Port guid was not found! Port skipped... $!" . __FILE__ . ":" . __LINE__ ) && next; }

        if ( exists $port_var->{name} && $port_var->{name} ne '' ) {
          $port_label .= " [$port_var->{name}]";
          $conf{$port_var->{guid}}{alias} = $port_var->{name};
        }

        # Architecture
        my %arch = (
          "item_id"    => $port_var->{guid},
          "label"      => $port_label,
          "hw_type"    => "sannav",
          "subsystem"  => "sanport",
          "hostcfg_id" => $sannav_uuid,
          "parents"    => [$switch_var->{guid}]
        );
        if ( exists $port_var->{wwn} && $port_var->{wwn} ne '' ) {
          $conf{$port_var->{guid}}{'wwn'} = Xormon::formatWWN($port_var->{wwn});
          $arch{wwn}                      = $conf{$port_var->{guid}}{'wwn'};
          $port_wwn                       = $port_var->{wwn};
        }
        if ( exists $port_var->{attachedPortWwn} && $port_var->{attachedPortWwn} ne '' ) {
          $conf{$port_var->{guid}}{'target_wwn'} = Xormon::formatWWN($port_var->{attachedPortWwn});
          push(@{ $arch{target_wwns} }, $conf{$port_var->{guid}}{'target_wwn'});
        }
        # conf
        if ( exists $port_var->{portLicense} ) { $conf{$port_var->{guid}}{license}            = $port_var->{portLicense}; }
        if ( exists $port_var->{speed} )       { $conf{$port_var->{guid}}{'speed'}            = $port_var->{speed}; }
        if ( exists $port_var->{state} )       { $conf{$port_var->{guid}}{operational_status} = $port_var->{state}; }
        if ( exists $port_var->{status} )      { $conf{$port_var->{guid}}{physical_status}    = $port_var->{status}; }
        if ( exists $port_var->{zoneAlias}    && $port_var->{zoneAlias} ne '' )    { $conf{$port_var->{guid}}{zone_alias}    = $port_var->{zoneAlias}; }
        if ( exists $port_var->{attachedPort} && $port_var->{attachedPort} ne '' ) { $conf{$port_var->{guid}}{'target_port'} = $port_var->{attachedPort}; }
        # switch ports
        if ( exists $conf{$port_var->{guid}}{operational_status} && exists $conf{$port_var->{guid}}{physical_status} ) {
          #if ( $conf{$port_var->{guid}}{physical_status} eq ""

          if ( $conf{$port_var->{guid}}{operational_status} eq "Offline" || $conf{$port_var->{guid}}{physical_status} eq "Disabled" ) {
            $unused_ports++;
            $conf{$port_var->{guid}}{'working_state'} = "unused";
          }
          if ( $conf{$port_var->{guid}}{operational_status} eq "Offline" && $conf{$port_var->{guid}}{physical_status} eq "No_Light" ) {
            $free_ports++;
            $avail_ports++;
          }
          if ( $conf{$port_var->{guid}}{physical_status} eq "In_Sync" ) {
            $avail_ports++;
          }
        }
        # host wwn, ISL
        if ( exists $port_var->{attachedPortWwn} ) {
          $conf{$port_var->{guid}}{host_wwn} = Xormon::formatWWN($port_var->{attachedPortWwn});
          push(@{ $arch{target_wwns} }, $conf{$port_var->{guid}}{host_wwn});

          if ( exists $port_var->{connectedDeviceType} && $port_var->{connectedDeviceType} eq "SWITCH" && exists $port_var->{connectedDevice} ) {
            my $target_switch_name = $port_var->{connectedDevice};
            my $target_port_wwn    = $port_var->{attachedPortWwn};

            if ( exists $conf_data{switches}{$target_switch_name} ) {
              my $target_switch_var = $conf_data{switches}{$target_switch_name};

              if ( exists $target_switch_var->{wwn} ) {
                my $target_switch_wwn = $target_switch_var->{wwn};

                if ( exists $ports_wwns{$target_port_wwn} ) {
                  my $target_port_idx = $ports_wwns{$target_port_wwn};

                  if ( exists $target_switch_var->{ports}{$target_port_idx} ) {
                    my $target_port_var  = $target_switch_var->{ports}{$target_port_idx};
                    my $target_port_name = set_port_name($target_switch_name,$target_port_var);

                    if ( defined $target_port_name ) {
                      $isl_map{$switch_name}{$port_name}{port_wwn}           = $port_wwn;
                      $isl_map{$switch_name}{$port_name}{switch_wwn}         = $switch_wwn;
                      $isl_map{$switch_name}{$port_name}{target_port_name}   = $target_port_name;
                      $isl_map{$switch_name}{$port_name}{target_port_wwn}    = $target_port_wwn;
                      $isl_map{$switch_name}{$port_name}{target_port_idx}    = $target_port_idx; # port is stored under this id in conf data ($conf_data{switches}{$target_switch_name}{ports}{$target_port_idx})
                      $isl_map{$switch_name}{$port_name}{target_switch_name} = $target_switch_name;
                      $isl_map{$switch_name}{$port_name}{target_switch_wwn}  = $target_switch_wwn;

                      $conf{$port_var->{guid}}{isl} = "$target_switch_name, $target_port_name";
                    }
                  }
                }
              }
            }
          }
        }
        push(@architecture, \%arch);
      }
      if ( exists $isl_map{$switch_name} ) { $isl_ports = keys %{ $isl_map{$switch_name} }; }
      $conf{$switch_var->{guid}}{'free_ports'}      = $free_ports;
      $conf{$switch_var->{guid}}{'available_ports'} = $avail_ports;
      $conf{$switch_var->{guid}}{'unused_ports'}    = $unused_ports;
      $conf{$switch_var->{guid}}{'isl_ports'}       = $isl_ports;
    }
  }

  #print "#################### Architecture\n";
  #print Dumper @architecture;
  #print "#################### CONF\n";
  #print Dumper \%conf;

  #save architecture
  my $rc_arch = $xormon->saveArchitecture(\@architecture);
  $logger->info("$switch_name: save architecture: $rc_arch");

  #save configuration
  my $rc_conf = $xormon->saveConfJSON(\%conf);
  $logger->info("$switch_name: save conf: $rc_conf");

  return 1;
}

sub get_conf_data {
  #
  # switches cfg
  #
  get_switches_conf();

  #
  # chassis cfg
  #
  get_chassis_conf();

  #
  # fabrics cfg
  #
  get_fabrics_conf();

  #
  # only once per hour
  #
  #if ( $cfg_time ) {
  #  #
  #  # fabrics cfg
  #  #
  #  get_fabrics_conf();
  #}
  #NOTE: it's not necessary to collect data only once per hour, it's only abbout one REST API request...

  return 1;
}

sub get_switches_conf {

  my $result = new_post_request("inventory/switches",{});

  if ( defined $result && ref($result) eq "ARRAY" ) {
    foreach my $var ( @{ $result } ) {
      #print Dumper $var;
      if ( exists $var->{name} ) {
        my $switch_name = $var->{name};
        $conf_data{switches}{$switch_name}{conf} = $var;

        if ( exists $var->{guid} && exists $var->{fabricGuid} ) {
          #GET https://10.22.111.20/inventory/products/?switchGuid=ee3cf260-367f-4548-88f7-d81c71b42669&fabricGuid=bd49d7db-acce-4f97-af61-bafeef2082c9
          my $sw_result = new_get_request("inventory/products/?switchGuid=$var->{guid}&fabricGuid=$var->{fabricGuid}");

          if ( defined $sw_result && ref($sw_result) eq "HASH" ) {
            foreach my $key ( sort keys %{ $sw_result } ) {

              if ( $key eq "ports" && ref($sw_result->{$key}) eq "ARRAY" ) {
                foreach my $port_var ( @{ $sw_result->{$key} } ) {
                  if ( exists $port_var->{portIndex} ) {
                    $conf_data{switches}{$switch_name}{ports}{ $port_var->{portIndex} } = $port_var;

                    if ( exists $port_var->{wwn} ) {
                      $ports_wwns{ $port_var->{wwn} } = $port_var->{portIndex}; # it's necessary to be able find port var by wwn
                    }
                  }
                }
              }
              else {
                unless ( exists $conf_data{switches}{$switch_name}{conf}{$key} ) {
                  # add additional info to switch configuratio (e.g. upTime is missing in the main conf "inventory/switches")
                  $conf_data{switches}{$switch_name}{conf}{$key} = $sw_result->{$key};
                }
              }
            }
          }
        }
      }
    }
  }

  return 1;
}

sub get_chassis_conf {
  my $result = new_post_request("inventory/chassis",{});

  if ( defined $result && ref($result) eq "ARRAY" ) {
    foreach my $var ( @{ $result } ) {
      #print Dumper $var;
      if ( exists $var->{chassisSerialNumber} ) {
        $conf_data{chassis}{ $var->{chassisSerialNumber} }{conf} = $var;
      }
    }
  }

  return 1;
}

sub get_fabrics_conf {
  my $result = new_post_request("inventory/fabrics",{});
  #print Dumper $result;

  if ( defined $result && ref($result) eq "ARRAY" ) {
    foreach my $var ( @{ $result } ) {
      #print Dumper $var;
      if ( exists $var->{name} ) {
        $conf_data{fabrics}{ $var->{name} }{conf} = $var;
      }
    }
  }

  return 1;
}

sub get_stats {
  my $stat_idx = 0;

  foreach my $line ( <$kafkadir/$kafka_ip-*.json> ) {
    chomp $line;

    my $file = $line;

    # /home/stor2rrd/stor2rrd/tmp/SANnav-kafka-perf/10.22.111.26-1654591500277-perfmon_metrics_switch-10:00:00:05:1E:DD:F0:EB.json
    $line =~ s/^$kafkadir\/$kafka_ip-//;
    $line =~ s/\.json$//;

    my ( $timestamp, $type, $wwn ) = split(/-/, $line);
    #$timestamp = int($timestamp / 1000); # from ms to s

    if ( int($timestamp / 1000) < ( int($act_time / $sample_rate) * $sample_rate ) ) {
      # replace NaN characters provided by python to have valid JSON
      # ERROR: malformed number (no digits after initial minus), at character offset 1070 (before "nfinity, "media_tx_p...")
      # --> "media_rx_power": -Infinity, "media_tx_power": -Infinity
      make_valid_json($file);

      my $stats = {};
      eval { $stats = Xormon::read_json($file); };

      #print Dumper $stats;
      if ( exists $stats->{statistics_collection_time} ) {
        #print "$timestamp : $stats->{statistics_collection_time}\n";
        if ( $stats->{statistics_collection_time} =~ m/^[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]$/ ) { # in seconds
          my $collection_timestamp = int($stats->{statistics_collection_time} / $sample_rate) * $sample_rate; # just to be sure, convert it to 5 minutes samples

          $perf_data{$collection_timestamp}{$wwn} = $stats;
        }
        elsif ( $stats->{statistics_collection_time} =~ m/^[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]$/ ) { # in ms
          my $collection_timestamp = int($stats->{statistics_collection_time} / 1000);         # from ms to s
          $collection_timestamp    = int($collection_timestamp / $sample_rate) * $sample_rate; # just to be sure, convert it to 5 minutes samples

          $perf_data{$collection_timestamp}{$wwn} = $stats;
        }
        else {
          $logger->error("Unsupported statistics_collection_time! Stats skippe... statistics_collection_time=\"$stats->{statistics_collection_time}\", file=$file $!" . __FILE__ . ":" . __LINE__ ) && next;
        }
      }

      unlink($file);
      $stat_idx++;
    }
  }

  if ( $stat_idx == 0 ) {
    $logger->error("Statistics have not been found! Exiting... $!" . __FILE__ . ":" . __LINE__ );
    return 0;
  }

  return 1;
}

sub make_valid_json {
  my $file = shift;

  open( my $f_in, "<$file" ) || $logger->error( "Couldn't open file $file $!" . __FILE__ . ":" . __LINE__ ) && exit;
  my $string = <$f_in>;
  close($f_in);

  $string =~ s/-Infinity/"-Infinity"/g;

  open( my $f_out, ">$file" ) || $logger->error( "Couldn't open file $file $!" . __FILE__ . ":" . __LINE__ ) && exit;
  print $f_out $string;
  close($f_out);

  return 1;
}

sub new_get_request {
  my $namespace = shift;

  my $url     = "https://$cfg->{'apiHost'}:$cfg->{'apiPort'}/$namespace";
  my $agent   = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );
  my $request = HTTP::Request->new( GET => $url );

  if ( defined $sessionId && $sessionId ne '' ) {
    $request->header(
      'Content-Type'  => 'application/json',
      'Accept'        => 'application/json',
      'Authorization' => $sessionId
    );
  }
  else {
    $request->header(
      'Content-Type' => 'application/json',
      'Accept'       => 'application/json',
      'username'     => $cfg->{'apiUsername'},
      'password'     => $cfg->{'apiPassword'}
    );
  }

  my $result = $agent->request($request);

  if ( !$result->is_success ) {
    my $st_line = $result->status_line;
    my $res_con = $result->content;
    $logger->error("GET url      : $url");
    $logger->error("Request error: $st_line");
    $logger->error("Request error: $res_con");

    if ( $connection_test ) {
      push(@rows, "POST url     : $url\n");
      push(@rows, "Request error: $st_line\n");
      push(@rows, "Request error: $res_con\n");
    }

    return undef;
  }

  #debug_print( "GET url      : ".$url_act );
  #debug_print( "status_line  : ".$result->status_line );

  eval { $result = decode_json( $result->content ); };

  #if ($debug) {
  #  my $filename = $url_act;
  #  $filename    =~ s/^$url\///;
  #  $filename    =~ s/\//_/g;
  #  Xorux_lib::save_as_json("$tmpdir/SANnav-debug/$filename.json",$result);
  #  debug_print( "Create file  : $tmpdir/SANnav-debug/$filename.json" );
  #}

  return $result;
}

sub new_post_request {
  my $namespace = shift;
  my $body      = shift;

  my $url     = "https://$cfg->{'apiHost'}:$cfg->{'apiPort'}/$namespace";
  my $agent   = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );
  my $request = HTTP::Request->new( POST => $url );

  if ( defined $sessionId && $sessionId ne '' ) {
    $request->header(
      'Content-Type'  => 'application/json',
      'Accept'        => 'application/json',
      'Authorization' => $sessionId
    );
  }
  else {
    $request->header(
      'Content-Type' => 'application/json',
      'Accept'       => 'application/json',
      'username'     => $cfg->{'apiUsername'},
      'password'     => $cfg->{'apiPassword'}
    );
  }

  if ( defined $body && $body ne '' ) {
    $request->content( encode_json($body) );
  }

  my $result = $agent->request($request);

  if ( !$result->is_success ) {
    my $st_line = $result->status_line;
    my $res_con = $result->content;
    $logger->error("POST url     : $url");
    $logger->error("Request error: $st_line");
    $logger->error("Request error: $res_con");
    if ( defined $body ) {
      $logger->error("JSON body    : ".encode_json($body));
    }

    if ( $connection_test ) {
      push(@rows, "POST url     : $url\n");
      push(@rows, "Request error: $st_line\n");
      push(@rows, "Request error: $res_con\n");
      if ( defined $body ) {
        push(@rows, "JSON body    : ".encode_json($body)."\n");
      }
    }

    return undef;
  }

  #debug_print( "POST url     : ".$url_act );
  #debug_print( "status_line  : ".$result->status_line );

  eval { $result = decode_json( $result->content ); };

  #if ($debug) {
  #  my $filename = $url_act;
  #  $filename    =~ s/^$url\///;
  #  $filename    =~ s/\//_/g;
  #  Xorux_lib::save_as_json("$tmpdir/SANnav-debug/$filename.json",$result);
  #  debug_print( "Create file  : $tmpdir/SANnav-debug/$filename.json" );
  #}

  if ( defined $result && ref($result) eq "HASH" && exists $result->{'sessionId'} ) {
    $sessionId = $result->{'sessionId'};
  }
  #print Dumper $result;

  return $result;
}

sub set_port_name {
  my $switch_name = shift;
  my $port_var    = shift;

  unless ( exists $port_var->{port} ) {
    $logger->error( "$switch_name: Somethink wrong, 'port' doesn't exist! $!" . __FILE__ . ":" . __LINE__ );
    $logger->error( Dumper $port_var );
    return undef;
  }

  if ( is_switch_director($switch_name) ) {
    if ( exists $port_var->{slot} ) {
      return $port_var->{slot}."-".$port_var->{port};
    }
    else {
      $logger->error( "$switch_name: Somethink wrong, slot of the port $port_var->{port} doesn't exists! $!" . __FILE__ . ":" . __LINE__ ) && return undef;
    }
  }
  else {
    return $port_var->{port};
  }
}

sub is_switch_director {
  my $switch_name = shift;

  if ( exists $directors{$switch_name} ) { # already discovered, don't do it twice
    return 1; # director
  }

  foreach my $port_id ( sort { $a <=> $b } keys %{ $conf_data{switches}{$switch_name}{ports} } ) {
    my $port_var = $conf_data{switches}{$switch_name}{ports}{$port_id};

    if ( exists $port_var->{slot} && Xormon::isdigit($port_var->{slot}) && $port_var->{slot} > 0 ) {
      $directors{$switch_name} = 1;
      return 1; # director
    }
  }

  return 0; # not
}

sub run_kafka_consumer {
  $cfg      = shift;
  $wrkdir   = shift;
  $kafkadir = shift;
  $act_dir  = shift;

  #
  # mandatory cfg
  #
  unless ( exists $cfg->{'hostcfg_id'} && $cfg->{'hostcfg_id'} ne '' ) { Xormon::error("Cannot continue, 'hostcfg_id' is not configured!") && exit 0; }
  unless ( exists $cfg->{'hostalias'}  && $cfg->{'hostalias'} ne '' )  { Xormon::error("Cannot continue, 'hostalias' is not configured!")  && exit 0; }
  $logger = XormonLogger->new( 'sannav', $cfg->{hostcfg_id}, $cfg->{hostalias} );
  unless ( exists $cfg->{'kafkaHost'}    && $cfg->{'kafkaHost'} ne '' )    { $logger->error("Cannot continue, 'kafkaHost' is not configured!")    && exit 0; }
  unless ( exists $cfg->{'registryPort'} && $cfg->{'registryPort'} ne '' ) { $logger->error("Cannot continue, 'registryPort' is not configured!") && exit 0; }
  unless ( exists $cfg->{'brokerPort'}   && $cfg->{'brokerPort'} ne '' )   { $logger->error("Cannot continue, 'brokerPort' is not configured!")   && exit 0; }

  #
  # Kafka consumer cfg
  #
  $ENV{'KAFKA_IP'}          = $cfg->{'kafkaHost'};
  $ENV{'KAFKA_SCHEMA_PORT'} = $cfg->{'registryPort'};
  $ENV{'KAFKA_BROKER_PORT'} = $cfg->{'brokerPort'};
  $ENV{'KAFKA_DIR'}         = $kafkadir;
  if ( $cfg->{kafka_debug} && Xormon::isdigit($cfg->{kafka_debug}) ) {
    $ENV{'KAFKA_DEBUG'} = $cfg->{kafka_debug};
    $logger->info( "KAFKA_DEBUG=$cfg->{kafka_debug}" );
  }

  #
  # check if the running Kafka configuration is up to date, restart consumer if not
  #
  my $last_kafka_cfg = "$kafkadir/last_kafka_cfg.json";
  if ( -f $last_kafka_cfg ) {
    my $last_cfg = Xormon::read_json($last_kafka_cfg);

    if ( exists $last_cfg->{'KAFKA_IP'} && exists $last_cfg->{'KAFKA_SCHEMA_PORT'} && exists $last_cfg->{'KAFKA_BROKER_PORT'} ) {
      unless ( $cfg->{'kafkaHost'} eq $last_cfg->{'KAFKA_IP'} && $cfg->{'registryPort'} eq $last_cfg->{'KAFKA_SCHEMA_PORT'} && $cfg->{'brokerPort'} eq $last_cfg->{'KAFKA_BROKER_PORT'} ) {
        # the configuration has changed
        stop_consumer("$wrkdir/sannavperf-run","sannavperf.py $last_cfg->{'KAFKA_IP'} $cfg->{'hostalias'} XormonNG"); # stop daemon for the old IP, if exists
        stop_consumer("$wrkdir/sannavperf-run","sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG");     # stop daemon for the current configuration, if exists
      }
    }
  }
  # save current configuration
  Xormon::save_as_json($last_kafka_cfg,{ 'KAFKA_IP' => $cfg->{'kafkaHost'}, 'KAFKA_SCHEMA_PORT' => $cfg->{'registryPort'}, 'KAFKA_BROKER_PORT' => $cfg->{'brokerPort'} });

  #
  # run kafka consumer as a daemon
  #
  if ( is_process_running("sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG") ) {
    # restart sannavperf.py process if it is the first run after midnight
    restart_consumer("$wrkdir/sannavperf-run", "sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG");
  }
  unless ( is_process_running("sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG") ) {
    $logger->info( "starting a background task: $act_dir/sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG &" );
    system "$PYTHON $act_dir/sannavperf.py $cfg->{'kafkaHost'} $cfg->{'hostalias'} XormonNG &";
    system ("touch", "$wrkdir/sannavperf-run");
  }

  return 1;
}

sub restart_consumer {
  my $file           = shift;
  my $process_string = shift;

  if ( -f $file ) {
    if ( ( localtime( ( stat($file) )[9] ) )[3] != ( localtime() )[3] ) {
      # day of the last modification --> ( localtime( ( stat($file) )[9] ) )[3]
      # today                        --> ( localtime() )[3] )
      #print "############# ".( localtime( ( stat($file) )[9] ) )[3]." : ".( localtime() )[3]."\n";

      $logger->info( "First run after midnight, restart old '$process_string' process:" );
      my @pids = ();
      foreach ( grep {/$process_string/} @{ get_ps_list() } ) {
        chomp $_;
        my ( undef, $pid, undef ) = split(/\s+/, $_);
        push(@pids, $pid);
        print "$_\n";
      }
      if ( scalar(@pids) > 0 ) {
        kill('TERM',@pids);
        $logger->info( "kill 'TERM', " . join(" ", @pids) );
      }
      else {
        $logger->info( "Old $process_string process has not been found..." );
      }
    }
  }

  return 1;
}

sub stop_consumer {
  my $file           = shift;
  my $process_string = shift;

  my @pids = ();
  foreach ( grep {/$process_string/} @{ get_ps_list() } ) {
    chomp $_;
    my ( undef, $pid, undef ) = split(/\s+/, $_);
    push(@pids, $pid);
    print "$_\n";
  }
  if ( scalar(@pids) > 0 ) {
    kill('TERM',@pids);
    $logger->info( "kill 'TERM', " . join(" ", @pids) );
  }
  else {
    $logger->info( "Old sannavperf.py process has not been found..." );
  }

  return 1;
}

sub terminate_old_processes {
  Xormon::log( "Terminate old sannavperf.py processes:" );
  my @pids = ();
  foreach ( @{ get_ps_list() } ) {
    if ( $_ =~ m/sannavperf\.py.*XormonNG/ ) {
      my ( undef, $pid, undef ) = split(/\s+/, $_);
      push(@pids, $pid);
      print "$_\n";
    }
  }
  if ( scalar(@pids) > 0 ) {
    kill('TERM',@pids);
    Xormon::log( "kill 'TERM', " . join(" ", @pids) );
  }
  else {
    Xormon::log( "Old sannavperf.py processes have not been found..." );
  }

  return 1;
}

sub is_process_running {
  my $process_string = shift;

  if ( ( grep {/$process_string/} @{ get_ps_list() } ) > 0 ) {
    return 1;
  }
  else {
    return 0;
  }
}

sub get_ps_list {
  open(my $ps,"ps -ef |") || Xormon::error("'ps -ef' command failed! $!" . __FILE__ . ":" . __LINE__ ) && exit 0;
  my @ps_list = <$ps>;
  chomp @ps_list;
  close($ps);

  return \@ps_list;
}

sub translate_brcd_type {
  my $type      = shift;
  my @all_types = (
    "2,1 Gbps,Brocade 2010",
    "3,1 Gbps,Brocade 2400",
    "6,1 Gbps,Brocade 2800",
    "9,2 Gbps,Brocade 3800",
    "10,2 Gbps,Brocade 12000",
    "12,2 Gbps,Brocade 3900",
    "16,2 Gbps,Brocade 3200",
    "21,2 Gbps,Brocade 24000",
    "22,2 Gbps,Brocade 3016",
    "26,2 Gbps,Brocade 3850",
    "27,2 Gbps,Brocade 3250",
    "29,2 Gbps,Brocade 4012",
    "32,4 Gbps,Brocade 4100",
    "34,2 Gbps,Brocade 200E",
    "37,2 Gbps,Brocade 4020",
    "38,2 Gbps,Brocade AP7420",
    "42,4 Gbps,Brocade 48000",
    "43,4 Gbps,Brocade 4024",
    "44,4 Gbps,Brocade 4900",
    "45,2 Gbps,Brocade 4016",
    "46,4 Gbps,Brocade 7500",
    "50,UN,Brocade CP8",
    "51,2 Gbps,Brocade 4018",
    "58,4 Gbps,Brocade 5000",
    "61,2 Gbps,Brocade 4424",
    "62,8 Gbps,Brocade DCX",
    "64,8 Gbps,Brocade 5300",
    "66,8 Gbps,Brocade 5100",
    "67,8 Gbps,Brocade Encryption Switch",
    "70,8 Gbps,Brocade 5410",
    "71,8 Gbps,Brocade 300",
    "72,8 Gbps,Brocade 5480",
    "73,8 Gbps,Brocade 5470",
    "75,8 Gbps,Brocade M5424",
    "76,UN,Brocade 8000",
    "77,8 Gbps,Brocade DCX-4S",
    "83,8 Gbps,Brocade 7800",
    "86,UN,Brocade 450",
    "87,UN,Brocade 460",
    "92,UN,Brocade A-40FC",
    "96,16 Gbps,Brocade FC16-48",
    "97,16 Gbps,Brocade FC16-32",
    "98,16 Gbps,Brocade CR16-8",
    "99,16 Gbps,Brocade CR16-4",
    "109,16 Gbps,Brocade 6510",
    "117,16 Gbps,Brocade 6547",
    "118,16 Gbps,Brocade 6505",
    "120,16 Gbps,Brocade DCX8510-8",
    "121,16 Gbps,Brocade DCX8510-4",
    "124,UN,Brocade 430",
    "125,8 Gbps,Brocade 5431",
    "129,16 Gbps,Brocade 6548",
    "130,16 Gbps,Brocade M6505",
    "133,16 Gbps,Brocade 6520",
    "134,8 Gbps,Brocade 5432",
    "148,16 Gbps,Brocade 7840",
    "149,16 Gbps,Brocade 6546",
    "150,16 Gbps,Brocade 6545",
    "152,16 Gbps,Brocade 6549",
    "153,16 Gbps,Brocade FC16-64",
    "156,16 Gbps,Brocade 6543",
    "157,16 Gbps,Brocade 6558",
    "158,16 Gbps,Brocade 6559",
    "162,32 Gbps,Brocade G620",
    "165,32 Gbps,Brocade X6-4",
    "166,32 Gbps,Brocade X6-8",
    "167,16 Gbps,Brocade 6542",
    "170,32 Gbps,Brocade G610",
    "173,128 Gbps,Brocade G630",
    "175,UN,Brocade CPX6",
    "176,32 Gbps,Brocade CR32-4",
    "177,32 Gbps,Brocade CR32-8",
    "178,32 Gbps,Brocade FC32-48",
    "179,64 Gbps,Brocade X7-4",
    "180,64 Gbps,Brocade X7-8",
    "181,128 Gbps,Brocade G720",
    "183,128 Gbps,Brocade G620",
    "184,128 Gbps,Brocade G630",
    "186,16 Gbps,Brocade SX6",
  );

  # 178 has already been used
  # 178 Brocade 7810 Extension Switch Extension switch with configurable scaling of 12 ports and supports 32G, 16G, 8G, and 4G media types. It supports 1GbE and 10GbE Ethernet interfaces.

  $type =~ s/\.[0-9]*$//;
  my $type_line = "";
  ($type_line) = grep {/^$type,/} @all_types;

  return ($type_line);
}

1;
