#
# FlashSystem.pm
#
# $Revision: 2.2.3 $ 2022.06.18

# Changes:
# Release 1.2 FlashSystem 840/900 outputs (_svcflashperf_).
# Release 1.3 Reduce ssh/scp login sessions
# Release 2.1 Implement Net::OpenSSH

package FlashSystem;    #
our $VERSION = 2.2.3;

# Modules
use strict;
use Data::Dumper;
use Date::Parse;
#
my $debug = 0;
my $cmd_ls   = "/bin/ls";
my $cmd_mv   = "/bin/mv";
my $cmd_rm   = "/bin/rm";
my $cmd_grep = "/bin/grep";
my $cmd_mkdir = "/bin/mkdir";
my $cmd_retry_max = 5;

# FCP Interface ('1','#0:04'), ('2','#0:08'), ('3','#1:0c'), ('4','#1:10') - Port a,b,c,d
# canister_id,adapter_id
my $interface2filename = {
  "1"=>{ "1"=>"interface825#0:04","2"=>"interface825#0:08" },
  "2"=>{ "1"=>"interface825#1:0c","2"=>"interface825#1:10" }
};

# canister_id,adapter_id,port_id
my $port2filename = {
  "1"=>{
    "1"=>{ "1"=>"if825_chan#0:04.a","2"=>"if825_chan#0:04.b","3"=>"if825_chan#0:04.c","4"=>"if825_chan#0:04.d" },
    "2"=>{ "1"=>"if825_chan#0:08.a","2"=>"if825_chan#0:08.b","3"=>"if825_chan#0:08.c","4"=>"if825_chan#0:08.d" }
  },
  "2"=>{
    "1"=>{ "1"=>"if825_chan#1:0c.a","2"=>"if825_chan#1:0c.b","3"=>"if825_chan#1:0c.c","4"=>"if825_chan#1:0c.d" },
    "2"=>{ "1"=>"if825_chan#1:10.a","2"=>"if825_chan#1:10.b","3"=>"if825_chan#1:10.c","4"=>"if825_chan#1:10.d" }
  }
};

my @perfstats = qw(ReadIOPS WriteIOPS IOPS ReadBW WriteBW BW ReadMaxLatency WriteMaxLatency);

# Return codes
use constant SVCPRF201E => 101;
use constant SVCPRF202E => 102;
# -------------------------------------------------------------------------------------------------

sub StatsSystemFilenames {
  my @a = (
  "iops","iops_read","iops_write",
  "bw","bw_read","bw_write",
  "rd_latency_avg","wr_latency_avg"
  );
  return("system","dumps/tejasstats/","current_short/system/",\@a);
}

sub StatsCpuFilenames {
  my @a = (
  "system_cpu","user_cpu","nice_cpu","idle_cpu",
  );
  return("cpu","dumps/tejasstats/","current_short/system/",\@a);
}

sub StatsFcInterfaceFilenames {
  my $cid = shift;
  my $aid = shift;
  my @a = (
  "iops","bw",
  "rd_latency_avg","wr_latency_avg",
  );
  # canister_id,adapter_id
  return("fcinterface","dumps/tejasstats/","current_short/".$interface2filename->{$cid}->{$aid}."/",\@a);
}

sub StatsFcPortFilenames {
  my $cid = shift;
  my $aid = shift;
  my $pid = shift;
  my @a = (
  "iops","iops_read","iops_write",
  "bw","bw_read","bw_write",
  "crc_errors",
  );
  # canister_id,adapter_id,port_id
  return("fcport","dumps/tejasstats/","current_short/".$port2filename->{$cid}->{$aid}->{$pid}."/",\@a);
}

sub Last_timestamps {
  my $d = shift;
  my $file = shift;
  my $cfg = shift;
  $debug = shift;
  my $fh;
  my $time    = localtime();
  if (defined $d->{timestamp}->{end}) {
    if($debug) { &message("Set last timestamp from previous run."); }
    if ( $d->{timestamp}->{'max'} - $d->{timestamp}->{'end'} > 2 * 3600 ) {
      $d->{timestamp}->{'end'} = $d->{timestamp}->{'max'} - 2 * 3600;
    }
    if ( $d->{timestamp}->{'max'} <= $d->{timestamp}->{'end'} ) {
      $d->{timestamp}->{'max'} = $d->{timestamp}->{'end'} + 300;
    }
    return($d->{timestamp});
  }
  if ( -r $file ) {
    if($debug) { &message("Read last timestamp from file: ".$file); }
    open($fh,'<',$file);
    while(my $line = <$fh>) {
      chomp($line);
      $d->{timestamp}->{'max'} = str2time($line);
      $d->{timestamp}->{'end'} = $d->{timestamp}->{'max'} - 6 * 300;
      return($d->{timestamp});
    }
    close($fh);
  } else {
    if($debug) { &message("Set last timestamp from current time."); }
    $d->{timestamp}->{'max'} = str2time($time) - 300;
    $d->{timestamp}->{'end'} = $d->{timestamp}->{'max'} - 6 * 300;
    return($d->{timestamp});
  }
}

sub Print_timestamps {
  my $h = shift;
  my @a;
  foreach my $k ( sort(keys(%{$h})) ) {
    push(@a,$k."(".$h->{$k}.")=".epoch2datestr($h->{$k}));
  }
  return(join(",",@a));
}

sub Upload_iostats_files_openssh {
  $debug = shift;
  my $net_ssh = shift;
  my $ssh_tmo = shift;
  my $iosdir = shift;
  my $cfg = shift;
  #
  my ($tmkey,$dir1,$dir2,$files);
  ($tmkey,$dir1,$dir2,$files) = &StatsSystemFilenames();
  Upload_files_openssh($net_ssh, $ssh_tmo, $iosdir, $dir1, $dir2, $files);
  ($tmkey,$dir1,$dir2,$files) = &StatsCpuFilenames();
  Upload_files_openssh($net_ssh, $ssh_tmo, $iosdir, $dir1, $dir2, $files);
  foreach my $int ( sort(keys(%{$cfg->{svc_interfacefc}})) ) {
    my $cid = $cfg->{svc_interfacefc}->{$int}->{canister_id};
    my $aid = $cfg->{svc_interfacefc}->{$int}->{adapter_id};
    ($tmkey,$dir1,$dir2,$files) = &StatsFcInterfaceFilenames($cid,$aid);
    Upload_files_openssh($net_ssh, $ssh_tmo, $iosdir, $dir1, $dir2, $files);
  }
  foreach my $port ( sort(keys(%{$cfg->{svc_portfc}})) ) {
    my $cid = $cfg->{svc_portfc}->{$port}->{canister_id};
    my $aid = $cfg->{svc_portfc}->{$port}->{adapter_id};
    my $pid = $cfg->{svc_portfc}->{$port}->{port_id};
    ($tmkey,$dir1,$dir2,$files) = &StatsFcPortFilenames($cid,$aid,$pid);
    Upload_files_openssh($net_ssh,$ssh_tmo,$iosdir,$dir1,$dir2,$files);
  }
} ## end sub Upload_iostats_files_openssh

sub Upload_files_openssh {
  my $ssh = shift;
  my $tmo = shift;
  my $iosdir = shift;
  my $dir1 = shift;
  my $dir2 = shift;
  my $files = shift;
  #
  if ( not -d $iosdir.$dir2 ) {
    runcmd( "$cmd_mkdir -p " . $iosdir.$dir2 );
  }
  # Remove local files
  my %upload;
  if ($debug) { &message( "Remove old files - dir: ".$dir2 ); }
  foreach my $file ( @{$files} ) {
    $upload{$file} = 1;
    runcmd( "$cmd_rm -f " . $iosdir.$dir2.$file ) if ( -r $iosdir.$dir2.$file );
  }
  # Retry uploud if file is not uploaded
  my $cmd_retry_count = $cmd_retry_max;
  while ( $cmd_retry_count ) {
    $cmd_retry_count--;
    # Create command
    my $download_files = [];
    foreach my $file ( keys(%upload) ) {
      if ( $upload{$file} ) {
        push( @{$download_files}, "/".$dir1.$dir2.$file );
      }
    }
    # Upload files
    if ($debug) { message( "Download files: ".join(" ", @{$download_files}).", in to directory ".$iosdir.$dir2 ); }
    my $out;
    $out = SVC::openssh_getfile($ssh, $download_files, $iosdir.$dir2, $tmo);
    if( not defined $out ) {
      &error("Download files: ".join(" ", @{$download_files}).", failed", 1);
      return undef;
    }
    # Check uploaded files
    my $cmd = $cmd_ls." ".$iosdir.$dir2;
    foreach my $line_LS (`$cmd`) {
      chomp($line_LS);
      $upload{$line_LS} = 0;
    }
    # Is all files uploaded
    my $missing = 0;
    foreach my $file ( @{$files} ) {
      if( $upload{$file} == 1 ) {
        $missing++;
      }
    }
    # All files downloaded
    if ( $missing == 0 ) { last; }
  } ## end while ( $cmd_retry_count )

  # Collect uploaded/missing files
  my @uploaded;
  my @missing;
  foreach my $file ( @{$files} ) {
    if( $upload{$file} ) {
      push(@missing,$dir2.$file);
    } else {
      push(@uploaded,$dir2.$file);
    }
  }
  # Files missing - exit
  if ( scalar @missing ) {
    foreach my $file ( @missing ) {
      warning("SVCPRF201E Upload IOSTAT file ".$file." failed.",SVCPRF201E);
    }
  }
  if ( scalar @uploaded ) {
    return(\@uploaded)
  }
}

sub Upload_iostats_files_cmdssh {
  $debug = shift;
  my $cmd_scp = shift;
  my $ssh_tmo = shift;
  my $iosdir = shift;
  my $cfg = shift;
  #
  my ($tmkey,$dir1,$dir2,$files);
  ($tmkey,$dir1,$dir2,$files) = &StatsSystemFilenames();
  Upload_files_cmdssh($cmd_scp,$ssh_tmo,$iosdir,$dir1,$dir2,$files);
  ($tmkey,$dir1,$dir2,$files) = &StatsCpuFilenames();
  Upload_files_cmdssh($cmd_scp,$ssh_tmo,$iosdir,$dir1,$dir2,$files);
  foreach my $int ( sort(keys(%{$cfg->{svc_interfacefc}})) ) {
    my $cid = $cfg->{svc_interfacefc}->{$int}->{canister_id};
    my $aid = $cfg->{svc_interfacefc}->{$int}->{adapter_id};
    ($tmkey,$dir1,$dir2,$files) = &StatsFcInterfaceFilenames($cid,$aid);
    Upload_files_cmdssh($cmd_scp,$ssh_tmo,$iosdir,$dir1,$dir2,$files);
  }
  foreach my $port ( sort(keys(%{$cfg->{svc_portfc}})) ) {
    my $cid = $cfg->{svc_portfc}->{$port}->{canister_id};
    my $aid = $cfg->{svc_portfc}->{$port}->{adapter_id};
    my $pid = $cfg->{svc_portfc}->{$port}->{port_id};
    ($tmkey,$dir1,$dir2,$files) = &StatsFcPortFilenames($cid,$aid,$pid);
    Upload_files_cmdssh($cmd_scp,$ssh_tmo,$iosdir,$dir1,$dir2,$files);
  }
} ## end sub Upload_iostats_files_cmdssh

sub Upload_files_cmdssh {
  my $scp = shift;
  my $tmo = shift;
  my $iosdir = shift;
  my $dir1 = shift;
  my $dir2 = shift;
  my $files = shift;
  #
  if ( not -d $iosdir.$dir2 ) {
    runcmd( "$cmd_mkdir -p " . $iosdir.$dir2 );
  }
  # Remove local files
  my %upload;
  if ($debug) { &message( "Remove old files - dir: ".$dir2 ); }
  foreach my $file ( @{$files} ) {
    $upload{$file} = 1;
    runcmd( "$cmd_rm -f " . $iosdir.$dir2.$file ) if ( -r $iosdir.$dir2.$file );
  }
  # Retry uploud if file is not uploaded
  my $cmd_retry_count = $cmd_retry_max;
  while ( $cmd_retry_count ) {
    $cmd_retry_count--;
    # Create command
    my $download_files = [];
    foreach my $file ( keys(%upload) ) {
      if ( $upload{$file} ) {
        push( @{$download_files}, "/".$dir1.$dir2.$file );
      }
    }
    # Upload files
    if ($debug) { message( "Download files: ".join(" ", @{$download_files}).", in to directory ".$iosdir.$dir2 ); }
    my $out;
    $out = SVC::cmdssh_getfile($scp, $download_files, $iosdir.$dir2, $tmo);
    if( not defined $out ) {
      &error("Download files: ".join(" ", @{$download_files}).", failed", 1);
      return undef;
    }
    # Check uploaded files
    my $cmd = $cmd_ls." ".$iosdir.$dir2;
    foreach my $line_LS (`$cmd`) {
      chomp($line_LS);
      $upload{$line_LS} = 0;
    }
    # Is all files uploaded
    my $missing = 0;
    foreach my $file ( @{$files} ) {
      if( $upload{$file} == 1 ) {
        $missing++;
      }
    }
    # All files downloaded
    if ( $missing == 0 ) { last; }
  } ## end while ( $cmd_retry_count )

  # Collect uploaded/missing files
  my @uploaded;
  my @missing;
  foreach my $file ( @{$files} ) {
    if( $upload{$file} ) {
      push(@missing,$dir2.$file);
    } else {
      push(@uploaded,$dir2.$file);
    }
  }
  # Files missing - exit
  if ( scalar @missing ) {
    foreach my $file ( @missing ) {
      warning("SVCPRF201E Upload IOSTAT file ".$file." failed.",SVCPRF201E);
    }
  }
  if ( scalar @uploaded ) {
    return(\@uploaded)
  }
}

sub Process_iostats_files {
  $debug = shift;
  my $timestamp = shift;
  my $iosdir = shift;
  my $file_base = shift;
  my $cfg = shift;
  my $uncomlete_data = 0;
  #
  my %data;
  while($timestamp->{'end'} < $timestamp->{'max'}) {
    # Timestamp:
    # max   ... upload start time - 300s
    # end   ... end of interval, end < max
    if($debug) { &message("Timestamp: ".&FlashSystem::Print_timestamps($timestamp)); }
    $uncomlete_data = Parse_iostats_files($timestamp,$iosdir,\%data,\&StatsCpuFilenames);
    $uncomlete_data = Parse_iostats_files($timestamp,$iosdir,\%data,\&StatsSystemFilenames);
    foreach my $port ( sort(keys(%{$cfg->{svc_portfc}})) ) {
      my $cid = $cfg->{svc_portfc}->{$port}->{canister_id};
      my $aid = $cfg->{svc_portfc}->{$port}->{adapter_id};
      my $pid = $cfg->{svc_portfc}->{$port}->{port_id};
      $uncomlete_data = Parse_iostats_files($timestamp,$iosdir,\%data,\&StatsFcPortFilenames,$port,$cid,$aid,$pid);
    }
    foreach my $interface ( sort(keys(%{$cfg->{svc_interfacefc}})) ) {
      my $cid = $cfg->{svc_interfacefc}->{$interface}->{canister_id};
      my $aid = $cfg->{svc_interfacefc}->{$interface}->{adapter_id};
      $uncomlete_data = Parse_iostats_files($timestamp,$iosdir,\%data,\&StatsFcInterfaceFilenames,$interface,$cid,$aid);
    }
    if ( $uncomlete_data ) {
      &warning("Zero count detected: Time: ".$timestamp->{'end'}."  ".epoch2datestr($timestamp->{'end'}));
      $timestamp->{'end'} += 300;
      next;
    }
    # Print data
    my $file_date;
    if( epoch2datestr($data{system}{'start'} ) =~ m/^(\d{4})-(\d{2})-(\d{2})\s+(\d{2}):(\d{2}):(\d{2})\s+/ ) {  # 2018-08-31 21:41:45 GMT+02:00
      $file_date = $1.$2.$3."_".$4.$5.$6.".out";
    } else {
      &warning("Unknown time detected. Timestamp (".$data{system}{'start'}."): ".epoch2datestr($data{system}{'start'}) );
      $timestamp->{'end'} += 300;
      next;
    }
    if($debug) {&message("Performance Output Filename: ".$file_base.$file_date)};
    #&message("__DBG__: Timestamp end  : ".epoch2datestr($timestamp->{'end'})."(".$timestamp->{'end'}.")");
    #&message("__DBG__: DataCpu        : ".epoch2datestr($data{'cpu'}{'end'})." - ".$data{'cpu'}{'system_cpu'});
    #&message("__DBG__: DataSys        : ".epoch2datestr($data{'system'}{'end'})." - ".$data{system}{'iops'});
    #&message("__DBG__: DataFcInterface: ".epoch2datestr($data{'fcinterface'}{'node1_interface1'}{'end'})." - ".$data{'fcinterface'}{'node1_interface1'}{'iops'});
    #&message("__DBG__: DataFcPort     : ".epoch2datestr($data{'fcport'}{'node1_p11'}{'end'})." - ".$data{'fcport'}{'node1_p11'}{'iops'});
    &PrintPerfData($file_base.$file_date,\%data,$cfg);
    $timestamp->{'end'} += 300;
  }
  return;
}

sub Parse_iostats_files {
  my $t = shift;
  my $iosdir = shift;
  my $data = shift;
  my $StatsFilenames_ref = shift;
  my $ent = shift;
  my $cid = shift;
  my $aid = shift;
  my $pid = shift;
  #
  my $uncomlete = 0;
  my ($tmkey,$dir1,$dir2,$files) = &{$StatsFilenames_ref}($cid,$aid,$pid);
  if($debug) {
    if(defined $ent){
      message("Process iostat files - ".$tmkey."|".$ent.": ".join(",",@{$files}));
    } else {
      message("Process iostat files - ".$tmkey.": ".join(",",@{$files}));
    }
  }
  foreach my $file ( @{$files} ) {
    open(my $fh,"<",$iosdir.$dir2.$file) || error("Cannot open file: ".$iosdir.$dir2.$file) && next;
    my $valstr;
    my $tmsstr;
    my $last = 0;
    my $row = 0;
    while( my $line = <$fh>  ) {
      $row++;
      my $isodd  = $row % 2 == 1;
      chomp($line);
      # Timestamps ... 1536025605.11,1536025605.0
      if( $isodd ) {
        # Values
        $valstr .= $line;
      }
      elsif( $line =~ m/^\d{10}\.\d{1,2},.*$/ ) {
        # Timestamps
        $tmsstr .= $line;
      } else {
        # Values
        &warning("Unexpected input - file: ".$iosdir.".".$dir2.".".$file."[".$row."] values: ".$line);
      }
    }
    my @tms = split(/,/,$tmsstr);
    my @val = split(/,/,$valstr);
    my $flag = 0;
    my $count = 0;
    my $sum = 0;
    my @int = ();
    # Set Timestamp MAX to last entry in iostat file
    $t->{'max'} = $tms[$#tms];
    #
    if( ($t->{'end'} ) < $tms[0] ) {
      $t->{'end'} = $tms[0] + 600;
      &warning("Start time is earlier than firt time in data - Correct: ".
              " First: ".$t->{'end'}." - ".epoch2datestr($t->{'end'}).
              " Last: ".$t->{'max'}." - ".epoch2datestr($t->{'max'}));
    }
    for(my $i=0; $i<$#tms; $i++) {
      # Validate time
      #&warning("_DBG_: ".$i." F=".$flag." S: ".($t->{'end'} - 300)." E: ".$t->{'end'}." V: ".$tms[$i] );
      if ( $tms[$i] =~ m/^\d{10}\.\d{1,2}$/ ) {
        # Start interval
        if( ($t->{'end'} - 300) < $tms[$i] ) { $flag = 1; }
        # End interval
        if ( $t->{'end'} < $tms[$i] )        { $flag = 0; }
        # Count average of values - create sum and count
        if ( $flag ) {
          $sum += $val[$i];
          $count++;
          push(@int,$tms[$i]);
        }
      }
      else {
        &warning("Bad timestamp format: ".$tms[$i]);
      }
    }
    #
    my $tmstart = $int[0];
    my $tmend = $int[$#int];
    if ( defined $ent ) {
      if ( $count > 0 ) {
        $data->{$tmkey}->{$ent}->{'name'} = $ent;
        $data->{$tmkey}->{$ent}->{'start'} = $tmstart;
        $data->{$tmkey}->{$ent}->{'end'} = $tmend;
        $data->{$tmkey}->{$ent}->{'count'} = $count;
        $data->{$tmkey}->{$ent}->{$file} = $sum / $count;
      } else {
        $uncomlete = 1;
        if ($debug) { &message("Zero count detected: Time: ".$t->{'end'}." First: ".$tms[0]." Last: ".$tms[$#tms]." Entry type: ".$tmkey."/".$ent."/".$file); }
        $data->{$tmkey}->{$ent}->{$file} = 0;
      }
    } else {
      if ( $count > 0 ) {
        $data->{$tmkey}->{'start'} = $tmstart;
        $data->{$tmkey}->{'end'} = $tmend;
        $data->{$tmkey}->{'count'} = $count;
        $data->{$tmkey}->{$file} = $sum / $count;
      } else {
        $uncomlete = 1;
        if ($debug) { &message("Zero count detected: Time: ".$t->{'end'}." First: ".$tms[0]." Last: ".$tms[$#tms]." Entry type: ".$tmkey."/".$file); }
        $data->{$tmkey}->{$file} = 0;
      }
    }
  }
  return($uncomlete);
}

# -------------------------------------------------------------------------------------------------
# PrintConfigData
# -------------------------------------------------------------------------------------------------
# PrintPerfData

sub PrintPerfData {
  my ${out_file} = shift;
  my $d_ref = shift;
  my $c_ref = shift;
  my $f_ref = shift;
  my $rflag = 0;
  #
  if ($debug) { &message("Output file: ${out_file}.tmp.") }
  open( PERFOUT, ">${out_file}.tmp" ) or die("Print data to file ${out_file}.tmp failed.");
  if ($debug) { print("CPUSystemStatistics - "); }
  if ( &CPUSystemStatistics( $d_ref ) ) { $rflag = 1; }
  if ($debug) { print("PortStatistics - "); }
  if ( &PortStatistics( $d_ref,$c_ref,$f_ref ) ) { $rflag = 1; }
  if ($debug) { print("InterfaceStatistics - "); }
  if ( &InterfaceStatistics( $d_ref,$c_ref ) ) { $rflag = 1; }
  if ($debug) { print("SystemStatistics - "); }
  if ( &SystemStatistics( $d_ref ) ) { $rflag = 1; }
  if ($debug) { print("MDiskGroupCapacityStatistics - "); }
  if ( &MDiskGroupCapacityStatistics( $d_ref,$c_ref ) ) { $rflag = 1; }
  if ($debug) { print("DONE\n"); }
  close(PERFOUT);
  if ($debug) { &message("Rename output file to: ${out_file}.") }
  if ( not -f ${out_file} . ".tmp" ) { return; }
  #my $ret = system( "$cmd_mv " . ${out_file} . ".tmp " . ${out_file} );
  my $ret = system( "$cmd_mv '${out_file}.tmp' '${out_file}'" );
  if ($ret) { error( "Couldn't rename file " . ${out_file} . ".tmp ." ); return; }
}

sub CPUSystemStatistics {
  my $d_ref = shift;
  #
  my $s_time   = epoch2datestr($d_ref->{cpu}->{'start'});
  my $e_time   = epoch2datestr($d_ref->{cpu}->{'end'});
  my $interval = $d_ref->{cpu}->{'end'} - $d_ref->{cpu}->{'start'};
  #
  print PERFOUT "\nCPU-System Level Statistics\n";
  print PERFOUT "\tInterval Start:   ", $s_time,   "\n";
  print PERFOUT "\tInterval End:     ", $e_time,   "\n";
  printf PERFOUT ("\tInterval Length: %4d seconds\n", $interval);
  print PERFOUT "---------------------\n";
  print PERFOUT "Node,Time,Interval,CPU Busy,CPU Limited,CPU Utilization - System,CPU Utilization - Compression,\n";
  #
  # Node,Time,Interval,
  printf PERFOUT ("%s,%s,%d,","system",$s_time,$interval);
  # CPU Busy,CPU Limited,CPU Utilization - System,CPU Utilization - Compression
  printf PERFOUT ("%d,0.000,%d,0.000,\n",$d_ref->{cpu}->{'system_cpu'}+$d_ref->{cpu}->{'user_cpu'},$d_ref->{cpu}->{'system_cpu'}+$d_ref->{cpu}->{'user_cpu'});
  print PERFOUT "\n";
}

sub PortStatistics {
  my $d_ref = shift;
  my $c_ref = shift;
  my $f_ref = shift;
  #
  my $s_time   = epoch2datestr($d_ref->{'system'}->{'start'});
  my $e_time   = epoch2datestr($d_ref->{'system'}->{'end'});
  my $interval = $d_ref->{'system'}->{'end'} - $d_ref->{'system'}->{'start'};
  #
  print PERFOUT "\nPort Level Statistics\n";
  print PERFOUT "\tInterval Start:   ", $s_time,   "\n";
  print PERFOUT "\tInterval End:     ", $e_time,   "\n";
  printf PERFOUT ("\tInterval Length: %4d seconds\n", $interval);
  print PERFOUT "---------------------\n";
  print PERFOUT "Port,Time,Interval (s),Port Send IO Rate (IO/s),Port Receive IO Rate (IO/s),Total Port IO Rate (IO/s),Port Send Data Rate (KB/s),Port Receive Data Rate (KB/s),Total Port Data Rate (KB/s),Port Send Transfer Size (kB),Port Receive Transfer Size (kB),Overall Port Transfer Size (kB),Port to Host Send IO Rate (IO/s),Port to Host Receive IO Rate (IO/s),Overall Port to Host IO Rate (IO/s),Port to Host Send Data Rate (KB/s),Port to Host Receive Data Rate (KB/s),Overall Port to Host Data Rate (KB/s),Port to Controller Send IO Rate (IO/s),Port to Controller Receive IO Rate (IO/s),Overall Port to Controller IO Rate (IO/s),Port to Controller Send Data Rate (KB/s),Port to Controller Receive Data Rate (KB/s),Overall Port to Controller Data Rate (KB/s),Port to Local Node Send IO Rate (IO/s),Port to Local Node Receive IO Rate (IO/s),Overall Port to Local Node IO Rate (IO/s),Port to Local Node Send Data Rate (KB/s),Port to Local Node Receive Data Rate (KB/s),Overall Port to Local Node Data Rate (KB/s),Port to Remote Node Send IO Rate (IO/s),Port to Remote Node Receive IO Rate (IO/s),Overall Port to Remote Node IO Rate (IO/s),Port to Remote Node Send Data Rate (KB/s),Port to Remote Node Receive Data Rate (KB/s),Overall Port to Reomte Node Data Rate (KB/s),Link Failure,Loss of Synch,Loss of Signal,Primitive Sequence Protocol Error,Invalid Transmission Word Count,Invalid CRC Count,Zero b2b (%),Node,Port type,WWPN,FC WWPN,FCoE WWPN,iSCSI IQN,\n";
  foreach my $ent ( sort( keys(%{$d_ref->{'fcport'}}) ) ) {
    if ( defined $d_ref->{'fcport'}->{$ent} ) {
      #&PortPrint($s_time,$interval,$d_ref->{'fcport'}->{$ent},$c_ref->{svc_portfc}->{$ent});
      &PortPrint($s_time,$interval,$d_ref->{'fcport'}->{$ent},$f_ref->{svc_portfc}->{$ent});
    }
  }
  return 0;
} ## end sub PortStatistics

sub PortPrint {
  my $time = shift;
  my $interval = shift;
  my $d = shift;
  my $c = shift;

  # Port,Time,Interval (s),
  printf PERFOUT "%s,%s,%d,",$d->{name},epoch2datestr($d->{start}),$d->{end} - $d->{start};
  # Port Send IO Rate (IO/s),Port Receive IO Rate (IO/s),Total Port IO Rate (IO/s),
  printf PERFOUT "%.3f,%.3f,%.3f,",$d->{iops_write},$d->{iops_read},$d->{iops};
  # Port Send Data Rate (KB/s),Port Receive Data Rate (KB/s),Total Port Data Rate (KB/s),
  printf PERFOUT "%.3f,%.3f,%.3f,",$d->{bw_write}*1024,$d->{bw_read}*1024,$d->{bw}*1024;
  # Port Send Transfer Size (kB),Port Receive Transfer Size (kB),Overall Port Transfer Size (kB),
  printf PERFOUT ("%.3f,%.3f,%.3f,",div($d->{bw_write}*1024,$d->{iops_write}),div($d->{bw_read}*1024,$d->{iops_read}),div($d->{bw}*1024,$d->{iops}));
  # Port to Host Send IO Rate (IO/s),Port to Host Receive IO Rate (IO/s),Overall Port to Host IO Rate (IO/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Host Send Data Rate (KB/s),Port to Host Receive Data Rate (KB/s),Overall Port to Host Data Rate (KB/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Controller Send IO Rate (IO/s),Port to Controller Receive IO Rate (IO/s),Overall Port to Controller IO Rate (IO/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Controller Send Data Rate (KB/s),Port to Controller Receive Data Rate (KB/s),Overall Port to Controller Data Rate (KB/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Local Node Send IO Rate (IO/s),Port to Local Node Receive IO Rate (IO/s),Overall Port to Local Node IO Rate (IO/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Local Node Send Data Rate (KB/s),Port to Local Node Receive Data Rate (KB/s),Overall Port to Local Node Data Rate (KB/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Remote Node Send IO Rate (IO/s),Port to Remote Node Receive IO Rate (IO/s),Overall Port to Remote Node IO Rate (IO/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Port to Remote Node Send Data Rate (KB/s),Port to Remote Node Receive Data Rate (KB/s),Overall Port to Reomte Node Data Rate (KB/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,",0,0,0);
  # Link Failure,Loss of Synch,Loss of Signal,Primitive Sequence Protocol Error,Invalid Transmission Word Count,Invalid CRC Count,Zero b2b (%)
  printf PERFOUT ("%d,%d,%d,%d,%d,%d,%.3f,",0,0,0,0,0,$d->{crc_errors},0);
  # Node,Port type,WWPN,FC WWPN,FCoE WWPN,iSCSI IQN,\n";
  if ( defined $c->{'WWPN'} ) {
    printf PERFOUT (",FC,0x%s,0x%s,,,,\n",$c->{'WWPN'},$c->{'WWPN'});
  } else {
    printf PERFOUT (",FC,,,,,,\n");
  }
}


sub InterfaceStatistics {
  my $d_ref = shift;
  my $c_ref = shift;
  #
  my $s_time   = epoch2datestr($d_ref->{'system'}->{'start'});
  my $e_time   = epoch2datestr($d_ref->{'system'}->{'end'});
  my $interval = $d_ref->{'system'}->{'end'} - $d_ref->{'system'}->{'start'};
  #
  print PERFOUT "\nInterface Level Statistics\n";
  print PERFOUT "\tInterval Start:   ", $s_time,   "\n";
  print PERFOUT "\tInterval End:     ", $e_time,   "\n";
  printf PERFOUT ("\tInterval Length: %4d seconds\n", $interval);
  print PERFOUT "---------------------\n";
  print PERFOUT "Interface,Time,Interval (s),Read IO Rate (IO/s),Write IO Rate (IO/s),Total IO Rate (IO/s),Read Data Rate (KB/s),Write Data Rate (KB/s),Total Data Rate (KB/s),Read Transfer Size (kB),Write Transfer Size (kB),Overall Transfer Size (kB),Read Response Time (ms),Write Response Time (ms),OverallResponse Time (ms),\n";
  #
  foreach my $ent ( sort( keys(%{$d_ref->{'fcinterface'}}) ) ) {
    if ( defined $d_ref->{'fcinterface'}->{$ent} ) {
      my $d = $d_ref->{'fcinterface'}->{$ent};
      # Count from ports?
      $d->{iops_read}=$d->{iops_write}=$d->{bw_read}=$d->{bw_write}=0;
      # Interface,Time,Interval (s),
      printf PERFOUT ("%s,%s,%d,",$ent,epoch2datestr($d->{start}),$d->{end} - $d->{start});
      # Read IO Rate (IO/s),Write IO Rate (IO/s),Total IO Rate (IO/s),Read Data Rate (KB/s),Write Data Rate (KB/s),Total Data Rate (KB/s),
      printf PERFOUT ("%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,",$d->{iops_read},$d->{iops_write},$d->{iops},$d->{bw_read}*1024,$d->{bw_write}*1024,$d->{bw}*1024);
      # Read Transfer Size (kB),Write Transfer Size (kB),Overall Transfer Size (kB),Read Response Time (ms),Write Response Time (ms),OverallResponse Time (ms),
      printf PERFOUT ("%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,\n",div($d->{bw_read}*1024,$d->{iops_read}),div($d->{bw_write}*1024,$d->{iops_write}),div($d->{bw}*1024,$d->{iops}),$d->{rd_latency_avg}/1000,$d->{wr_latency_avg}/1000,($d->{rd_latency_avg}+$d->{wr_latency_avg})/1000);
    }
  }
  print PERFOUT "\n";
}

sub SystemStatistics {
  my $d_ref = shift;
  my $c_ref = shift;
  #
  my $s_time   = epoch2datestr($d_ref->{'system'}->{'start'});
  my $e_time   = epoch2datestr($d_ref->{'system'}->{'end'});
  my $interval = $d_ref->{'system'}->{'end'} - $d_ref->{'system'}->{'start'};
  #
  print PERFOUT "\nSystem Level Statistics\n";
  print PERFOUT "\tInterval Start:   ", $s_time,   "\n";
  print PERFOUT "\tInterval End:     ", $e_time,   "\n";
  printf PERFOUT ("\tInterval Length: %4d seconds\n", $interval);
  print PERFOUT "---------------------\n";
  print PERFOUT "System,Time,Interval (s),Backend Read IO Rate (IO/s),Backend Write IO Rate (IO/s),Backend Total IO Rate (IO/s),Backend Read Data Rate (KB/s),Backend Write Data Rate (KB/s),Backend Total Data Rate (KB/s),Backend Read Transfer Size (kB),Backend Write Transfer Size (kB),Backend Overall Transfer Size (kB),Backend Read Response Time (ms),Backend Write Response Time (ms),Overall Backend Response Time (ms),\n";
  #
  my $d = $d_ref->{'system'};
  # System,Time,Interval (s),
  printf PERFOUT ("%s,%s,%d,","system",$s_time,$interval);
  # Backend Read IO Rate (IO/s),Backend Write IO Rate (IO/s),Backend Total IO Rate (IO/s),Backend Read Data Rate (KB/s),Backend Write Data Rate (KB/s),Backend Total Data Rate (KB/s),
  printf PERFOUT ("%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,",$d->{iops_read},$d->{iops_write},$d->{iops},$d->{bw_read}*1024,$d->{bw_write}*1024,$d->{bw}*1024);
  # Backend Read Transfer Size (kB),Backend Write Transfer Size (kB),Backend Overall Transfer Size (kB),Backend Read Response Time (ms),Backend Write Response Time (ms),Overall Backend Response Time (ms),
  printf PERFOUT ("%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,\n",div($d->{bw_read}*1024,$d->{iops_read}),div($d->{bw_write}*1024,$d->{iops_write}),div($d->{bw}*1024,$d->{iops}),$d->{rd_latency_avg}/1000,$d->{wr_latency_avg}/1000,($d->{rd_latency_avg}+$d->{wr_latency_avg})/1000);
  print PERFOUT "\n";
}

sub MDiskGroupCapacityStatistics {
  my $d_ref = shift;
  my $c_ref = shift;
  #
  my $s_time   = epoch2datestr($d_ref->{'system'}->{'start'});
  my $e_time   = epoch2datestr($d_ref->{'system'}->{'end'});
  my $interval = $d_ref->{'system'}->{'end'} - $d_ref->{'system'}->{'start'};
  #
  print PERFOUT "\nPool Capacity Statistics\n";
  print PERFOUT "\tInterval Start:   ", $s_time,   "\n";
  print PERFOUT "\tInterval End:     ", $e_time,   "\n";
  printf PERFOUT ("\tInterval Length: %4d seconds\n", $interval);
  print PERFOUT "---------------------\n";

  print PERFOUT "name,id,status,mdisk count,volume count,capacity (TB),extent size,free capacity (TB),virtual capacity (TB),used capacity (TB),real capacity (TB),overallocation,warning (%),easy tier,easy tier status,TIER-0 type,TIER-0 mdisk count,TIER-0 capacity (TB),TIER-0 free capacity (TB),TIER-1 type,TIER-1 mdisk count,TIER-1 capacity (TB),TIER-1 free capacity (TB),TIER-2 type,TIER-2 mdisk count,TIER-2 capacity (TB),TIER-2 free capacity (TB),compression active,compression virtual capacity (TB),compression compressed capacity (TB),compression uncompressed capacity (TB),TIER-0-NL type,TIER-0-NL mdisk count,TIER-0-NL capacity (TB),TIER-0-NL free capacity (TB)\n";
  #
  foreach my $ent ( sort( keys(%{$c_ref->{svc_pool}}) ) ) {
    my $d = $c_ref->{svc_pool}->{$ent};
    # name,id,status,mdisk count,volume count,
    printf PERFOUT ("%s,%d,%s,%d,%d,",$d->{name},$d->{id},$d->{status},$d->{mdisk_count},$d->{vdisk_count});
    # capacity (TB),extent size,free capacity (TB),virtual capacity (TB),used capacity (TB),real capacity (TB),
    printf PERFOUT ("%.3f,%d,%.3f,%.3f,%.3f,%.3f,",$d->{capacity}/1099511627776,$d->{extent_size},$d->{free_capacity}/1099511627776,$d->{virtual_capacity}/1099511627776,$d->{used_capacity}/1099511627776,$d->{real_capacity}/1099511627776 );
    # overallocation,warning (%),easy tier,easy tier status,
    printf PERFOUT ("%d,%d,no,,",100,100);
    # TIER-0 type,TIER-0 mdisk count,TIER-0 capacity (TB),TIER-0 free capacity (TB),
    printf PERFOUT ("%s,%d,%.3f,%.3f,","tier0_flash",1,$d->{capacity}/1099511627776,$d->{free_capacity}/1099511627776);
    # TIER-1 type,TIER-1 mdisk count,TIER-1 capacity (TB),TIER-1 free capacity (TB),
    printf PERFOUT ("%s,%d,%.3f,%.3f,","tier_enterprise",0,0,0);
    # TIER-2 type,TIER-2 mdisk count,TIER-2 capacity (TB),TIER-2 free capacity (TB),
    printf PERFOUT ("%s,%d,%.3f,%.3f,","tier_nearline",0,0,0);
    # compression active,compression virtual capacity (TB),compression compressed capacity (TB),compression uncompressed capacity (TB),
    printf PERFOUT ("%s,%.3f,%.3f,%.3f,","no",0,0,0);
    # TIER-0-NL type,TIER-0-NL mdisk count,TIER-0-NL capacity (TB),TIER-0-NL free capacity (TB)
    printf PERFOUT ("%s,%d,%.3f,%.3f,\n","tier1_flash",0,0,0);
  }
  print PERFOUT "\n";
}

sub div {
	my $a = shift;
	my $b = shift;
	if ( $b > 0 ) {
		return($a/$b);
	} else {
		return(0);
	}
}

# -------------------------------------------------------------------------------------------------

sub epoch2datestr {
  # Vystup: 2012-04-25 09:50:00 GMT+01:00
  my $tm = shift;
  #
  if( not defined $tm ) { return "Undefined_time_epoch"; }
  if( $tm =~ m/^\d{10}$/ || $tm =~ m/^\d{10}\.\d{1,2}$/ ) {
    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) = localtime($tm);
    my $y   = $year + 1900;
    my $m   = $mon + 1;
    my $tz  = tzoffset();
    my $str = sprintf( "%4s-%02s-%02s %02s:%02s:%02s %s", $y, $m, $mday, $hour, $min, $sec, $tz );
    return ($str);
  } else {
    return "Unknown_time_epoch";
  }
} ## end sub epoch2datestr

sub tzoffset {
  my $sign       = "";
  my $t          = time();
  my $local_time = localtime($t);
  my $gm_time    = gmtime($t);
  my $local      = str2time($local_time);
  my $utc        = str2time($gm_time);
  my $tzo        = ( ( $local - $utc ) / 3600 );
  if ( $tzo > 0 ) { $sign = "+"; }
  return sprintf( "%s%s%02d:00", "GMT", $sign, $tzo );
} ## end sub tzoffset


sub message {
  my $msg = shift;
  my $tm  = localtime();
  print( $tm. " - INFO    - FlashSystem.pm: " . $msg . "\n" );
}

sub warning {
  my ( $msg, $rc ) = @_;
  my $tm = localtime();
  print ( $tm . " - WARNING - FlashSystem.pm: " . $msg . "\n" );
  print STDERR ( $tm . " - WARNING - FlashSystem.pm: " . $msg . "\n" );
}

sub error {
  my ( $msg, $rc ) = @_;
  my $tm = localtime();
  if ( defined $rc ) {
    print STDERR ( $tm . " - ERROR   - FlashSystem.pm: " . $msg . " - Return code: " . $rc . "\n" );
  }
  else {
    print STDERR ( $tm . " - ERROR   - FlashSystem.pm: " . $msg . "\n" );
  }
} ## end sub error

sub cmd_exec_ntimes {
  my $cmd = shift;
  my $max = shift;
  my $ret = 1;
  my $out;
  my $count = 0;    # let it run only 5 times ( 10 x 80 sec timeout in ssh = 800secs)
  while ( $ret ) {
    $count++;
    ($ret,$out) = &runcmd("$cmd 2>&1");
    if ( $ret && $count < $max ) {
      for my $line ( @{$out}) {
        chomp($line);
        if ( $line =~ /No such file or directory/ ) {
          return 2;
        } else {
          warning("Command output: ".$line);
        }
      }
      warning("Wait 5s and Retry $cmd");
      sleep 5;
      next;
    }
    if ( $ret && $count == $max ) {
      error("Failed: $cmd : too many attempts : $count");
      return 1;
    }
  } ## end while ( $ret && !$terminate_pgm)
  return $ret;
} ## end sub cmd_exec_ntimes

sub runcmd {
  my $cmd = "@_";
  my $ret_value; my $msg;
  my $rc = 0;
  my @out;
  #
  (@out = `@_`);
  $ret_value = $?;
  if ($ret_value == -1) {
    $msg = sprintf "runcmd: Command \'%s\' Failed to execute: $!\n",$cmd;
    if ($debug) { message($msg); }
    $rc = -1;
  } elsif ($ret_value & 127) {
    $msg = sprintf "runcmd: Command \'%s\' died with signal %d\n",$cmd,($ret_value & 127);
    if ($debug) { message($msg); }
    $rc = ($ret_value & 127);
  } else {
    $msg = sprintf "runcmd: Command \'%s\' exited with value %d\n",$cmd,$ret_value >> 8;
    if ($debug) { message($msg); }
    $rc = ($ret_value >> 8);
  }
  return ($rc,\@out);
}

return 1;
