package SWIZ;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";

use Xormon;
use Data::Dumper;
use JSON qw(encode_json decode_json);
use Date::Parse; # str2time
use POSIX qw(mktime);
use XormonLogger;

my $xormon;
my $cfg;
my $wrkdir;
my $sample_rate = 300; # default
my %capacity;
my $translator;
my %tier_arch;
my $logger;
my $last_perf_timestamp;

sub process_collected_data {
  $xormon = shift;
  $cfg    = shift;
  $wrkdir = shift;

  my $status = "successful"; # status code

  #
  # test main variables
  #
  unless ( exists $cfg->{'hostcfg_id'} ) { Xormon::error("Cannot continue, 'hostcfg_id' is not configured!") && return "error"; }
  unless ( exists $cfg->{'hostalias'} )  { Xormon::error("Cannot continue, 'hostalias' is not configured!")  && return "error"; }
  $logger = XormonLogger->new( 'swiz', $cfg->{hostcfg_id}, $cfg->{hostalias} );

  if ( exists $cfg->{'sampleRate'} && Xormon::isdigit($cfg->{'sampleRate'}) ) { $sample_rate = $cfg->{'sampleRate'} * 60; } # from minutes to seconds

  opendir( my $dir, "$wrkdir" ) || $logger->error( "Cannot open directory '$wrkdir' " . __FILE__ . ":" . __LINE__ ) && return "error";
  my @files = grep !/^\.\.?$/, readdir($dir);
  closedir($dir);
  chomp @files;

  my @conf_files         = grep( /$cfg->{'hostalias'}\_svcconf\_20.*out$/, @files );
  my @svcperf_files      = grep( /$cfg->{'hostalias'}\_svcperf\_20.*out$/, @files );
  my @svcflashperf_files = grep( /$cfg->{'hostalias'}\_svcflashperf\_20.*out$/, @files );
  my @svccapacity_files  = grep( /$cfg->{'hostalias'}\_svccapacity\_20.*out$/, @files );

  #print "$$: ".localtime().": svcperf=".scalar(@svcperf_files).", svcflashperf=".scalar(@svcflashperf_files).", svccapacity=".scalar(@svccapacity_files).", svcconf=".scalar(@conf_files)."\n";

  #
  # get microservice data
  #
  my $service_data = $xormon->getData();

  #
  # Hostcfg status
  #
  my $conn_ret = "";
  my $data_ret = "";
  # connection status: test conf file which is created every agent run
  if ( scalar(@conf_files) > 0 || scalar(@svcperf_files) > 0 || scalar(@svcflashperf_files) > 0 ) {
    $conn_ret = $xormon->isConnected(1);
  }
  else {
    $conn_ret = $xormon->isConnected(0, "conf/perf file has not been created");
  }
  $logger->info("isConnected: $conn_ret");
  # data check status: sometimes there is no any perf file, so therefore it's necessary to test last perf timestamp
  if ( scalar(@svcperf_files) > 0 || scalar(@svcflashperf_files) > 0 ) {
    $data_ret = $xormon->hasData(1);
  }
  elsif ( exists $service_data->{last_perf_timestamp} ) {
    if ( ( time() - $service_data->{last_perf_timestamp} ) < 900 ) { # test if perf is not older than 15 minutes
      $data_ret = $xormon->hasData(1);
    }
    else {
      $data_ret = $xormon->hasData(0, "no current stats, last available stats: $service_data->{last_perf_timestamp} : ".localtime($service_data->{last_perf_timestamp}));
    }
  }
  else {
    $data_ret = $xormon->hasData(0, "no available stats");
  }
  $logger->info("hasData: $data_ret");

  #
  # get microservice data
  #
  #my $service_data_in  = $xormon->getData();
  #my $service_data_out = {};
  #if ( exists $service_data_in->{last_conf_timestamp} ) {
  #  $service_data_out->{last_conf_timestamp} = $service_data_in->{last_conf_timestamp};
  #}

  #
  # capacity files
  #
  if ( scalar(@svccapacity_files) > 0 ) {
    chomp @svccapacity_files;

    foreach my $file ( sort { lc $a cmp lc $b } @svccapacity_files ) {
      $logger->info("parse capacity file: $wrkdir/$file");
      my $data = parse_cap("$wrkdir/$file");
      prepare_cap($data); # data stored to %stats
      remove_file("$wrkdir/$file");
    }
  }

  #
  # conf files, send architecture and conf data (only once per hour, use only the latest conf file)
  #
  #my $act_timestamp = time();
  #if ( ! exists $service_data_in->{last_conf_timestamp} || ! Xormon::isdigit($service_data_in->{last_conf_timestamp}) || $act_timestamp - $service_data_in->{last_conf_timestamp} > 3600 ) {
  #  $service_data_out->{last_conf_timestamp} = $act_timestamp;

  #  if ( scalar(@conf_files) > 0 ) {
  #    chomp @conf_files;

  #    my $idx = 0;
  #    foreach my $file ( sort { lc $b cmp lc $a } @conf_files ) {
  #      $idx++;
  #      if ( $idx == 1 ) { # use only the last conf file
  #        $logger->info("parse conf file: $wrkdir/$file");
  #        my $data = parse_conf("$wrkdir/$file");
  #        send_conf($data);
  #      }
  #      remove_file("$wrkdir/$file");
  #    }
  #  }
  #}

  # KZ: cannot be used as above, bcs information about hosts come only once per day, so therefore we have to parse each conf file... grrr
  if ( scalar(@conf_files) > 0 ) {
    chomp @conf_files;

    foreach my $file ( sort { lc $b cmp lc $a } @conf_files ) {
      $logger->info("parse conf file: $wrkdir/$file");
      my $data = parse_conf("$wrkdir/$file");
      send_conf($data);
      remove_file("$wrkdir/$file");
    }
  }

  #
  # save microservice data
  #
  #$xormon->saveData(\%{ $service_data_out });

  #
  # perf files
  #
  if ( scalar(@svcperf_files) > 0 ) {
    chomp @svcperf_files;

    foreach my $file ( sort { lc $a cmp lc $b } @svcperf_files ) {
      $logger->info("parse perf file: $wrkdir/$file");
      my $data     = parse_perf("$wrkdir/$file");
      my $perf_out = prepare_perf($data);
      my $rc       = $xormon->saveJSON2($perf_out);

      if ( defined $rc ) {
        $logger->info("save perf: $rc");

        if ( $rc eq "201" ) {
          $status = "successful";
        }
        elsif ( $rc eq "401" ) {
          $status = "unauthorized";
          last;
        }
        else {
          $status = "error";
          last;
        }
      }
      else {
        $status = "error";
        last;
      }

      remove_file("$wrkdir/$file");
    }

    # stop if data saving fails
    if ( $status ne 'successful' ) { return $status; }
  }

  #
  # health status
  #
  health_status();

  #
  # save microservice data
  #
  if ( defined $last_perf_timestamp ) { $service_data->{last_perf_timestamp} = $last_perf_timestamp; }
  $xormon->saveData(\%{ $service_data });

  return $status;
}

sub health_status {
  my @hs = ();

  my %hs_data;
  my %hs_data_subsystem;
  my $main_status = "";
  my $main_reason = "";

  if ( -f "$wrkdir/health_status.json" ) {
    %hs_data = %{ Xormon::read_json("$wrkdir/health_status.json") };
  }

  if ( -f "$wrkdir/health_status_subsystem.json" ) {
    %hs_data_subsystem = %{ Xormon::read_json("$wrkdir/health_status_subsystem.json") };
  }

  if ( exists $hs_data{status} && exists $hs_data{reason} ) {
    $main_status = $hs_data{status};
    $main_reason = $hs_data{reason};
  }
  else {
    # sometimes HS is not available
    $logger->error("Health status has not been found!" . __FILE__ . ":" . __LINE__ ) && return 1;
  }

  # add events to error message
  if ( exists $hs_data{events} && ref($hs_data{events}) eq "ARRAY" ) {
    my @events = ("", "time : event_id : object_type : object_name : description : status : fixed : sequence_number : object_id : copy_id : error_code");
    foreach my $event ( @{ $hs_data{events} } ) {
      my $time            = ( exists $event->{time} )            ? $event->{time}            : "";
      my $event_id        = ( exists $event->{event_id} )        ? $event->{event_id}        : "";
      my $object_type     = ( exists $event->{object_type} )     ? $event->{object_type}     : "";
      my $object_name     = ( exists $event->{object_name} )     ? $event->{object_name}     : "";
      my $description     = ( exists $event->{description} )     ? $event->{description}     : "";
      my $status          = ( exists $event->{status} )          ? $event->{status}          : "";
      my $fixed           = ( exists $event->{fixed} )           ? $event->{fixed}           : "";
      my $sequence_number = ( exists $event->{sequence_number} ) ? $event->{sequence_number} : "";
      my $object_id       = ( exists $event->{object_id} )       ? $event->{object_id}       : "";
      my $copy_id         = ( exists $event->{copy_id} )         ? $event->{copy_id}         : "";
      my $error_code      = ( exists $event->{error_code} )      ? $event->{error_code}      : "";

      push(@events, "$time : $event_id : $object_type : $object_name : $description : $status : $fixed : $sequence_number : $object_id : $copy_id : $error_code");
    }

    $main_reason .= "\n".join("\n", @events)."\n";
    #push(@hs, { "item_id" => $cfg->{'hostcfg_id'}, "status" => $hs_data{status}, "data" => $hs_data{reason}."\n".join("\n", @events)."\n" });
  }

  # add events to error message
  if ( exists $hs_data_subsystem{hs_data} && ref($hs_data_subsystem{hs_data}) eq "ARRAY" ) {
    my @events = ("", "section : id : name : status : description");
    foreach my $event ( @{ $hs_data_subsystem{hs_data} } ) {
      my $section     = ( exists $event->{section} )     ? $event->{section}     : "";
      my $id          = ( exists $event->{id} )          ? $event->{id}          : "";
      my $name        = ( exists $event->{name} )        ? $event->{name}        : "";
      my $status      = ( exists $event->{status} )      ? $event->{status}      : "";
      my $description = ( exists $event->{description} ) ? $event->{description} : "";

      push(@events, "$section : $id : $name : $status : $description");
    }

    $main_reason .= "\n".join("\n", @events)."\n";
    #push(@hs, { "item_id" => $cfg->{'hostcfg_id'}, "status" => $hs_data{status}, "data" => $hs_data{reason}."\n".join("\n", @events)."\n" });
  }

  push(@hs, { "item_id" => $cfg->{'hostcfg_id'}, "status" => $main_status, "data" => $main_reason });

  # send statuses
  my $rc = $xormon->saveStatus(\@hs);
  $logger->info("save health status: $rc");

  # save HS data
  my $rc_hs = $xormon->saveConfJSON( { $cfg->{'hostcfg_id'} => { "health_status_data" => encode_json(\%hs_data) } }, 1 );
  $logger->info("save health status table: $rc_hs");

  # save HS data subsystem
  my $rc_hs_subsys = $xormon->saveConfJSON( { $cfg->{'hostcfg_id'} => { "health_status_subsystem" => encode_json(\@{$hs_data_subsystem{hs_data}}) } }, 1 );
  $logger->info("save health status subsystem table: $rc_hs_subsys");

  return 1;
}

sub prepare_cap {
  my $data_in = shift;

  foreach my $timestamp ( sort { $a <=> $b } keys %{ $data_in } ) {
    if ( exists $data_in->{$timestamp}{storage} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{storage} } ) {
        my $var = $data_in->{$timestamp}{storage}{$line_idx};

        if ( exists $var->{'Kapacitni vrstva'} && $var->{'Kapacitni vrstva'} ne '' && $var->{'Kapacitni vrstva'} ne '-' ) {
          #
          # TIER conf
          #
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."TIER".$var->{'Kapacitni vrstva'}) );
          #my @tier_arch;
          # Architecture
          %tier_arch = (
            "item_id"   => $id,
            "label"     => $var->{'Kapacitni vrstva'},
            "hw_type"   => "swiz",
            "subsystem" => "tier",
            "parents"   => [$cfg->{'hostcfg_id'}],
            "hostcfg_id" => $cfg->{'hostcfg_id'}
          );
          if ( $var->{'Kapacitni vrstva'} eq "tier0_flash" || $var->{'Kapacitni vrstva'} eq "ssd" ) {
            $tier_arch{label} = "Tier 0 Flash";
          }
          elsif ( $var->{'Kapacitni vrstva'} eq "tier1_flash" ) {
            $tier_arch{label} = "Tier 1 Flash";
          }
          elsif ( $var->{'Kapacitni vrstva'} eq "tier_scm" ) {
            $tier_arch{label} = "Tier SCM";
          }
          elsif ( $var->{'Kapacitni vrstva'} eq "tier_enterprise" || $var->{'Kapacitni vrstva'} eq "hdd" || $var->{'Kapacitni vrstva'} eq "enterprise" ) {
            $tier_arch{label} = "Enterprise Disk";
          }
          elsif ( $var->{'Kapacitni vrstva'} eq "tier_nearline" || $var->{'Kapacitni vrstva'} eq "nearline" ) {
            $tier_arch{label} = "Nearline Disk";
          }
          #push(@tier_arch, \%arch);
          #save architecture
          #$xormon->saveArchitecture(\@tier_arch);
          #print "#################### Architecture\n";
          #print Dumper @tier_arch;

          #
          # TIER perf (cap)
          #
          if ( exists $var->{'Hruba kapacita'} )                           { $capacity{tier}{$id}{raw_capacity_total}          = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita kapacita'} )                   { $capacity{tier}{$id}{raw_capacity_used}           = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita kapacita'},"GiB")); }
          if ( exists $var->{'Hruba RAID volna kapacita'} )                { $capacity{tier}{$id}{raw_raid_capacity_free}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba RAID volna kapacita'},"GiB")); }
          if ( exists $var->{'Hruba RAID kapacita'} )                      { $capacity{tier}{$id}{raw_raid_capacity_total}     = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba RAID kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita snapshot kapacita'} )          { $capacity{tier}{$id}{raw_snapshot_capacity_used}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita snapshot kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita datova kapacita'} )            { $capacity{tier}{$id}{raw_data_capacity_used}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita datova kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita replikovana kapacita'} )       { $capacity{tier}{$id}{raw_replication_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita replikovana kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana RAID volna kapacita'} )          { $capacity{tier}{$id}{format_raid_capacity_free}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana RAID volna kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana RAID kapacita'} )                { $capacity{tier}{$id}{format_raid_capacity_total}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana RAID kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana pouzita replikovana kapacita'} ) { $capacity{tier}{$id}{format_replication_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana pouzita replikovana kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana pouzita snapshot kapacita'} )    { $capacity{tier}{$id}{format_snapshot_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana pouzita snapshot kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana Pouzita datova kapacita'} )      { $capacity{tier}{$id}{format_data_capacity_used}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana Pouzita datova kapacita'},"GiB")); }
          if ( exists $var->{'Celkova efektivita'} )                       { $capacity{tier}{$id}{effective_capacity_total}    = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Celkova efektivita'},"GiB")); }
          if ( exists $var->{'Uspora deduplikace'} )                       { $capacity{tier}{$id}{dedupe_saving}               = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora deduplikace'},"GiB")); }
          if ( exists $var->{'Uspora komprese'} )                          { $capacity{tier}{$id}{compression_saving}          = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora komprese'},"GiB")); }
          if ( exists $var->{'Uspora thin-provisioning'} )                 { $capacity{tier}{$id}{thin_provisioning}           = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora thin-provisioning'},"GiB")); }
          if ( exists $var->{'Spare kapacita'} )                           { $capacity{tier}{$id}{spare_capacity_total}        = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Spare kapacita'},"GiB")); }
          #if ( exists $var->{'Hruba pouzita system kapacita'} )            { $capacity{tier}{$id}{raw_capacity_used} = $var->{'Hruba pouzita system kapacita'}; }
          #if ( exists $var->{'Formatovana pouzita system kapacita'} )      { $capacity{tier}{$id}{} = $var->{'Formatovana pouzita system kapacita'}; }

          if ( exists $capacity{tier}{$id}{raw_capacity_total} && Xormon::isdigit($capacity{tier}{$id}{raw_capacity_total}) && exists $capacity{tier}{$id}{raw_capacity_used} && Xormon::isdigit($capacity{tier}{$id}{raw_capacity_used}) ) {
            $capacity{tier}{$id}{raw_capacity_free} = Xormon::formatNumberByType("int",$capacity{tier}{$id}{raw_capacity_total} - $capacity{tier}{$id}{raw_capacity_used});
          }
          if ( exists $capacity{tier}{$id}{raw_raid_capacity_total} && Xormon::isdigit($capacity{tier}{$id}{raw_raid_capacity_total}) && exists $capacity{tier}{$id}{raw_raid_capacity_free} && Xormon::isdigit($capacity{tier}{$id}{raw_raid_capacity_free}) ) {
            $capacity{tier}{$id}{raw_raid_capacity_used} = Xormon::formatNumberByType("int",$capacity{tier}{$id}{raw_raid_capacity_total} - $capacity{tier}{$id}{raw_raid_capacity_free});
          }
          if ( exists $capacity{tier}{$id}{format_raid_capacity_total} && Xormon::isdigit($capacity{tier}{$id}{format_raid_capacity_total}) && exists $capacity{tier}{$id}{format_raid_capacity_free} && Xormon::isdigit($capacity{tier}{$id}{format_raid_capacity_free}) ) {
            $capacity{tier}{$id}{format_raid_capacity_used} = Xormon::formatNumberByType("int",$capacity{tier}{$id}{format_raid_capacity_total} - $capacity{tier}{$id}{format_raid_capacity_free});
          }
        }
        else {
          #
          # DEVICE
          #
          my $id = $cfg->{'hostcfg_id'};

          if ( exists $var->{'Hruba kapacita'} )                           { $capacity{device}{$id}{raw_capacity_total}          = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita kapacita'} )                   { $capacity{device}{$id}{raw_capacity_used}           = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita kapacita'},"GiB")); }
          if ( exists $var->{'Hruba RAID volna kapacita'} )                { $capacity{device}{$id}{raw_raid_capacity_free}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba RAID volna kapacita'},"GiB")); }
          if ( exists $var->{'Hruba RAID kapacita'} )                      { $capacity{device}{$id}{raw_raid_capacity_total}     = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba RAID kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita snapshot kapacita'} )          { $capacity{device}{$id}{raw_snapshot_capacity_used}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita snapshot kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita datova kapacita'} )            { $capacity{device}{$id}{raw_data_capacity_used}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita datova kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita replikovana kapacita'} )       { $capacity{device}{$id}{raw_replication_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita replikovana kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana RAID volna kapacita'} )          { $capacity{device}{$id}{format_raid_capacity_free}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana RAID volna kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana RAID kapacita'} )                { $capacity{device}{$id}{format_raid_capacity_total}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana RAID kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana pouzita replikovana kapacita'} ) { $capacity{device}{$id}{format_replication_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana pouzita replikovana kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana pouzita snapshot kapacita'} )    { $capacity{device}{$id}{format_snapshot_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana pouzita snapshot kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana Pouzita datova kapacita'} )      { $capacity{device}{$id}{format_data_capacity_used}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana Pouzita datova kapacita'},"GiB")); }
          if ( exists $var->{'Celkova efektivita'} )                       { $capacity{device}{$id}{effective_capacity_total}    = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Celkova efektivita'},"GiB")); }
          if ( exists $var->{'Uspora deduplikace'} )                       { $capacity{device}{$id}{dedupe_saving}               = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora deduplikace'},"GiB")); }
          if ( exists $var->{'Uspora komprese'} )                          { $capacity{device}{$id}{compression_saving}          = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora komprese'},"GiB")); }
          if ( exists $var->{'Uspora thin-provisioning'} )                 { $capacity{device}{$id}{thin_provisioning}           = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora thin-provisioning'},"GiB")); }
          if ( exists $var->{'Spare kapacita'} )                           { $capacity{device}{$id}{spare_capacity_total}        = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Spare kapacita'},"GiB")); }
          #if ( exists $var->{'Hruba pouzita system kapacita'} )            { $capacity{device}{$id}{raw_capacity_used} = $var->{'Hruba pouzita system kapacita'}; }
          #if ( exists $var->{'Formatovana pouzita system kapacita'} )      { $capacity{device}{$id}{} = $var->{'Formatovana pouzita system kapacita'}; }

          if ( exists $capacity{device}{$id}{raw_capacity_total} && Xormon::isdigit($capacity{device}{$id}{raw_capacity_total}) && exists $capacity{device}{$id}{raw_capacity_used} && Xormon::isdigit($capacity{device}{$id}{raw_capacity_used}) ) {
            $capacity{device}{$id}{raw_capacity_free} = Xormon::formatNumberByType("int",$capacity{device}{$id}{raw_capacity_total} - $capacity{device}{$id}{raw_capacity_used});
          }
          if ( exists $capacity{device}{$id}{raw_raid_capacity_total} && Xormon::isdigit($capacity{device}{$id}{raw_raid_capacity_total}) && exists $capacity{device}{$id}{raw_raid_capacity_free} && Xormon::isdigit($capacity{device}{$id}{raw_raid_capacity_free}) ) {
            $capacity{device}{$id}{raw_raid_capacity_used} = Xormon::formatNumberByType("int",$capacity{device}{$id}{raw_raid_capacity_total} - $capacity{device}{$id}{raw_raid_capacity_free});
          }
          if ( exists $capacity{device}{$id}{format_raid_capacity_total} && Xormon::isdigit($capacity{device}{$id}{format_raid_capacity_total}) && exists $capacity{device}{$id}{format_raid_capacity_free} && Xormon::isdigit($capacity{device}{$id}{format_raid_capacity_free}) ) {
            $capacity{device}{$id}{format_raid_capacity_used} = Xormon::formatNumberByType("int",$capacity{device}{$id}{format_raid_capacity_total} - $capacity{device}{$id}{format_raid_capacity_free});
          }
        }

      }
    }

    #
    # POOL
    #
    if ( exists $data_in->{$timestamp}{pool} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{pool} } ) {
        my $var = $data_in->{$timestamp}{pool}{$line_idx};

        if ( exists $var->{'Kapacitni vrstva'} && $var->{'Kapacitni vrstva'} eq '-' && exists $var->{'Cislo poolu'} && $var->{'Cislo poolu'} ne '' ) {
          my $id  = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$var->{'Cislo poolu'}) );

          if ( exists $var->{'Hruba kapacita'} )                           { $capacity{pool}{$id}{raw_capacity_total}          = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita kapacita'} )                   { $capacity{pool}{$id}{raw_capacity_used}           = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita kapacita'},"GiB")); }
          if ( exists $var->{'Hruba RAID volna kapacita'} )                { $capacity{pool}{$id}{raw_raid_capacity_free}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba RAID volna kapacita'},"GiB")); }
          if ( exists $var->{'Hruba RAID kapacita'} )                      { $capacity{pool}{$id}{raw_raid_capacity_total}     = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba RAID kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita snapshot kapacita'} )          { $capacity{pool}{$id}{raw_snapshot_capacity_used}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita snapshot kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita datova kapacita'} )            { $capacity{pool}{$id}{raw_data_capacity_used}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita datova kapacita'},"GiB")); }
          if ( exists $var->{'Hruba pouzita replikovana kapacita'} )       { $capacity{pool}{$id}{raw_replication_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Hruba pouzita replikovana kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana RAID volna kapacita'} )          { $capacity{pool}{$id}{format_raid_capacity_free}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana RAID volna kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana RAID kapacita'} )                { $capacity{pool}{$id}{format_raid_capacity_total}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana RAID kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana pouzita replikovana kapacita'} ) { $capacity{pool}{$id}{format_replication_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana pouzita replikovana kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana pouzita snapshot kapacita'} )    { $capacity{pool}{$id}{format_snapshot_capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana pouzita snapshot kapacita'},"GiB")); }
          if ( exists $var->{'Formatovana Pouzita datova kapacita'} )      { $capacity{pool}{$id}{format_data_capacity_used}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Formatovana Pouzita datova kapacita'},"GiB")); }
          if ( exists $var->{'Celkova efektivita'} )                       { $capacity{pool}{$id}{effective_capacity_total}    = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Celkova efektivita'},"GiB")); }
          if ( exists $var->{'Uspora deduplikace'} )                       { $capacity{pool}{$id}{dedupe_saving}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora deduplikace'},"GiB")); }
          if ( exists $var->{'Uspora komprese'} )                          { $capacity{pool}{$id}{compression_saving} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora komprese'},"GiB")); }
          if ( exists $var->{'Uspora thin-provisioning'} )                 { $capacity{pool}{$id}{thin_provisioning}  = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Uspora thin-provisioning'},"GiB")); }
          if ( exists $var->{'Spare kapacita'} )                           { $capacity{pool}{$id}{spare_capacity_total}        = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'Spare kapacita'},"GiB")); }
          #if ( exists $var->{'Hruba pouzita system kapacita'} )            { $capacity{pool}{$id}{raw_capacity_used} = $var->{'Hruba pouzita system kapacita'}; }
          #if ( exists $var->{'Formatovana pouzita system kapacita'} )      { $capacity{pool}{$id}{} = $var->{'Formatovana pouzita system kapacita'}; }

          if ( exists $capacity{pool}{$id}{raw_capacity_total} && Xormon::isdigit($capacity{pool}{$id}{raw_capacity_total}) && exists $capacity{pool}{$id}{raw_capacity_used} && Xormon::isdigit($capacity{pool}{$id}{raw_capacity_used}) ) {
            $capacity{pool}{$id}{raw_capacity_free} = Xormon::formatNumberByType("int",$capacity{pool}{$id}{raw_capacity_total} - $capacity{pool}{$id}{raw_capacity_used});
          }
          if ( exists $capacity{pool}{$id}{raw_raid_capacity_total} && Xormon::isdigit($capacity{pool}{$id}{raw_raid_capacity_total}) && exists $capacity{pool}{$id}{raw_raid_capacity_free} && Xormon::isdigit($capacity{pool}{$id}{raw_raid_capacity_free}) ) {
            $capacity{pool}{$id}{raw_raid_capacity_used} = Xormon::formatNumberByType("int",$capacity{pool}{$id}{raw_raid_capacity_total} - $capacity{pool}{$id}{raw_raid_capacity_free});
          }
          if ( exists $capacity{pool}{$id}{format_raid_capacity_total} && Xormon::isdigit($capacity{pool}{$id}{format_raid_capacity_total}) && exists $capacity{pool}{$id}{format_raid_capacity_free} && Xormon::isdigit($capacity{pool}{$id}{format_raid_capacity_free}) ) {
            $capacity{pool}{$id}{format_raid_capacity_used} = Xormon::formatNumberByType("int",$capacity{pool}{$id}{format_raid_capacity_total} - $capacity{pool}{$id}{format_raid_capacity_free});
          }

        }
      }
    }
  }

  return 1;
}

sub parse_cap {
  my $cap_file = shift;

  open( IN, "<$cap_file" ) || $logger->error( "Couldn't open file $cap_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  my @lines = <IN>;
  close(IN);

  my %data_in;
  my $section    = "";
  my $act_header = "";
  my $stat_time  = "";

  my $line_idx = 0;
  foreach my $line (@lines) {
    chomp $line;
    $line_idx++;

    #print $line."\n";

    $line =~ s/^\s+//g;
    $line =~ s/\s+$//g;

    if ( $line =~ m/^----------/ )          { next; }

    if ( $line =~ m/^ULOZISTE/ )           { $section = "device";  next; }
    if ( $line =~ m/^Kapacita uloziste/ )  { $section = "storage"; next; }
    if ( $line =~ m/^Kapacita data pool/ ) { $section = "pool";    next; }
    if ( $line =~ m/^Kapacita Volumu/ )    { $section = "volume";  next; }

    if ( $line =~ m/^Jmeno uloziste,Kapacitni vrstva,/ ) { $act_header = $line; next; }
    if ( $line =~ m/^Cislo poolu,Jmeno poolu,/ )         { $act_header = $line; next; }
    if ( $line =~ m/^Cislo volumu,Jmeno volumu,/ )       { $act_header = $line; next; }

    # clear out type when empty line
    if ( $line =~ m/^$/ && $section ne "device" ) { $section = ""; $act_header = ""; }

    if ( $section eq "device" ) {
      #my ( $key, $value ) = split( /\s+/, $line );
      #if ( defined $key && $key ne '' && defined $value && $value ne '' ) {
      #  $data_in{$stat_time}{$section}{$key} = $value;
      #}
      if ( $line =~ m/^Date:\s+/ ) {
        # Date:           19.04.2022 16:37:33   --> cap file
        # Interval End:     2022-05-10 15:40:59 --> perf file
        $line =~ s/^Date:\s+//;
        my ( $date, $time ) = split( /\s+/, $line );
        my ( $day,  $month, $year ) = split( /\./, $date );
        my ( $hour, $min,   $sec )  = split( /:/,  $time );
        $stat_time = mktime( $sec, $min, $hour, $day, $month - 1, $year - 1900 );
        #print "$line : $stat_time : ".localtime($stat_time)."\n";
        $stat_time = int( $stat_time / $sample_rate ) * $sample_rate; # just set 5 minutes interval
      }
    }
    elsif ( defined $section && $section ne '' && defined $stat_time && $stat_time ne '' ) {
      $data_in{$stat_time}{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }
  }
  #print Dumper \%data_in;

  return \%data_in;
}

sub prepare_perf {
  my $data_in = shift;

  my %data_out;

  # load translator (id to uuid, etc.)
  my $translator;
  if ( -f "$wrkdir/translator.txt" ) {
    $translator = Xormon::read_json("$wrkdir/translator.txt");
  }

  foreach my $timestamp ( sort { $a <=> $b } keys %{ $data_in } ) {
    $last_perf_timestamp = $timestamp;
    #
    # VOLUME
    #
    if ( exists $data_in->{$timestamp}{volume} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{volume} } ) {
        my $var = $data_in->{$timestamp}{volume}{$line_idx};

        if ( exists $var->{'Volume ID'} && exists $translator->{volume}{ $var->{'Volume ID'} } ) {
          my $id = $translator->{volume}{ $var->{'Volume ID'} };

          if ( exists $var->{'Read IO Rate (IO/s)'} )     { $data_out{swiz}{volume}{$id}{$timestamp}{read_io}        = Xormon::formatNumberByType("float4",$var->{'Read IO Rate (IO/s)'}); }
          if ( exists $var->{'Write IO Rate (IO/s)'} )    { $data_out{swiz}{volume}{$id}{$timestamp}{write_io}       = Xormon::formatNumberByType("float4",$var->{'Write IO Rate (IO/s)'}); }
          if ( exists $var->{'Total IO Rate (IO/s)'} )    { $data_out{swiz}{volume}{$id}{$timestamp}{io_rate}        = Xormon::formatNumberByType("float4",$var->{'Total IO Rate (IO/s)'}); }
          if ( exists $var->{'Read Data Rate (KB/s)'} )   { $data_out{swiz}{volume}{$id}{$timestamp}{read}           = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Read Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Write Data Rate (KB/s)'} )  { $data_out{swiz}{volume}{$id}{$timestamp}{write}          = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Write Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Total Data Rate (KB/s)'} )  { $data_out{swiz}{volume}{$id}{$timestamp}{data_rate}      = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Total Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Read Response Time (ms)'} ) { $data_out{swiz}{volume}{$id}{$timestamp}{resp_t_r}       = Xormon::formatNumberByType("float4",$var->{'Read Response Time (ms)'}); }
          if ( exists $var->{'Write Response Time (ms)'} ){ $data_out{swiz}{volume}{$id}{$timestamp}{resp_t_w}       = Xormon::formatNumberByType("float4",$var->{'Write Response Time (ms)'}); }
          if ( exists $var->{'Capacity (B)'} )            { $data_out{swiz}{volume}{$id}{$timestamp}{capacity_total} = Xormon::formatNumberByType("int",$var->{'Capacity (B)'}); }
          if ( exists $var->{'Real Capacity (B)'} )       { $data_out{swiz}{volume}{$id}{$timestamp}{capacity_real}  = Xormon::formatNumberByType("int",$var->{'Real Capacity (B)'}); }
          if ( exists $var->{'Used Capacity (B)'} )       { $data_out{swiz}{volume}{$id}{$timestamp}{capacity_used}  = Xormon::formatNumberByType("int",$var->{'Used Capacity (B)'}); }

          my $vol_stats = $data_out{swiz}{volume}{$id}{$timestamp};

          # calculate capacity_free, capacity_used_percent
          if ( exists $data_out{swiz}{volume}{$id}{$timestamp}{capacity_total} && Xormon::isdigit($data_out{swiz}{volume}{$id}{$timestamp}{capacity_total}) && exists $data_out{swiz}{volume}{$id}{$timestamp}{capacity_used} && Xormon::isdigit($data_out{swiz}{volume}{$id}{$timestamp}{capacity_used}) ) {
            $data_out{swiz}{volume}{$id}{$timestamp}{capacity_free}         = $data_out{swiz}{volume}{$id}{$timestamp}{capacity_total} - $data_out{swiz}{volume}{$id}{$timestamp}{capacity_used};
            if ( $data_out{swiz}{volume}{$id}{$timestamp}{capacity_total} > 0 ) {
              $data_out{swiz}{volume}{$id}{$timestamp}{capacity_used_percent} = $data_out{swiz}{volume}{$id}{$timestamp}{capacity_used} / ( $data_out{swiz}{volume}{$id}{$timestamp}{capacity_total} / 100 );
            }
          }
          # calculate resp_t
          if ( exists $vol_stats->{resp_t_r} && Xormon::isdigit($vol_stats->{resp_t_r}) && exists $vol_stats->{resp_t_w} && Xormon::isdigit($vol_stats->{resp_t_w}) && exists $vol_stats->{read_io} && Xormon::isdigit($vol_stats->{read_io}) && exists $vol_stats->{write_io} && Xormon::isdigit($vol_stats->{write_io}) ) {
            my $resp_t = Xormon::calculate_weighted_avg( [ { count => $vol_stats->{read_io}, value => $vol_stats->{resp_t_r} }, { count => $vol_stats->{write_io}, value => $vol_stats->{resp_t_w} } ] );
            if ( defined $resp_t ) { $data_out{swiz}{volume}{$id}{$timestamp}{resp_t} = Xormon::formatNumberByType("float4",$resp_t); }
          }
          # calculate block size
          if ( exists $vol_stats->{io_rate} && Xormon::isdigit($vol_stats->{io_rate}) && exists $vol_stats->{data_rate} ) {
            if ( $vol_stats->{io_rate} == 0 ) {
              $data_out{swiz}{volume}{$id}{$timestamp}{block_size} = 0;
            }
            else {
              $data_out{swiz}{volume}{$id}{$timestamp}{block_size} = Xormon::formatNumberByType("float4",$vol_stats->{data_rate} / $vol_stats->{io_rate});
            }
          }
          if ( exists $vol_stats->{read_io} && Xormon::isdigit($vol_stats->{read_io}) && exists $vol_stats->{read} ) {
            if ( $vol_stats->{read_io} == 0 ) {
              $data_out{swiz}{volume}{$id}{$timestamp}{block_size_r} = 0;
            }
            else {
              $data_out{swiz}{volume}{$id}{$timestamp}{block_size_r} = Xormon::formatNumberByType("float4",$vol_stats->{read} / $vol_stats->{read_io});
            }
          }
          if ( exists $vol_stats->{write_io} && Xormon::isdigit($vol_stats->{write_io}) && exists $vol_stats->{write} ) {
            if ( $vol_stats->{write_io} == 0 ) {
              $data_out{swiz}{volume}{$id}{$timestamp}{block_size_w} = 0;
            }
            else {
              $data_out{swiz}{volume}{$id}{$timestamp}{block_size_w} = Xormon::formatNumberByType("float4",$vol_stats->{write} / $vol_stats->{write_io});
            }
          }
          # find cache metrics
          if ( exists $data_in->{$timestamp}{"volume-cache"} ) {
            foreach my $cache_line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{"volume-cache"} } ) {
              my $cache_var = $data_in->{$timestamp}{"volume-cache"}{$cache_line_idx};

              if ( exists $cache_var->{'Volume ID'} && $cache_var->{'Volume ID'} eq $var->{'Volume ID'} ) {
                if ( exists $cache_var->{'Cache read hits (%)'} )    { $data_out{swiz}{volume}{$id}{$timestamp}{r_cache_hit} = Xormon::formatNumberByType("float4",$cache_var->{'Cache read hits (%)'}); }
                if ( exists $cache_var->{'Cache write hits (%)'} )   { $data_out{swiz}{volume}{$id}{$timestamp}{w_cache_hit} = Xormon::formatNumberByType("float4",$cache_var->{'Cache write hits (%)'}); }
                if ( exists $cache_var->{'Total cache usage (kB)'} ) { $data_out{swiz}{volume}{$id}{$timestamp}{cache_usage} = Xormon::formatNumberByType("float4",Xormon::returnBytes($cache_var->{'Total cache usage (kB)'},"kB")); }
                last;
              }
            }
          }

          # add volume uuid to pool
          if ( exists $var->{'Managed Disk Group ID'} ) {
            # there can be 2 or more pools separated by a space
            if ( $var->{'Managed Disk Group ID'} =~ m/\s+/ ) {
              my @pool_ids   = split(/\s+/, $var->{'Managed Disk Group ID'});
              my $pool_count = scalar(@pool_ids);
              foreach my $pool_id (@pool_ids) {
                $data_in->{$timestamp}{pool}{$pool_id}{volumes}{$id} = $pool_count;
              }
            }
            else {
              $data_in->{$timestamp}{pool}{ $var->{'Managed Disk Group ID'} }{volumes}{$id} = 1;
            }
            #push( @{ $data_in->{$timestamp}{pool}{ $var->{'Managed Disk Group ID'} }{volumes} }, $id );
          }
          # add volume uuid to IO group
          if ( exists $var->{'IO Group ID'} ) {
            push( @{ $data_in->{$timestamp}{'io_group'}{ $var->{'IO Group ID'} }{volumes} }, $id );
          }
        }
      }
    }

    #
    # POOL
    #
    if ( exists $data_in->{$timestamp}{pool} ) {
      foreach my $pool_id ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{pool} } ) {
        my $var = $data_in->{$timestamp}{pool}{$pool_id};
        my $id  = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$pool_id) );

        # calculate data from the assigned volumes stats
        #
        #my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate","cache_usage");
        #my @met_avg          = ("r_cache_hit","w_cache_hit");
        my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate");
        my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

        my $values2avg;
        my $values2weighted_avg;

        if ( exists $var->{'volumes'} ) {
          foreach my $volume_id ( keys %{ $var->{'volumes'} } ) {
            my $divider = $var->{'volumes'}{$volume_id}; # if volume belongs to 2 pools then data and IO must be devided by 2
            #print "$pool_id : $volume_id : $divider\n";

            #if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp} ) { # this creates an empty hash when $data_out{swiz}{volume}{$volume_id} doesn't exist
            if ( exists $data_out{swiz}{volume}{$volume_id} && exists $data_out{swiz}{volume}{$volume_id}{$timestamp} ) { # must be like this
              # aggregate stats
              foreach my $met_act ( @met_aggregate ) {
                if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} ) {
                  if ( exists $data_out{swiz}{pool}{$id}{$timestamp}{$met_act} ) {
                    $data_out{swiz}{pool}{$id}{$timestamp}{$met_act} += $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} / $divider;
                  }
                  else {
                    $data_out{swiz}{pool}{$id}{$timestamp}{$met_act} = $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} / $divider;
                  }
                }
              }
              ## avg stats
              #foreach my $met_act ( @met_avg ) {
              #  if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} ) {
              #    push(@{ $values2avg->{$met_act} }, $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act});
              #  }
              #}
              # weighted avg stats
              foreach my $met_act ( @met_weighted_avg ) {
                if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
                  push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
                }
              }
            }
          }
        }

        ## calculate avg
        #foreach my $met_act ( @met_avg ) {
        #  if ( exists $values2avg->{ $met_act } ) {
        #    my $value = Xormon::calculate_avg($values2avg->{ $met_act });
        #    #print "$met_act = $value\n";
        #    if ( defined $value ) { $data_out{swiz}{pool}{$id}{$timestamp}{ $met_act } = $value; }
        #  }
        #}
        # calculate weighted avg
        foreach my $met_act ( @met_weighted_avg ) {
          if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
            my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
            #print "$met_act->{value_key} = $value\n";
            if ( defined $value ) { $data_out{swiz}{pool}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
          }
        }
        # calculate block size
        if ( exists $data_out{swiz}{pool}{$id}{$timestamp}{io_rate} && exists $data_out{swiz}{pool}{$id}{$timestamp}{data_rate} ) {
          if ( $data_out{swiz}{pool}{$id}{$timestamp}{io_rate} == 0 ) {
            $data_out{swiz}{pool}{$id}{$timestamp}{block_size} = 0;
          }
          else {
            $data_out{swiz}{pool}{$id}{$timestamp}{block_size} = $data_out{swiz}{pool}{$id}{$timestamp}{data_rate} / $data_out{swiz}{pool}{$id}{$timestamp}{io_rate};
          }
        }
        if ( exists $data_out{swiz}{pool}{$id}{$timestamp}{read_io} && exists $data_out{swiz}{pool}{$id}{$timestamp}{read} ) {
          if ( $data_out{swiz}{pool}{$id}{$timestamp}{read_io} == 0 ) {
            $data_out{swiz}{pool}{$id}{$timestamp}{block_size_r} = 0;
          }
          else {
            $data_out{swiz}{pool}{$id}{$timestamp}{block_size_r} = $data_out{swiz}{pool}{$id}{$timestamp}{read} / $data_out{swiz}{pool}{$id}{$timestamp}{read_io};
          }
        }
        if ( exists $data_out{swiz}{pool}{$id}{$timestamp}{write_io} && exists $data_out{swiz}{pool}{$id}{$timestamp}{write} ) {
          if ( $data_out{swiz}{pool}{$id}{$timestamp}{write_io} == 0 ) {
            $data_out{swiz}{pool}{$id}{$timestamp}{block_size_w} = 0;
          }
          else {
            $data_out{swiz}{pool}{$id}{$timestamp}{block_size_w} = $data_out{swiz}{pool}{$id}{$timestamp}{write} / $data_out{swiz}{pool}{$id}{$timestamp}{write_io};
          }
        }
        # format aggregated values
        foreach my $metric_key ( sort keys %{ $data_out{swiz}{pool}{$id}{$timestamp} } ) {
          #if ( $metric_key =~ m/^resp_t/ || $metric_key =~ m/_cache_hit$/ ) {
          #  $data_out{swiz}{pool}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{pool}{$id}{$timestamp}{$metric_key});
          #}
          #else {
          #  $data_out{swiz}{pool}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{pool}{$id}{$timestamp}{$metric_key});
          #}
          $data_out{swiz}{pool}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{pool}{$id}{$timestamp}{$metric_key});
        }

        # find capacity metrics
        #
        if ( exists $data_in->{$timestamp}{"pool-cap"} ) {
          foreach my $pool_line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{"pool-cap"} } ) {
            my $pool_var = $data_in->{$timestamp}{"pool-cap"}{$pool_line_idx};

            if ( exists $pool_var->{'id'} && $pool_var->{'id'} eq $pool_id ) {
              #print Dumper $pool_var;
              if ( exists $pool_var->{'capacity (TB)'} )         { $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total}         = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'capacity (TB)'},"TiB")); }
              if ( exists $pool_var->{'free capacity (TB)'} )    { $data_out{swiz}{pool}{$id}{$timestamp}{capacity_free}          = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'free capacity (TB)'},"TiB")); }
              if ( exists $pool_var->{'used capacity (TB)'} )    { $data_out{swiz}{pool}{$id}{$timestamp}{capacity_used}          = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'used capacity (TB)'},"TiB")); }
              if ( exists $pool_var->{'real capacity (TB)'} )    { $data_out{swiz}{pool}{$id}{$timestamp}{capacity_real}          = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'real capacity (TB)'},"TiB")); }
              if ( exists $pool_var->{'virtual capacity (TB)'} ) { $data_out{swiz}{pool}{$id}{$timestamp}{virtual_capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'virtual capacity (TB)'},"TiB")); }
              if ( exists $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total} && exists $data_out{swiz}{pool}{$id}{$timestamp}{capacity_used} ) {
                $data_out{swiz}{pool}{$id}{$timestamp}{capacity_used_percent} = $data_out{swiz}{pool}{$id}{$timestamp}{capacity_used} / ( $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total} / 100 );
              }
              if ( exists $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total} && exists $data_out{swiz}{pool}{$id}{$timestamp}{virtual_capacity_total} ) {
                if ( $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total} < $data_out{swiz}{pool}{$id}{$timestamp}{virtual_capacity_total} ) {
                  $data_out{swiz}{pool}{$id}{$timestamp}{capacity_oversubscr}         = $data_out{swiz}{pool}{$id}{$timestamp}{virtual_capacity_total} - $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total};
                  $data_out{swiz}{pool}{$id}{$timestamp}{capacity_oversubscr_percent} = ( $data_out{swiz}{pool}{$id}{$timestamp}{virtual_capacity_total} / $data_out{swiz}{pool}{$id}{$timestamp}{capacity_total} ) * 100;
                }
                else {
                  $data_out{swiz}{pool}{$id}{$timestamp}{capacity_oversubscr}         = 0;
                  $data_out{swiz}{pool}{$id}{$timestamp}{capacity_oversubscr_percent} = 0;
                }
              }
              last;
            }
          }
        }

        # add additional capacity information from the last parsed capacity file
        if ( exists $capacity{pool}{$id} ) {
          # merge 2 hashes
          %{ $data_out{swiz}{pool}{$id}{$timestamp} } = ( %{ $capacity{pool}{$id} }, %{ $data_out{swiz}{pool}{$id}{$timestamp} });
        }

      }
    }

    #
    # IO_GROUP
    #
    if ( exists $data_in->{$timestamp}{io_group} ) {
      foreach my $id ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{io_group} } ) {
        my $var = $data_in->{$timestamp}{io_group}{$id};
        my $id  = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."io_group".$id) );

        # calculate data from the assigned volumes stats
        #
        #my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate","cache_usage");
        #my @met_avg          = ("r_cache_hit","w_cache_hit");
        my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate");
        my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

        my $values2avg;
        my $values2weighted_avg;

        if ( exists $var->{'volumes'} && ref($var->{'volumes'}) eq "ARRAY" ) {
          foreach my $volume_id ( @{ $var->{'volumes'} } ) {
            if ( exists $data_out{swiz}{volume}{$volume_id} && exists $data_out{swiz}{volume}{$volume_id}{$timestamp} ) { # must be like this
              # aggregate stats
              foreach my $met_act ( @met_aggregate ) {
                if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} && Xormon::isdigit($data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act}) ) {
                  if ( exists $data_out{swiz}{io_group}{$id}{$timestamp}{$met_act} ) {
                    $data_out{swiz}{io_group}{$id}{$timestamp}{$met_act} += $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act};
                  }
                  else {
                    $data_out{swiz}{io_group}{$id}{$timestamp}{$met_act} = $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act};
                  }
                }
              }
              ## avg stats
              #foreach my $met_act ( @met_avg ) {
              #  if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} ) {
              #    push(@{ $values2avg->{$met_act} }, $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act});
              #  }
              #}
              # weighted avg stats
              foreach my $met_act ( @met_weighted_avg ) {
                if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
                  push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
                }
              }
            }
          }
        }

        ## calculate avg
        #foreach my $met_act ( @met_avg ) {
        #  if ( exists $values2avg->{ $met_act } ) {
        #    my $value = Xormon::calculate_avg($values2avg->{ $met_act });
        #    #print "$met_act = $value\n";
        #    if ( defined $value ) { $data_out{swiz}{io_group}{$id}{$timestamp}{ $met_act } = $value; }
        #  }
        #}
        # calculate weighted avg
        foreach my $met_act ( @met_weighted_avg ) {
          if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
            my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
            #print "$met_act->{value_key} = $value\n";
            if ( defined $value ) { $data_out{swiz}{io_group}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
          }
        }
        # format aggregated values
        foreach my $metric_key ( sort keys %{ $data_out{swiz}{io_group}{$id}{$timestamp} } ) {
          #if ( $metric_key =~ m/^resp_t/ || $metric_key =~ m/_cache_hit$/ ) {
          #  $data_out{swiz}{io_group}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{io_group}{$id}{$timestamp}{$metric_key});
          #}
          #else {
          #  $data_out{swiz}{io_group}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{io_group}{$id}{$timestamp}{$metric_key});
          #}
          $data_out{swiz}{io_group}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{io_group}{$id}{$timestamp}{$metric_key});
        }
      }
    }

    #
    # DEVICE
    #
    if ( exists $data_out{swiz}{volume} ) {
      my $id = $cfg->{'hostcfg_id'};

      # calculate data from the volumes stats
      #
      my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate","cache_usage");
      my @met_avg          = ("r_cache_hit","w_cache_hit");
      my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

      my $values2avg;
      my $values2weighted_avg;

      foreach my $volume_id ( %{ $data_out{swiz}{volume} } ) {
        if ( exists $data_out{swiz}{volume}{$volume_id} && exists $data_out{swiz}{volume}{$volume_id}{$timestamp} ) { # must be like this
          # aggregate stats
          foreach my $met_act ( @met_aggregate ) {
            if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} && Xormon::isdigit($data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act}) ) {
              if ( exists $data_out{swiz}{device}{$id}{$timestamp}{$met_act} ) {
                $data_out{swiz}{device}{$id}{$timestamp}{$met_act} += $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act};
              }
              else {
                $data_out{swiz}{device}{$id}{$timestamp}{$met_act} = $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act};
              }
            }
          }
          # avg stats
          foreach my $met_act ( @met_avg ) {
            if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} ) {
              push(@{ $values2avg->{$met_act} }, $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act});
            }
          }
          # weighted avg stats
          foreach my $met_act ( @met_weighted_avg ) {
            if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
              push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
            }
          }
        }
      }

      # calculate avg
      foreach my $met_act ( @met_avg ) {
        if ( exists $values2avg->{ $met_act } ) {
          my $value = Xormon::calculate_avg($values2avg->{ $met_act });
          #print "$met_act = $value\n";
          if ( defined $value ) { $data_out{swiz}{device}{$id}{$timestamp}{ $met_act } = $value; }
        }
      }
      # calculate weighted avg
      foreach my $met_act ( @met_weighted_avg ) {
        if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
          my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
          #print "$met_act->{value_key} = $value\n";
          if ( defined $value ) { $data_out{swiz}{device}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
        }
      }
      # format aggregated values
      foreach my $metric_key ( sort keys %{ $data_out{swiz}{device}{$id}{$timestamp} } ) {
        #if ( $metric_key =~ m/^resp_t/ || $metric_key =~ m/_cache_hit$/ ) {
        #  $data_out{swiz}{device}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{device}{$id}{$timestamp}{$metric_key});
        #}
        #else {
        #  $data_out{swiz}{device}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{device}{$id}{$timestamp}{$metric_key});
        #}
        $data_out{swiz}{device}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{device}{$id}{$timestamp}{$metric_key});
      }

      # aggregate capacity metrics from all pools
      #
      if ( exists $data_in->{$timestamp}{"pool-cap"} ) {
        foreach my $pool_line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{"pool-cap"} } ) {
          my $pool_var = $data_in->{$timestamp}{"pool-cap"}{$pool_line_idx};

          if ( exists $pool_var->{'capacity (TB)'} && $pool_var->{'capacity (TB)'} ne '' ) {
            if ( exists $data_out{swiz}{device}{$id}{$timestamp}{capacity_total} ) {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_total} += Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'capacity (TB)'},"TiB"));
            }
            else {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'capacity (TB)'},"TiB"));
            }
          }
          if ( exists $pool_var->{'free capacity (TB)'} && $pool_var->{'free capacity (TB)'} ne '' ) {
            if ( exists $data_out{swiz}{device}{$id}{$timestamp}{capacity_free} ) {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_free} += Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'free capacity (TB)'},"TiB"));
            }
            else {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_free} = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'free capacity (TB)'},"TiB"));
            }
          }
          if ( exists $pool_var->{'used capacity (TB)'} && $pool_var->{'used capacity (TB)'} ne '' ) {
            if ( exists $data_out{swiz}{device}{$id}{$timestamp}{capacity_used} ) {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_used} += Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'used capacity (TB)'},"TiB"));
            }
            else {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_used} = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'used capacity (TB)'},"TiB"));
            }
          }
          if ( exists $data_out{swiz}{device}{$id}{$timestamp}{capacity_total} && $data_out{swiz}{device}{$id}{$timestamp}{capacity_total} > 0 && exists $data_out{swiz}{device}{$id}{$timestamp}{capacity_used} ) {
            $data_out{swiz}{device}{$id}{$timestamp}{capacity_used_percent} = $data_out{swiz}{device}{$id}{$timestamp}{capacity_used} / ( $data_out{swiz}{device}{$id}{$timestamp}{capacity_total} / 100 );
          }
          if ( exists $pool_var->{'real capacity (TB)'} && $pool_var->{'real capacity (TB)'} ne '' ) {
            if ( exists $data_out{swiz}{device}{$id}{$timestamp}{capacity_real} ) {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_real} += Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'real capacity (TB)'},"TiB"));
            }
            else {
              $data_out{swiz}{device}{$id}{$timestamp}{capacity_real} = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'real capacity (TB)'},"TiB"));
            }
          }
          if ( exists $pool_var->{'virtual capacity (TB)'} && $pool_var->{'virtual capacity (TB)'} ne '' ) {
            if ( exists $data_out{swiz}{device}{$id}{$timestamp}{virtual_capacity_total} ) {
              $data_out{swiz}{device}{$id}{$timestamp}{virtual_capacity_total} += Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'virtual capacity (TB)'},"TiB"));
            }
            else {
              $data_out{swiz}{device}{$id}{$timestamp}{virtual_capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($pool_var->{'virtual capacity (TB)'},"TiB"));
            }
          }
        }
      }

      # add additional capacity information from the last parsed capacity file
      if ( exists $capacity{device}{$id} ) {
        # merge 2 hashes
        %{ $data_out{swiz}{device}{$id}{$timestamp} } = ( %{ $capacity{device}{$id} }, %{ $data_out{swiz}{device}{$id}{$timestamp} });
      }
    }

    #
    # NODE (CPU-NODE)
    #
    if ( exists $data_in->{$timestamp}{'cpu-node'} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{'cpu-node'} } ) {
        my $var = $data_in->{$timestamp}{'cpu-node'}{$line_idx};
        #print Dumper $var;

        if ( exists $var->{'Node'} && exists $translator->{'cpu-node'}{ $var->{'Node'} } ) {
          my $id = $translator->{'cpu-node'}{ $var->{'Node'} };

          #if ( exists $var->{'CPU Busy'}                      && Xormon::isdigit($var->{'CPU Busy'}) )                      { $data_out{SWIZ}{'node'}{$id}{$timestamp}{'bussy'}    = $var->{'CPU Busy'}; } # unused
          #if ( exists $var->{'CPU Limited'}                   && Xormon::isdigit($var->{'CPU Limited'}) )                   { $data_out{swiz}{'node'}{$id}{$timestamp}{'lim'}      = $var->{'CPU Limited'}; } # unused
          if ( exists $var->{'CPU Utilization - System'} )      { $data_out{swiz}{'node'}{$id}{$timestamp}{'cpu_system'}   = Xormon::formatNumberByType("float4",$var->{'CPU Utilization - System'}); }
          if ( exists $var->{'CPU Utilization - Compression'} ) { $data_out{swiz}{'node'}{$id}{$timestamp}{'cpu_compress'} = Xormon::formatNumberByType("float4",$var->{'CPU Utilization - Compression'}); }
        }
      }
    }

    #
    # PORT
    #
    if ( exists $data_in->{$timestamp}{'port'} ) {
      my $ports;
      my $port_idx = 0;
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{'port'} } ) {
        my $var = $data_in->{$timestamp}{'port'}{$line_idx};
        #print Dumper $var; next;

        if ( exists $var->{'Port'} && $var->{'Port'} ne '' ) {
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."port".$var->{'Port'}) );

          $port_idx++;
          $ports->{ $var->{'Port'} }{id} = $id;
          if ( exists $var->{'Port type'} ) {
            $ports->{ $var->{'Port'} }{port_type} = $var->{'Port type'};

            # stats available only for FC ports
            if ( $var->{'Port type'} eq "FC" ) {
              if ( exists $var->{'Invalid CRC Count'} )                   { $data_out{swiz}{port}{$id}{$timestamp}{'crc'} = Xormon::formatNumberByType("float4",$var->{'Invalid CRC Count'}); }
              if ( exists $var->{'Zero b2b / Delayed Transmission (%)'} ) { $data_out{swiz}{port}{$id}{$timestamp}{'b2b'} = Xormon::formatNumberByType("float4",$var->{'Zero b2b / Delayed Transmission (%)'}); }
            }
          }

          if ( exists $var->{'Total Port IO Rate (IO/s)'} )     { $data_out{swiz}{port}{$id}{$timestamp}{'io_rate'}   = Xormon::formatNumberByType("float4",$var->{'Total Port IO Rate (IO/s)'}); }
          if ( exists $var->{'Port Receive IO Rate (IO/s)'} )   { $data_out{swiz}{port}{$id}{$timestamp}{'read_io'}   = Xormon::formatNumberByType("float4",$var->{'Port Receive IO Rate (IO/s)'}); }
          if ( exists $var->{'Port Send IO Rate (IO/s)'} )      { $data_out{swiz}{port}{$id}{$timestamp}{'write_io'}  = Xormon::formatNumberByType("float4",$var->{'Port Send IO Rate (IO/s)'}); }
          if ( exists $var->{'Total Port Data Rate (KB/s)'} )   { $data_out{swiz}{port}{$id}{$timestamp}{'data_rate'} = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Total Port Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Port Receive Data Rate (KB/s)'} ) { $data_out{swiz}{port}{$id}{$timestamp}{'read'}      = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Port Receive Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Port Send Data Rate (KB/s)'} )    { $data_out{swiz}{port}{$id}{$timestamp}{'write'}     = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Port Send Data Rate (KB/s)'},"kB")); }
        }
      }

      # save port info to json file for the further use
      # it's necessary bcs there are not all ports presented in the conf files (only FC ports are in conf files)
      if ( $port_idx > 0 ) {
        Xormon::save_as_json("$wrkdir/ports.json",$ports);
      }
    }

    #
    # DRIVE
    #
    if ( exists $data_in->{$timestamp}{'drive'} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{'drive'} } ) {
        my $var = $data_in->{$timestamp}{'drive'}{$line_idx};
        #print Dumper $var; next;

        if ( exists $var->{'Drive'} && $var->{'Drive'} ne '' ) {
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."drive".$var->{Drive}) );

          if ( exists $var->{'Total IO Rate (IO/s)'} )             { $data_out{swiz}{drive}{$id}{$timestamp}{'io_rate'}   = Xormon::formatNumberByType("float4",$var->{'Total IO Rate (IO/s)'}); }
          if ( exists $var->{'Read IO Rate (IO/s)'} )              { $data_out{swiz}{drive}{$id}{$timestamp}{'read_io'}   = Xormon::formatNumberByType("float4",$var->{'Read IO Rate (IO/s)'}); }
          if ( exists $var->{'Write IO Rate (IO/s)'} )             { $data_out{swiz}{drive}{$id}{$timestamp}{'write_io'}  = Xormon::formatNumberByType("float4",$var->{'Write IO Rate (IO/s)'}); }
          if ( exists $var->{'Total Data Rate (KB/s)'} )           { $data_out{swiz}{drive}{$id}{$timestamp}{'data_rate'} = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Total Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Read Data Rate (KB/s)'} )            { $data_out{swiz}{drive}{$id}{$timestamp}{'read'}      = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Read Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Write Data Rate (KB/s)'} )           { $data_out{swiz}{drive}{$id}{$timestamp}{'write'}     = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Write Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Backend Read Response Time (ms)'} )  { $data_out{swiz}{drive}{$id}{$timestamp}{'resp_t_r'}  = Xormon::formatNumberByType("float4",$var->{'Backend Read Response Time (ms)'}); }
          if ( exists $var->{'Backend Write Response Time (ms)'} ) { $data_out{swiz}{drive}{$id}{$timestamp}{'resp_t_w'}  = Xormon::formatNumberByType("float4",$var->{'Backend Write Response Time (ms)'}); }

          # calculate resp_t
          if ( exists $data_out{swiz}{drive}{$id}{$timestamp}{resp_t_r} && exists $data_out{swiz}{drive}{$id}{$timestamp}{resp_t_w} && exists $data_out{swiz}{drive}{$id}{$timestamp}{read_io} && exists $data_out{swiz}{drive}{$id}{$timestamp}{write_io} ) {
            my $resp_t = Xormon::calculate_weighted_avg( [ { count => $data_out{swiz}{drive}{$id}{$timestamp}{read_io}, value => $data_out{swiz}{drive}{$id}{$timestamp}{resp_t_r} }, { count => $data_out{swiz}{drive}{$id}{$timestamp}{write_io}, value => $data_out{swiz}{drive}{$id}{$timestamp}{resp_t_w} } ] );
            if ( defined $resp_t ) { $data_out{swiz}{drive}{$id}{$timestamp}{resp_t} = Xormon::formatNumberByType("float4",$resp_t); }
          }
        }
      }
    }

    #
    # MDISK (RANK)
    #
    if ( exists $data_in->{$timestamp}{'rank'} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{'rank'} } ) {
        my $var = $data_in->{$timestamp}{'rank'}{$line_idx};
        #print Dumper $var; next;

        if ( exists $var->{'Managed Disk ID'} && $var->{'Managed Disk ID'} ne '' ) {
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."mdisk".$var->{'Managed Disk ID'}) );

          if ( exists $var->{'Total IO Rate (IO/s)'} )             { $data_out{swiz}{mdisk}{$id}{$timestamp}{'io_rate'}   = Xormon::formatNumberByType("float4",$var->{'Total IO Rate (IO/s)'}); }
          if ( exists $var->{'Read IO Rate (IO/s)'} )              { $data_out{swiz}{mdisk}{$id}{$timestamp}{'read_io'}   = Xormon::formatNumberByType("float4",$var->{'Read IO Rate (IO/s)'}); }
          if ( exists $var->{'Write IO Rate (IO/s)'} )             { $data_out{swiz}{mdisk}{$id}{$timestamp}{'write_io'}  = Xormon::formatNumberByType("float4",$var->{'Write IO Rate (IO/s)'}); }
          if ( exists $var->{'Total Data Rate (KB/s)'} )           { $data_out{swiz}{mdisk}{$id}{$timestamp}{'data_rate'} = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Total Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Read Data Rate (KB/s)'} )            { $data_out{swiz}{mdisk}{$id}{$timestamp}{'read'}      = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Read Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Write Data Rate (KB/s)'} )           { $data_out{swiz}{mdisk}{$id}{$timestamp}{'write'}     = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Write Data Rate (KB/s)'},"kB")); }
          if ( exists $var->{'Backend Read Response Time (ms)'} )  { $data_out{swiz}{mdisk}{$id}{$timestamp}{'resp_t_r'}  = Xormon::formatNumberByType("float4",$var->{'Backend Read Response Time (ms)'}); }
          if ( exists $var->{'Backend Write Response Time (ms)'} ) { $data_out{swiz}{mdisk}{$id}{$timestamp}{'resp_t_w'}  = Xormon::formatNumberByType("float4",$var->{'Backend Write Response Time (ms)'}); }

          # calculate resp_t
          if ( exists $data_out{swiz}{mdisk}{$id}{$timestamp}{resp_t_r} && exists $data_out{swiz}{mdisk}{$id}{$timestamp}{resp_t_w} && exists $data_out{swiz}{mdisk}{$id}{$timestamp}{read_io} && exists $data_out{swiz}{mdisk}{$id}{$timestamp}{write_io} ) {
            my $resp_t = Xormon::calculate_weighted_avg( [ { count => $data_out{swiz}{mdisk}{$id}{$timestamp}{read_io}, value => $data_out{swiz}{mdisk}{$id}{$timestamp}{resp_t_r} }, { count => $data_out{swiz}{mdisk}{$id}{$timestamp}{write_io}, value => $data_out{swiz}{mdisk}{$id}{$timestamp}{resp_t_w} } ] );
            if ( defined $resp_t ) { $data_out{swiz}{mdisk}{$id}{$timestamp}{resp_t} = Xormon::formatNumberByType("float4",$resp_t); }
          }
        }
      }
    }

    #
    # CPU (CPU-CORE)
    #
    if ( exists $data_in->{$timestamp}{'cpu-core'} ) {
      my $cpu;
      my $cpu_idx = 0;
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{'cpu-core'} } ) {
        my $var = $data_in->{$timestamp}{'cpu-core'}{$line_idx};
        #print Dumper $var; next;

        if ( exists $var->{'CPU Core'} && $var->{'CPU Core'} ne '' ) {
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."cpu".$var->{'CPU Core'}) );

          $cpu_idx++;
          $cpu->{ $var->{'CPU Core'} }{id} = $id;

          if ( exists $var->{'CPU Utilization - System'} )      { $data_out{swiz}{'cpu'}{$id}{$timestamp}{'cpu_system'}   = Xormon::formatNumberByType("float4",$var->{'CPU Utilization - System'}); }
          if ( exists $var->{'CPU Utilization - Compression'} ) { $data_out{swiz}{'cpu'}{$id}{$timestamp}{'cpu_compress'} = Xormon::formatNumberByType("float4",$var->{'CPU Utilization - Compression'}); }
        }
      }

      # save port info to json file for the further use
      # it's necessary bcs there are not all ports presented in the conf files (only FC ports are in conf files)
      if ( $cpu_idx > 0 ) {
        Xormon::save_as_json("$wrkdir/cpu.json",$cpu);
      }
    }

    #
    # TIER (only capacity stats)
    #
    # add capacity information from the last parsed capacity file
    if ( exists $capacity{tier} ) {
      foreach my $id ( %{ $capacity{tier} } ) {
        if ( exists $capacity{tier}{$id} && scalar( keys(%{ $capacity{tier}{$id} }) ) ) {
          $data_out{swiz}{tier}{$id}{$timestamp} = $capacity{tier}{$id};
        }
      }
    }

    #
    # HOST (aggregated stats from volumes)
    #
    if ( -f "$wrkdir/host.json" && Xormon::file_time_diff("$wrkdir/host.json") < 86400 ) {
      my %hosts = %{ Xormon::read_json("$wrkdir/host.json") };
      foreach ( sort keys %hosts ) {
        my $var = $hosts{$_};
        #print Dumper $var;
        if ( exists $var->{host_id} && exists $var->{name}) {
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."host".$var->{host_id}) );
          if ( exists $var->{'Volume IDs'} ) {
            # calculate data from the assigned volumes stats
            #
            my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate");
            my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

            my $values2avg;
            my $values2weighted_avg;

            foreach my $volume_idx ( split(/\s+/, $var->{'Volume IDs'}) ) {
              if ( exists $translator->{volume}{$volume_idx} ) {
                my $volume_id = $translator->{volume}{$volume_idx};

                if ( exists $data_out{swiz}{volume}{$volume_id} && exists $data_out{swiz}{volume}{$volume_id}{$timestamp} ) { # must be like this
                  # aggregate stats
                  foreach my $met_act ( @met_aggregate ) {
                    if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} && Xormon::isdigit($data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act}) ) {
                      if ( exists $data_out{swiz}{host}{$id}{$timestamp}{$met_act} ) {
                        $data_out{swiz}{host}{$id}{$timestamp}{$met_act} += $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act};
                      }
                      else {
                        $data_out{swiz}{host}{$id}{$timestamp}{$met_act} = $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act};
                      }
                    }
                  }
                  ## avg stats
                  #foreach my $met_act ( @met_avg ) {
                  #  if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act} ) {
                  #    push(@{ $values2avg->{$met_act} }, $data_out{swiz}{volume}{$volume_id}{$timestamp}{$met_act});
                  #  }
                  #}
                  # weighted avg stats
                  foreach my $met_act ( @met_weighted_avg ) {
                    if ( exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
                      push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{swiz}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
                    }
                  }
                }
              }
            }

            # calculate weighted avg
            foreach my $met_act ( @met_weighted_avg ) {
              if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
                my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
                #print "$met_act->{value_key} = $value\n";
                if ( defined $value ) { $data_out{swiz}{host}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
              }
            }
            # format aggregated values
            foreach my $metric_key ( sort keys %{ $data_out{swiz}{host}{$id}{$timestamp} } ) {
              $data_out{swiz}{host}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{swiz}{host}{$id}{$timestamp}{$metric_key});
            }
          }
        }
      }
    }


  }

  #print "#################### PERF\n";
  #print Dumper \%data_out;
  #print Dumper \%{$data_out{swiz}{host}};


  return \%data_out;
}

sub parse_perf {
  my $perf_file = shift;

  open( IN, "<$perf_file" ) || $logger->error( "Couldn't open file $perf_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  my @lines = <IN>;
  close(IN);

  my %data_in;
  my $section    = "";
  my $act_header = "";
  my $stat_time  = "";

  my $line_idx = 0;
  foreach my $line (@lines) {
    chomp $line;
    $line_idx++;

    #print $line."\n";

    $line =~ s/^\s+//g;
    $line =~ s/\s+$//g;

    if ( $line =~ m/^----------/ )          { next; }

    if ( $line =~ m/Pool Capacity Statistics/ )          { $section = "pool-cap";      next; }
    if ( $line =~ m/Volume Cache Level Statistics/ )     { $section = "volume-cache";  next; }
    if ( $line =~ m/Managed Disk Level Statistics/ )     { $section = "rank";          next; }
    if ( $line =~ m/Port Level Statistics/ )             { $section = "port";          next; }
    if ( $line =~ m/Volume Level Statistics/ )           { $section = "volume";        next; }
    if ( $line =~ m/Drive Level Statistics/ )            { $section = "drive";         next; }
    if ( $line =~ m/CPU-Node Level Statistics/ )         { $section = "cpu-node";      next; }
    if ( $line =~ m/CPU-Core Level Statistics/ )         { $section = "cpu-core";      next; }

    if ( $line =~ m/^Port,Time,Interval / )            { $act_header = $line; next; }
    if ( $line =~ m/^Managed Disk ID,Time,Interval / ) { $act_header = $line; next; }
    if ( $line =~ m/^Volume ID,Time,Interval / )       { $act_header = $line; next; }
    if ( $line =~ m/^Drive,Time,Interval / )           { $act_header = $line; next; }
    if ( $line =~ m/^Node,Time,Interval/ )             { $act_header = $line; next; }
    if ( $line =~ m/^CPU Core,Time,Interval/ )         { $act_header = $line; next; }
    if ( $line =~ m/^name,id,/ )                       { $act_header = $line; next; }

    if ( $line =~ m/Interval End:/ ) {
      $line =~ s/Interval End:\s+//;
      $line =~ s/GMT.*//;
      $stat_time = str2time($line);
      unless ( Xormon::isdigit($stat_time) ) { $stat_time = ""; }
      #print "####### $sevtion : $stat_time : ".localtime($stat_time)."\n";
      next;
    }
    # clear out type when empty line
    if ( $line =~ m/^$/ ) { $section = ""; $act_header = ""; }
    # avoid other trash
    #if ( $line =~ /^name,id/ || $line =~ m/^>-Interval/ || $line =~ m/^-----------/ || $line =~ m/^Node,Time/ || $line =~ m/^Node,cfav/ || $line =~ m/^CPU Core,Time/ || $line =~ m/^Port,Time/ || $line =~ m/^Drive,Time/ || $line =~ m/^Managed Disk ID,Time/ || $line =~ m/^Volume ID,Time/ ) {
    if ( $line =~ m/^Interval\s+/ ) { next; }

    if ( $section eq "port" || $section eq "rank" || $section eq "volume" || $section eq "drive"|| $section eq "cpu-node" || $section eq "cpu-core" || $section eq "pool-cap" || $section eq "volume-cache" ) {
      if ( defined $stat_time && $stat_time ne '' ) {
        $data_in{$stat_time}{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
      }
    }
  }
  #print Dumper \%data_in;

  return \%data_in;
}

sub send_conf {
  my $data = shift;

  my @architecture;
  my @volumes_arch;
  my %conf;
  my $io_group;
  my $cluster;
  my $host;

  #
  # DEVICE
  #
  # Architecture
  my %device_arch = (
    "item_id"    => $cfg->{'hostcfg_id'},
    "hostcfg_id" => $cfg->{'hostcfg_id'},
    "label"      => $cfg->{'hostalias'},
    "hw_type"    => "swiz",
    "subsystem"  => "device",
    "parents"    => []
  );
  push(@architecture, \%device_arch);
  # conf
  if ( exists $data->{device}{'Machine Serial'} )                  { $conf{ $cfg->{'hostcfg_id'} }{'serial'}                          = $data->{device}{'Machine Serial'}; }
  if ( exists $data->{device}{'Machine Type-Model'} )              { $conf{ $cfg->{'hostcfg_id'} }{'model'}                           = $data->{device}{'Machine Type-Model'}; }
  if ( exists $data->{device}{'Code Level'} )                      { $conf{ $cfg->{'hostcfg_id'} }{'version'}                         = $data->{device}{'Code Level'}; }
  if ( exists $data->{device}{'Machine Name'} )                    { $conf{ $cfg->{'hostcfg_id'} }{'machine_name'}                    = $data->{device}{'Machine Name'}; }
  if ( exists $data->{device}{'Product Name'} )                    { $conf{ $cfg->{'hostcfg_id'} }{'product_name'}                    = $data->{device}{'Product Name'}; }
  if ( exists $data->{device}{'Performance Statistic Status'} )    { $conf{ $cfg->{'hostcfg_id'} }{'performance_statistic_status'}    = $data->{device}{'Performance Statistic Status'}; }
  if ( exists $data->{device}{'Performance Statistic Frequency'} ) { $conf{ $cfg->{'hostcfg_id'} }{'performance_statistic_frequency'} = $data->{device}{'Performance Statistic Frequency'}; }
  if ( exists $data->{device}{'Date'} )                            { $conf{ $cfg->{'hostcfg_id'} }{'last_conf_date'}                  = $data->{device}{'Date'}; } # 17.04.2023 12:25:00
  if ( exists $data->{device}{'Space in Pools'} )                  { $conf{ $cfg->{'hostcfg_id'} }{'capacity_pools'}                  = format_cap_unit($data->{device}{'Space in Pools'}); } # 1.6TB
  if ( exists $data->{device}{'Space Allocated to Volumes'} )      { $conf{ $cfg->{'hostcfg_id'} }{'volumes_allocated_capacity'}      = format_cap_unit($data->{device}{'Space Allocated to Volumes'}); }
  if ( exists $data->{device}{'Total Overallocation'} )            { $conf{ $cfg->{'hostcfg_id'} }{'total_overallocation'}            = format_cap_unit($data->{device}{'Total Overallocation'}); }
  if ( exists $data->{device}{'Total Mdisk Capacity'} )            { $conf{ $cfg->{'hostcfg_id'} }{'total_mdisk_capacity'}            = format_cap_unit($data->{device}{'Total Mdisk Capacity'}); }
  if ( exists $data->{device}{'Total Volumecopy Capacity'} )       { $conf{ $cfg->{'hostcfg_id'} }{'total_volumecopy_capacity'}       = format_cap_unit($data->{device}{'Total Volumecopy Capacity'}); }
  if ( exists $data->{device}{'Total Free_space'} )                { $conf{ $cfg->{'hostcfg_id'} }{'total_free_capacity'}             = format_cap_unit($data->{device}{'Total Free_space'}); }
  if ( exists $data->{device}{'Total Allocated Extent Capacity'} ) { $conf{ $cfg->{'hostcfg_id'} }{'total_allocated_extent_capacity'} = format_cap_unit($data->{device}{'Total Allocated Extent Capacity'}); }
  if ( exists $data->{device}{'Total Used_capacity'} )             { $conf{ $cfg->{'hostcfg_id'} }{'total_used_capacity'}             = format_cap_unit($data->{device}{'Total Used_capacity'}); }
  if ( exists $data->{device}{'Total Volume Capacity'} )           { $conf{ $cfg->{'hostcfg_id'} }{'total_volume_capacity'}           = format_cap_unit($data->{device}{'Total Volume Capacity'}); }
  if ( exists $cfg->{'host'} )                                     { $conf{ $cfg->{'hostcfg_id'} }{'ip'}                              = $cfg->{'host'}; }
  # add node names
  my @nodes = ();
  if ( exists $data->{'cpu-node'} ) {
    foreach ( sort keys %{ $data->{'cpu-node'} } ) {
      my $var = $data->{'cpu-node'}{$_};
      if ( exists $var->{id} && exists $var->{name} ) {
        push(@nodes, $var->{name});
      }
    }
  }
  if ( scalar(@nodes) > 0 ) { $conf{ $cfg->{'hostcfg_id'} }{node_name} = join(", ", @nodes); }

  #
  # POOL
  #
  if ( exists $data->{pool} ) {
    foreach ( sort keys %{ $data->{pool} } ) {
      my $var = $data->{pool}{$_};
      if ( exists $var->{id} && exists $var->{name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$var->{id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "swiz",
          "subsystem"  => "pool",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # conf
        #if ( exists $var->{'vdisk_count'} )        { $conf{$id}{'vdisk_count'}        = $var->{'vdisk_count'}; }
        #if ( exists $var->{'mdisk_count'} )        { $conf{$id}{'mdisk_count'}        = $var->{'mdisk_count'}; }
        #if ( exists $var->{'status'} )             { $conf{$id}{'status'}             = $var->{'status'}; }
        #if ( exists $var->{'compression_active'} ) { $conf{$id}{'compression_active'} = $var->{'compression_active'}; }
        #if ( exists $var->{'extent_size'} )        { $conf{$id}{'extent_size'}        = $var->{'extent_size'}; }
        #if ( exists $var->{'warning'} )            { $conf{$id}{'warning'}            = $var->{'warning'}; }
        #if ( exists $var->{'easy_tier'} )          { $conf{$id}{'easy_tier'}          = $var->{'easy_tier'}; }
        #if ( exists $var->{'easy_tier_status'} )   { $conf{$id}{'easy_tier_status'}   = $var->{'easy_tier_status'}; }

        #if ( exists $var->{'capacity'} )           { $conf{$id}{'capacity_total'}     = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'capacity'},"MB")); }
        #if ( exists $var->{'free_capacity'} )      { $conf{$id}{'capacity_free'}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'free_capacity'},"MB")); }
        #if ( exists $var->{'virtual_capacity'} )   { $conf{$id}{'capacity_virtual'}   = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'virtual_capacity'},"MB")); }
        #if ( exists $var->{'used_capacity'} )      { $conf{$id}{'capacity_used'}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'used_capacity'},"MB")); }
        #if ( exists $var->{'real_capacity'} )      { $conf{$id}{'capacity_real'}      = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'real_capacity'},"MB")); }

        push(@architecture, \%arch);
      }
    }
  }

  #
  # VOLUME
  #
  if ( exists $data->{volume} ) {
    foreach ( sort keys %{ $data->{volume} } ) {
      my $var = $data->{volume}{$_};
      if ( exists $var->{vdisk_UID} && exists $var->{name} ) {
        my $id = lc( $var->{vdisk_UID} );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "swiz",
          "subsystem"  => "volume",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # WWN for mapping
        $arch{wwn} = Xormon::formatWWN($var->{'vdisk_UID'});
        # conf
        $conf{$id}{'lun_id'} = $var->{'vdisk_UID'};
        #if ( exists $var->{'id'} )                    { $conf{$id}{'id'}                    = $var->{'id'}; }
        #if ( exists $var->{'fc_map_count'} )          { $conf{$id}{'fc_map_count'}          = $var->{'fc_map_count'}; }
        #if ( exists $var->{'RC_change'} )             { $conf{$id}{'rc_change'}             = $var->{'RC_change'}; }
        #if ( exists $var->{'status'} )                { $conf{$id}{'status'}                = $var->{'status'}; }
        #if ( exists $var->{'IO_group_id'} )           { $conf{$id}{'io_group_id'}           = $var->{'IO_group_id'}; }
        #if ( exists $var->{'RC_name'} )               { $conf{$id}{'rc_name'}               = $var->{'RC_name'}; }
        #if ( exists $var->{'se_copy_count'} )         { $conf{$id}{'se_copy_count'}         = $var->{'se_copy_count'}; }
        #if ( exists $var->{'IO_group_name'} )         { $conf{$id}{'io_group_name'}         = $var->{'IO_group_name'}; }
        #if ( exists $var->{'FC_id'} )                 { $conf{$id}{'fc_id'}                 = $var->{'FC_id'}; }
        #if ( exists $var->{'FC_name'} )               { $conf{$id}{'fc_name'}               = $var->{'FC_name'}; }
        #if ( exists $var->{'fast_write_state'} )      { $conf{$id}{'fast_write_state'}      = $var->{'fast_write_state'}; }
        #if ( exists $var->{'capacity'} )              { $conf{$id}{'capacity'}              = $var->{'capacity'}; }
        #if ( exists $var->{'compressed_copy_count'} ) { $conf{$id}{'compressed_copy_count'} = $var->{'compressed_copy_count'}; }
        #if ( exists $var->{'RC_id'} )                 { $conf{$id}{'rc_id'}                 = $var->{'RC_id'}; }
        #if ( exists $var->{'copy_count'} )            { $conf{$id}{'copy_count'}            = $var->{'copy_count'}; }
        #if ( exists $var->{'type'} )                  { $conf{$id}{'type'}                  = $var->{'type'}; }
        # there can be 2 or more pools separated by a space
        #if ( exists $var->{'mdisk_grp_name'} ) {
        #  $conf{$id}{'pool_name'} = $var->{'mdisk_grp_name'}; # there can be more pools separated by a space
        #}
        if ( exists $var->{'mdisk_grp_id'} ) {
          #$conf{$id}{'pool_id'} = $var->{'mdisk_grp_id'}; # there can be more pools separated by a space
          if ( $var->{'mdisk_grp_id'} =~ m/\s+/ ) {
            foreach my $pool_id ( split(/\s+/, $var->{'mdisk_grp_id'}) ) {
              push(@{ $arch{parents} }, lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$pool_id) ));
            }
          }
          else {
            push(@{ $arch{parents} }, lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$var->{'mdisk_grp_id'}) ));
          }
        }
        if ( exists $var->{'volume_id'} ) {
          #$conf{$id}{'volume_id'} = $var->{'volume_id'};
          # translator
          $translator->{volume}{ $var->{'volume_id'} } = $id; # find this uuid by volume_id in the perf processing
        }
        # assign volume to IO group
        if ( exists $var->{'IO_group_id'} && exists $var->{'IO_group_name'} ) {
          my $io_id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."io_group".$var->{'IO_group_id'}) );
          $io_group->{$io_id}{label} = $var->{'IO_group_name'};
          $io_group->{$io_id}{id}    = $var->{'IO_group_id'};
          #push(@{ $conf{$id}{'parents'} }, $io_id);
          push(@{ $arch{parents} }, $io_id);
        }

        #push(@architecture, \%arch); # arch for volumes must be after io_group, bcs parents must be stored firts
        push(@volumes_arch, \%arch);
      }
    }
  }

  #
  # IO_GROUP
  #
  foreach my $io_group_id ( keys %{ $io_group } ) {
    # Architecture
    my %arch = (
      "item_id"    => $io_group_id,
      "label"      => $io_group->{$io_group_id}{label},
      "hw_type"    => "swiz",
      "subsystem"  => "io_group",
      "hostcfg_id" => $cfg->{'hostcfg_id'},
      "parents"    => [$cfg->{'hostcfg_id'}]
    );
    # conf
    #$conf{$io_group_id}{'io_group_id'} = $io_group->{$io_group_id}{id};

    push(@architecture, \%arch);
  }

  #
  # add arch for volumes
  # arch for volumes must be after io_group, bcs parents must be stored firts
  if ( scalar(@volumes_arch) > 0 ) {
    push(@architecture, @volumes_arch);
  }

  #
  # NODE (CPU-NODE)
  #
  if ( exists $data->{'cpu-node'} ) {
    foreach ( sort keys %{ $data->{'cpu-node'} } ) {
      my $var = $data->{'cpu-node'}{$_};
      #print Dumper $var;
      if ( exists $var->{id} && exists $var->{name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."node".$var->{id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "swiz",
          "subsystem"  => "node",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # translator
        $translator->{'cpu-node'}{ $var->{'name'} } = $id; # find this id in the perf processing
        # conf
        #if ( exists $var->{'IO_group_name'} )           { $conf{$id}{'io_group_name'} = $var->{'IO_group_name'}; }
        #if ( exists $var->{'enclosure_serial_number'} ) { $conf{$id}{'serial'}        = $var->{'enclosure_serial_number'}; }
        #if ( exists $var->{'status'} )                  { $conf{$id}{'status'}        = $var->{'status'}; }
        if ( exists $var->{'WWNN'} ) {
          #$conf{$id}{'wwn'} = $var->{'WWNN'};
          # WWN for mapping
          $arch{wwn} = Xormon::formatWWN($var->{'WWNN'});
        }

        #'product_mtm' => '6099-24C',
        #'config_node' => 'yes',
        #'IO_group_id' => '0',
        #'iscsi_name' => 'iqn.1986-03.com.ibm:2145.cluster01.node1',
        #'canister_id' => '1',
        #'panel_name' => '01-1',
        #'hardware' => 'TB4',
        #'enclosure_id' => '1',

        push(@architecture, \%arch);
      }
    }
  }

  #
  # PORT
  #
  # only FC ports are presented in conf files
  my %fc_ports;
  if ( exists $data->{'port'} ) {
    foreach ( sort keys %{ $data->{'port'} } ) {
      my $var = $data->{'port'}{$_};
      #print Dumper $var; next;
      if ( exists $var->{port_name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."port".$var->{port_name}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{port_name},
          "hw_type"    => "swiz",
          "subsystem"  => "port",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # WWN for mapping
        if ( exists $var->{'WWPN'} ) {
          $arch{wwn}      = Xormon::formatWWN($var->{'WWPN'});
          $conf{$id}{wwn} = $var->{'WWPN'};
        }
        # conf
        $fc_ports{$id} = $id;
        #if ( exists $var->{'status'} )        { $conf{$id}{'status'}        = $var->{'status'}; }
        #if ( exists $var->{'fc_io_port_id'} ) { $conf{$id}{'fc_io_port_id'} = $var->{'fc_io_port_id'}; }
        #if ( exists $var->{'id'} )            { $conf{$id}{'id'}            = $var->{'id'}; }
        #if ( exists $var->{'node_name'} )     { $conf{$id}{'node_name'}     = $var->{'node_name'}; }
        #if ( exists $var->{'node_id'} )       { $conf{$id}{'node_id'}       = $var->{'node_id'}; }
        #if ( exists $var->{'type'} )          { $conf{$id}{'type'}          = $var->{'type'}; }
        #if ( exists $var->{'port_id'} )       { $conf{$id}{'port_id'}       = $var->{'port_id'}; }

        #if ( exists $conf{$id}{'type'} && $conf{$id}{'type'} eq "fc" && exists $conf{$id}{'WWPN'} ) {
        if ( exists $var->{'type'} && $var->{'type'} eq "fc" && exists $var->{'WWPN'} ) {
          #$arch{label} = "$var->{port_name} (FC $conf{$id}{'WWPN'})";
          $arch{label} = "$var->{port_name} (FC $var->{'WWPN'})";
        }

        #'port_name' => 'node1_p01',
        #'port_speed' => '4Gb',
        #'attachment' => 'switch',
        #'nportid' => '750700',

        push(@architecture, \%arch);
      }
    }
  }
  # find the rest of the ports
  if ( -f "$wrkdir/ports.json" ) {
    my $ports = Xormon::read_json("$wrkdir/ports.json");
    foreach my $port_name ( keys %{ $ports } ) {
      my $id = $ports->{$port_name}{id};

      #if ( exists $conf{$id} ) { next; } # this FC port has been already added
      if ( exists $fc_ports{$id} ) { next; } # this FC port has been already added

      # Architecture
      my %arch = (
        "item_id"    => $id,
        "label"      => $port_name,
        "hw_type"    => "swiz",
        "subsystem"  => "port",
        "hostcfg_id" => $cfg->{'hostcfg_id'},
        "parents"    => [$cfg->{'hostcfg_id'}]
      );
      # conf
      if ( exists $ports->{$port_name}{port_type} ) {
        #$conf{$id}{'type'} = $ports->{$port_name}{port_type};
        $arch{label} = "$port_name ($ports->{$port_name}{port_type})";
      }

      push(@architecture, \%arch);
    }
  }

  #
  # DRIVE
  #
  if ( exists $data->{'drive'} ) {
    foreach ( sort keys %{ $data->{'drive'} } ) {
      my $var = $data->{'drive'}{$_};
      #print Dumper $var; next;
      if ( exists $var->{drive_id} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."drive".$var->{drive_id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{drive_id},
          "hw_type"    => "swiz",
          "subsystem"  => "drive",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # conf
        #if ( exists $var->{'mdisk_name'} ) { $conf{$id}{'mdisk_name'}     = $var->{'mdisk_name'}; }
        #if ( exists $var->{'mdisk_id'} )   { $conf{$id}{'mdisk_id'}       = $var->{'mdisk_id'}; }
        #if ( exists $var->{'status'} )     { $conf{$id}{'status'}         = $var->{'status'}; }
        #if ( exists $var->{'member_id'} )  { $conf{$id}{'member_id'}      = $var->{'member_id'}; }
        #if ( exists $var->{'id'} )         { $conf{$id}{'id'}             = $var->{'id'}; }
        #if ( exists $var->{'slot_id'} )    { $conf{$id}{'slot_id'}        = $var->{'slot_id'}; }
        #if ( exists $var->{'capacity'} )   { $conf{$id}{'capacity_total'} = $var->{'capacity'}; }

        #'tech_type' => 'tier_enterprise',
        #'use' => 'member',
        #'enclosure_id' => '1',

        push(@architecture, \%arch);
      }
    }
  }

  #
  # MDISK (RANK)
  #
  if ( exists $data->{'rank'} ) {
    foreach ( sort keys %{ $data->{'rank'} } ) {
      my $var = $data->{'rank'}{$_};
      #print Dumper $var; next;
      if ( exists $var->{mdisk_id} && exists $var->{name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."mdisk".$var->{mdisk_id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "swiz",
          "subsystem"  => "mdisk",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # conf
        #if ( exists $var->{'mdisk_grp_name'} ) { $conf{$id}{'mdisk_grp_name'} = $var->{'mdisk_grp_name'}; }
        #if ( exists $var->{'mdisk_grp_id'} )   { $conf{$id}{'mdisk_grp_id'}   = $var->{'mdisk_grp_id'}; }
        #if ( exists $var->{'status'} )         { $conf{$id}{'status'}         = $var->{'status'}; }
        #if ( exists $var->{'id'} )             { $conf{$id}{'id'}             = $var->{'id'}; }
        #if ( exists $var->{'tier'} )           { $conf{$id}{'tier'}           = $var->{'tier'}; }
        #if ( exists $var->{'capacity'} )       { $conf{$id}{'capacity_total'} = $var->{'capacity'}; }

        #'dedupe' => 'no',
        #'mode' => 'array',
        #'hw_compressed' => 'no',

        push(@architecture, \%arch);
      }
    }
  }

  #
  # CPU (CPU-CORE)
  #
  if ( -f "$wrkdir/cpu.json" ) {
    my $cpu = Xormon::read_json("$wrkdir/cpu.json");
    foreach my $cpu_name ( keys %{ $cpu } ) {
      my $id = $cpu->{$cpu_name}{id};

      # Architecture
      my %arch = (
        "item_id"    => $id,
        "label"      => $cpu_name,
        "hw_type"    => "swiz",
        "subsystem"  => "cpu",
        "hostcfg_id" => $cfg->{'hostcfg_id'},
        "parents"    => [$cfg->{'hostcfg_id'}]
      );
      # conf

      push(@architecture, \%arch);
    }
  }

  #
  # HOST (this info is collected only once per day)
  #
  my $update_hosts = 0;
  if ( exists $data->{'host'} ) {
    foreach ( sort keys %{ $data->{'host'} } ) {
      my $var = $data->{'host'}{$_};
      #print Dumper $var; next;
      if ( exists $var->{host_id} && exists $var->{name} ) {
        $update_hosts++
      }
    }
  }
  if ( $update_hosts > 0 ) {
    Xormon::save_as_json("$wrkdir/host.json",$data->{'host'}); # save it as json for the further purposes
  }
  if ( -f "$wrkdir/host.json" && Xormon::file_time_diff("$wrkdir/host.json") < 86400 ) {
    my %hosts = %{ Xormon::read_json("$wrkdir/host.json") };
    foreach ( sort keys %hosts ) {
      my $var = $hosts{$_};
      #print Dumper $var;
      if ( exists $var->{host_id} && exists $var->{name}) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."host".$var->{host_id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "swiz",
          "subsystem"  => "host",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # conf
        #if ( exists $var->{'status'} )         { $conf{$id}{'status'}         = $var->{'status'}; }
        #if ( exists $var->{'id'} )             { $conf{$id}{'id'}             = $var->{'id'}; }
        #'iogrp_count' => '1',
        #'IQN WWPN' => '50014380167D073C 50014380167D073E',
        #'id' => '1',
        #'host_id' => '0001',
        #'port_count' => '2',
        #'Volume Names' => 'vmware-dev01'
        #'Volume IDs' => '0002',
        if ( exists $var->{'Volume IDs'} ) {
          my %vols;
          foreach my $volume_id ( split(/\s+/, $var->{'Volume IDs'}) ) {
            if ( exists $vols{$volume_id}) {
              next; # already discovered volume, there was an issue for some devices, where children (volumes) were twice...
            }
            else {
              $vols{$volume_id} = $volume_id;
            }
            if ( exists $translator->{volume}{$volume_id} ) {
              #print "host: $var->{name} : volume_id=$volume_id --> $translator->{volume}{$volume_id}\n";
              push(@{ $arch{children} }, $translator->{volume}{$volume_id});
            }
          }
        }

        push(@architecture, \%arch);
      }
    }
  }

  #
  # tier arch from capacity file
  #
  if ( scalar(keys %tier_arch) > 0 ) {
    push(@architecture, \%tier_arch);
  }

  #save architecture
  my $rc_arch = $xormon->saveArchitecture(\@architecture);
  $logger->info("save architecture: $rc_arch");

  #save configuration
  my $rc_conf = $xormon->saveConfJSON(\%conf);
  $logger->info("save conf: $rc_conf");

  # save translator
  Xormon::save_as_json("$wrkdir/translator.txt",$translator);

  #print "#################### Architecture\n";
  #print Dumper @architecture;
  #print "#################### CONF\n";
  #print Dumper \%conf;

  return 1;
}

sub parse_conf {
  my $conf_file = shift;

  open( IN, "<$conf_file" ) || $logger->error( "Couldn't open file $conf_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  my @lines = <IN>;
  close(IN);

  my %data_in;
  my $section    = "";
  my $act_header = "";

  my $line_idx = 0;
  foreach my $line (@lines) {
    chomp $line;
    $line_idx++;

    $line =~ s/^\s+//g;
    $line =~ s/\s+$//g;

    if ( $line =~ m/^----------/ )          { next; }

    if ( $line =~ m/^$/ && $section ne "device" )        { $section = "NA";        next; }
    if ( $line =~ m/Configuration Data/ )                { $section = "device";    next; }
    if ( $line =~ m/Node Level Configuration/ )          { $section = "cpu-node";  next; }
    if ( $line =~ m/Port Level Configuration/ )          { $section = "port";      next; }
    if ( $line =~ m/Drive Level Configuration/ )         { $section = "drive";     next; }
    if ( $line =~ m/Pool Level Configuration/ )          { $section = "pool";      next; }
    if ( $line =~ m/Managed Disk Level Configuration/ )  { $section = "rank";      next; }
    if ( $line =~ m/Volume Level Configuration/ )        { $section = "volume";    next; }
    if ( $line =~ m/Host Level Configuration/ )          { $section = "host";      next; }

    # main cfg, Configuration Data
    if ( $section eq "device" ) {
      if ( $line =~ m/^Machine Name:\s+/ ) {
        $line =~ s/^Machine Name:\s+//;
        $data_in{device}{'Machine Name'} = $line;
      }
      if ( $line =~ m/^Machine Type-Model:\s+/ ) {
        $line =~ s/^Machine Type-Model:\s+//;
        $data_in{device}{'Machine Type-Model'} = $line;
      }
      if ( $line =~ m/^Machine Serial:\s+/ ) {
        $line =~ s/^Machine Serial:\s+//;
        $data_in{device}{'Machine Serial'} = $line;
      }
      if ( $line =~ m/^Code Level:\s+/ ) {
        $line =~ s/^Code Level:\s+//;
        $data_in{device}{'Code Level'} = $line;
      }
      if ( $line =~ m/^Product Name:\s+/ ) {
        $line =~ s/^Product Name:\s+//;
        $data_in{device}{'Product Name'} = $line;
      }
      if ( $line =~ m/^Performance Statistic Status:\s+/ ) {
        $line =~ s/^Performance Statistic Status:\s+//;
        $data_in{device}{'Performance Statistic Status'} = $line;
      }
      if ( $line =~ m/^Performance Statistic Frequency:\s+/ ) {
        $line =~ s/^Performance Statistic Frequency:\s+//;
        $data_in{device}{'Performance Statistic Frequency'} = $line;
      }
      if ( $line =~ m/^Date:\s+/ ) {
        $line =~ s/^Date:\s+//;
        $data_in{device}{'Date'} = $line;
      }
      if ( $line =~ m/^Total Mdisk Capacity:\s+/ ) {
        $line =~ s/^Total Mdisk Capacity:\s+//;
        $data_in{device}{'Total Mdisk Capacity'} = $line;
      }
      if ( $line =~ m/^Space in Pools:\s+/ ) {
        $line =~ s/^Space in Pools:\s+//;
        $data_in{device}{'Space in Pools'} = $line;
      }
      if ( $line =~ m/^Space Allocated to Volumes:\s+/ ) {
        $line =~ s/^Space Allocated to Volumes:\s+//;
        $data_in{device}{'Space Allocated to Volumes'} = $line;
      }
      if ( $line =~ m/^Total Free_space:\s+/ ) {
        $line =~ s/^Total Free_space:\s+//;
        $data_in{device}{'Total Free_space'} = $line;
      }
      if ( $line =~ m/^Total Used_capacity:\s+/ ) {
        $line =~ s/^Total Used_capacity:\s+//;
        $data_in{device}{'Total Used_capacity'} = $line;
      }
      if ( $line =~ m/^Total Overallocation:\s+/ ) {
        $line =~ s/^Total Overallocation:\s+//;
        $data_in{device}{'Total Overallocation'} = $line;
      }
      if ( $line =~ m/^Total Volume Capacity:\s+/ ) {
        $line =~ s/^Total Volume Capacity:\s+//;
        $data_in{device}{'Total Volume Capacity'} = $line;
      }
      if ( $line =~ m/^Total Volumecopy Capacity:\s+/ ) {
        $line =~ s/^Total Volumecopy Capacity:\s+//;
        $data_in{device}{'Total Volumecopy Capacity'} = $line;
      }
      if ( $line =~ m/^Total Allocated Extent Capacity:\s+/ ) {
        $line =~ s/^Total Allocated Extent Capacity:\s+//;
        $data_in{device}{'Total Allocated Extent Capacity'} = $line;
      }
    }

    # Node Level Configuration
    if ( $section eq "cpu-node" ) {
      # parse header
      if ( $line =~ m/^id,name,UPS_serial_number,/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Ports Level Configuration
    if ( $section eq "port" ) {
      # parse header
      if ( $line =~ m/^port_name,id,fc_io_port_id,port_id/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Drive Level Configuration
    if ( $section eq "drive" ) {
      # parse header
      if ( $line =~ m/^drive_id,id,status,error_sequence_number/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Pool Level Configuration
    if ( $section eq "pool" ) {
      # parse header
      if ( $line =~ m/^name,id,status,mdisk_count/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Managed Disk Level Configuration
    if ( $section eq "rank" ) {
      # parse header
      if ( $line =~ m/^mdisk_id,id,name,status/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Volume Level Configuration
    if ( $section eq "volume" ) {
      # parse header
      if ( $line =~ m/^volume_id,id,name,IO_group_id/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Host Level Configuration
    if ( $section eq "host" ) {
      # parse header
      if ( $line =~ m/^host_id,id,name,port_count/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }
  }
  #print Dumper \%data_in;

  return \%data_in;
}

sub remove_file {
  my $file = shift;

  # do not remove perf and conf files for debug purposes
  if ( exists $ENV{KEEP_OUT_FILES} && ( $ENV{KEEP_OUT_FILES} eq "1" || $ENV{KEEP_OUT_FILES} eq "$cfg->{'hostalias'}" ) ) {
    return 1;
  }

  if ( -f $file ) {
    unlink($file) || $logger->error( "Couldn't remove file $file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
    $logger->info("remove file: $file");
  }

  return 1;
}

sub format_cap_unit {
  my $value = shift;

  if ( defined $value ) {
    $value =~ s/PB/ PiB/;
    $value =~ s/TB/ TiB/;
    $value =~ s/GB/ GiB/;
    $value =~ s/MB/ MiB/;
  }

  return $value;
}

1;
