#!/usr/bin/env bash
#
# XorMon NG build helper script
#

BASEDIR=$PWD
ERRLOG=$BASEDIR/error.log

# list of commands available in production environment (see NODE_ENV in server-nest/.env)
PRODUCTION_CMDS="start upgrade_start stop restart status ver logs upgrade password mspatch"

RED="\e[1;91m"
ENDCOLOR="\e[0m"

if [ -f "$ERRLOG" ]; then
  ERR_START=$(grep -vc nohup "$ERRLOG")
else
  ERR_START=0
fi

if [ "$EUID" -eq 0 ]
  then
    read -rp "WARNING: You're running the script as root, do you want to proceed? (y/n) " yn
    case $yn in
      y ) echo ok, we will proceed;;
      n ) echo exiting...;
        exit;;
      * ) echo invalid response;
        exit 1;;
    esac
fi

#UV_THREADPOOL_SIZE=12

if [ -f server-nest/.env ]; then
  . server-nest/.env
fi


build_be()
{
  echo "============================================================================="
  echo "BUILDING BACKEND..."
  rm -rf "$BASEDIR"/server-nest/dist/*
  cd "$BASEDIR"/server-nest || exit
  npm ci && npx nest build
  git describe > "$BASEDIR/server-nest/logs/git_version.txt"
  echo "BUILDING BACKEND FINISHED"
  echo "============================================================================="
  echo
}

rebuild_be()
{
  echo "============================================================================="
  echo "REBUILDING BACKEND..."
  rm -rf "$BASEDIR"/server-nest/dist/*
  cd "$BASEDIR"/server-nest || exit
  npx nest build
  git describe > "$BASEDIR/server-nest/logs/git_version.txt"
  echo "REBUILDING BACKEND FINISHED"
  echo "============================================================================="
  echo "PM2 STARTING UP..."
  npm run start:prod
}

build_fe()
{
  echo "============================================================================="
  echo "BUILDING FRONTEND..."
  cd "$BASEDIR"/ui-react || exit
  npm ci && CI='' npm run build
  rm -rf ../server-nest/client/
  cp -a build/. ../server-nest/client/
  echo "BUILDING FRONTEND FINISHED"
  echo "============================================================================="
  echo
}

migration()
{
  echo "============================================================================="
  echo "MIGRATION..."
  cd "$BASEDIR"/server-nest && npm run typeorm:migration:run
  echo "MIGRATION FINISHED"
  echo "============================================================================="
  echo
}

versions()
{
  echo "Versions"
  echo "--------------------"
  echo "Backend  : $(grep '"version"' ./server-nest/package.json | awk '{gsub(/"|,/, "", $2); print $2}')"
  echo "Frontend : $(grep '"version"' ./ui-react/package.json | awk '{gsub(/"|,/, "", $2); print $2}')"
  echo "Node.js  : $(node -v)"
  echo "NPM      : $(npm -v  2>/dev/null)"
  echo
}

start_pm2()
{
  if [[ $# -gt 0 ]]
  then
    echo "============================================================================="
    echo "STARTING UP PROCESS ID $1..."
    cd "$BASEDIR"/server-nest && npx pm2 start $1
  else
    echo "============================================================================="
    echo "PM2 STARTING UP..."
    cd "$BASEDIR"/server-nest && npm run start:prod
  fi
}

stop_pm2()
{
  if [[ $# -gt 0 ]]
  then
    echo "============================================================================="
    echo "SHUTTING DOWN PROCESS ID $1..."
    cd "$BASEDIR"/server-nest && npx pm2 stop $1

  else
    echo "============================================================================="
    echo "PM2 SHUTTING DOWN..."
    cd "$BASEDIR"/server-nest && npx pm2 stop all
  fi
}

upgrade_start()
{
  echo "============================================================================="
  echo "STARTING AFTER UPGRADE..."
  cd "$BASEDIR"/server-nest && npm run start:prod
}

usage()
{
  echo
  printf 'usage: %s [arguments] \n' "$(basename "$0")" >&2
  if [[ $NODE_ENV && $NODE_ENV == "production" ]]; then
    echo "  start [ID]                start all processes (or just ID)"
    echo "  stop [ID]                 stop all processes (or just ID)"
    echo "  restart [ID]              restart application (or just ID)"
    echo "  status                    process list"
    echo "  ver                       print versions"
    echo "  logs                      backend logs"
    echo "  upgrade                   install new version from package"
    echo "  mspatch                   patch microservice code from package"
  else
    echo "  all                       build all"
    echo "  be                        build backend"
    echo "  fe                        build frontend"
    echo "  ver                       print versions"
    echo "  logs                      backend logs"
    echo "  migration                 run DB migrations"
    echo "  start [ID]                npm run start all (or just ID)"
    echo "  stop [ID]                 stop all PM2 processes (or just ID)"
    echo "  rebuild                   nest build && reset (without npm ci)"
    echo "  status                    PM2 process list"
  fi
  echo "  password [user] [pass]    change user password"
  echo
}

show_errors()
{
  if [ -f "$ERRLOG" ]; then
    ERR_END=$(grep -cv nohup "$ERRLOG")
  else
    ERR_END=0
  fi
  ERR_TOT=$((ERR_END - ERR_START))
  if [ $ERR_TOT -gt 0 ]; then
    echo -e "${RED}>>> !!!${ENDCOLOR} Some error(s) occured, check $ERRLOG and output of $0"
    echo ""
    echo "tail -$ERR_TOT $ERRLOG"
    echo "------------------------------------"
    tail -$ERR_TOT "$ERRLOG"
    echo ""
    date >> "$ERRLOG"
  fi
}


if (($# < 1))
then
  usage
  exit 2;
fi

node_version_check()
{
NODEVERSION=$(node -v | sed "s/v\([^.]*\).*/\1/")

if ((NODEVERSION < 20))
then
  echo "Node.js should be at least version 20, detected version is $NODEVERSION"
  exit 2;
fi
}

if [[ "$NODE_ENV" && "$NODE_ENV" == "production" ]]; then
  if echo "$PRODUCTION_CMDS" | grep -qvw "$1"; then
    echo
    echo "Since you are in PRODUCTION environment, you can only use the following commands:"
    usage;
    exit;
  fi
fi

case $1 in
  all)
    node_version_check
    { build_be; build_fe; cd "$BASEDIR"/server-nest && npm run start:prod; } 2> >(tee -a "$ERRLOG" >&2)
    show_errors;
    exit;
    ;;
  be)
    node_version_check
    { build_be; cd "$BASEDIR"/server-nest && npm run start:prod; } 2> >(tee -a "$ERRLOG" >&2)
    show_errors;
    exit;
    ;;
  fe)
    node_version_check
    { build_fe; cd "$BASEDIR"/server-nest && npm run start:prod; } 2> >(tee -a "$ERRLOG" >&2)
    show_errors;
    ;;
  logs)
    { cd "$BASEDIR"/server-nest; npx pm2 logs; } 2> >(tee -a "$ERRLOG" >&2)
    show_errors;
    exit;
    ;;
  migration)
    node_version_check
    migration 2> >(tee -a "$ERRLOG" >&2)
    show_errors;
    exit;
    ;;
  ver)   versions;
    ;;
  start)
    # Clean up /tmp directory
    user=$(id -un)
    if [ -n "$user" ]; then
      if [ "$user" != "root" ]; then
        find /tmp -user "$user" -delete 2>/dev/null
      fi
    fi
    start_pm2 $2;
    ;;
  stop)  stop_pm2 $2;
    ;;
  restart)
    # Clean up /tmp directory
    user=$(id -un)
    if [ -n "$user" ]; then
      if [ "$user" != "root" ]; then
        find /tmp -user "$user" -delete 2>/dev/null
      fi
    fi
    start_pm2 $2;
    ;;
  upgrade_start) upgrade_start;
    ;;
  rebuild)
    node_version_check
    rebuild_be 2> >(tee -a "$ERRLOG" >&2)
    show_errors;
    exit;
    ;;
  status) cd "$BASEDIR"/server-nest || exit;
    npx pm2 status;
    exit;
    ;;
  password)
    if (($# == 3))
    then
      echo "============================================================================="
      echo "CHANGING password for user $2 ..."
      cd "$BASEDIR"/server-nest && node dist/change-password "$2" "$3"
    else
      echo "------------------------------------"
      echo "You must specify both username and password:"
      echo "$(basename "$0") password [user] [pass]"
      echo
      exit;
    fi
    exit;
    ;;
  upgrade)
    if (($# == 2))
    then
      echo "============================================================================="
      echo "Installing upgrade ..."
      cd "$BASEDIR"/server-nest && node dist/cli-upgrade "$2"
    else
      echo "------------------------------------"
      echo "You must specify path to the upgrade package:"
      echo "$(basename "$0") upgrade [path_to_package]"
      echo
      exit;
    fi
    exit;
    ;;
  mspatch)
    if (($# == 2))
    then
      MSNAME=`basename $2`
      MSNAME=${MSNAME%.tar}
      MSNAME=${MSNAME//xormon_ms_patch-/}
      if [ ! -d "$BASEDIR/server-nest/microservices/$MSNAME" ]; then
        echo "Microservice $MSNAME not found, exiting..."
        echo
        exit;
      fi
      echo "============================================================================="
      echo "Patching microservice $MSNAME..."
      cd "$BASEDIR" && tar xvf "$2"
      echo
      echo "Restarting microservice $MSNAME..."
      start_pm2 $MSNAME;

    else
      echo "------------------------------------"
      echo "You must specify path to the patch package:"
      echo "$(basename "$0") mspatch [path_to_package]"
      echo
      exit;
    fi
    exit;
    ;;
  *)  usage;
    exit 2
    ;;
esac
